/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Functions to detect attacks on square, king or castle.
//-------------------------------------------------------------------------------------------------

// Masks for attacks of pieces(knight, king and pawn) on white king side castle 
#define BB_WKSN     ((U64)0x00000000001F3B00)
#define BB_WKSK     ((U64)0x0000000000001F1F)
#define BB_WKSP     ((U64)0x0000000000001F00)

// Masks for attacks of pieces(knight, king and pawn) on white queen side castle
#define BB_WQSN     ((U64)0x00000000007CEE00)
#define BB_WQSK     ((U64)0x0000000000007C7C)
#define BB_WQSP     ((U64)0x0000000000007C00)

// Masks for attacks of pieces(knight, king and pawn) on black king side castle
#define BB_BKSN     ((U64)0x003B1F0000000000)
#define BB_BKSK     ((U64)0x1F1F000000000000)
#define BB_BKSP     ((U64)0x001F000000000000)

// Masks for attacks of pieces(knight, king and pawn) on black queen side castle
#define BB_BQSN     ((U64)0x00EE7C0000000000)
#define BB_BQSK     ((U64)0x7C7C000000000000)
#define BB_BQSP     ((U64)0x007C000000000000)

//-------------------------------------------------------------------------------------------------
//	Verify if move made leaves king in check, or in case of castle move, the castle is under attack
//-------------------------------------------------------------------------------------------------
int is_illegal(MOVE move)
{
    assert(move);

    switch(TYPE(move))
    {
    case MT_CSBQS: return is_black_queenside_attacked();
    case MT_CSBKS: return is_black_kingside_attacked();
    case MT_CSWQS: return is_white_queenside_attacked();
    case MT_CSWKS: return is_white_kingside_attacked();
    }
	if (is_incheck(flip_color(board.side_on_move)))
		return 1;
	return 0;
}

//-------------------------------------------------------------------------------------------------
//  Verify if side is in check.
//-------------------------------------------------------------------------------------------------
int is_incheck(int color)
{
    assert(valid_color(color));
    assert(valid_square(bb_first(board.state[color].piece[KING])));
    
	return is_square_attacked(bb_first(board.state[color].piece[KING]), flip_color(color));
}

//-------------------------------------------------------------------------------------------------
//	Verify if "square" is beign attacked by side of "by_color"
//-------------------------------------------------------------------------------------------------
int	is_square_attacked(int square, int by_color)
{
	//BBIX    temp;
	//int	    idx;

    assert(valid_square(square));
    assert(valid_color(by_color));
 
	// Attacked by knight
	if (knight_moves[square] & board.state[by_color].piece[KNIGHT].u64)
		return 1;
    // Attacked by sliding piece
    if (bb_rook_attacks(square, board.occupied_squares) & (board.state[by_color].piece[QUEEN].u64 | board.state[by_color].piece[ROOK].u64))
        return 1;
    if (bb_bishop_attacks(square, board.occupied_squares) & (board.state[by_color].piece[QUEEN].u64 | board.state[by_color].piece[BISHOP].u64))
        return 1;
	// Attacked by pawn
	if (pawn_attack[by_color][square] & board.state[by_color].piece[PAWN].u64)
		return 1;
	// Attacked by king
	if (king_moves[square] & board.state[by_color].piece[KING].u64)
		return 1;
    // No attacks
	return 0;
}

//-------------------------------------------------------------------------------------------------
//  Verify if squares E1,F1,G1 are attacked by black player
//-------------------------------------------------------------------------------------------------
int	is_white_kingside_attacked(void)
{
    int     square;
    U64     pc_qr;
    U64     pc_qb;

	// Attacked by knight
	if (board.state[BLACK].piece[KNIGHT].u64 & BB_WKSN)
		return 1;
	// Attacked by rook, bishop or queen
    pc_qr = (board.state[BLACK].piece[QUEEN].u64 | board.state[BLACK].piece[ROOK].u64);
    pc_qb = (board.state[BLACK].piece[QUEEN].u64 | board.state[BLACK].piece[BISHOP].u64);
	for (square = E1; square <= G1; square++)
    {
        if (bb_rook_attacks(square, board.occupied_squares) & pc_qr)
            return 1;
        if (bb_bishop_attacks(square, board.occupied_squares) & pc_qb)
            return 1;
	}
	// Attacked by pawn
	if (board.state[BLACK].piece[PAWN].u64 & BB_WKSP)
		return 1;
	// Attacked by king
	if (board.state[BLACK].piece[KING].u64 & BB_WKSK)
		return 1;
	// No attacks
	return 0;
}

//-------------------------------------------------------------------------------------------------
//  Verify if squares C1,D1,E1 are attacked by black player
//-------------------------------------------------------------------------------------------------
int	is_white_queenside_attacked(void)
{
    int     square;
    U64     pc_qr;
    U64     pc_qb;

	// Attacked by knight
	if (board.state[BLACK].piece[KNIGHT].u64 & BB_WQSN)
		return 1;
	// Attacked by rook, bishop or queen
    pc_qr = (board.state[BLACK].piece[QUEEN].u64 | board.state[BLACK].piece[ROOK].u64);
    pc_qb = (board.state[BLACK].piece[QUEEN].u64 | board.state[BLACK].piece[BISHOP].u64);
	for (square = C1; square <= E1; square++)
    {
        if (bb_rook_attacks(square, board.occupied_squares) & pc_qr)
            return 1;
        if (bb_bishop_attacks(square, board.occupied_squares) & pc_qb)
            return 1;
	}
    // Attacked by pawn
	if (board.state[BLACK].piece[PAWN].u64 & BB_WQSP)
		return 1;
    // Attacked by king
	if (board.state[BLACK].piece[KING].u64 & BB_WQSK)
		return 1;
    // No attacks
	return 0;
}

//-------------------------------------------------------------------------------------------------
//  Verify if squares E8,F8,G8 are attacked by white player
//-------------------------------------------------------------------------------------------------
int	is_black_kingside_attacked(void)
{
    int     square;
    U64     pc_qr;
    U64     pc_qb;

    // Attacked by knight
	if (board.state[WHITE].piece[KNIGHT].u64 & BB_BKSN)
		return 1;
    // Attacked by rook, bishop or queen
    pc_qr = (board.state[WHITE].piece[QUEEN].u64 | board.state[WHITE].piece[ROOK].u64);
    pc_qb = (board.state[WHITE].piece[QUEEN].u64 | board.state[WHITE].piece[BISHOP].u64);
	for (square = E8; square <= G8; square++)
    {
        if (bb_rook_attacks(square, board.occupied_squares) & pc_qr)
            return 1;
        if (bb_bishop_attacks(square, board.occupied_squares) & pc_qb)
            return 1;
	}
    // Attacked by pawn
	if (board.state[WHITE].piece[PAWN].u64 & BB_BKSP)
		return 1;
    // Attacked by king
	if (board.state[WHITE].piece[KING].u64 & BB_BKSK)
		return 1;
    // No attacks
	return 0;
}

//-------------------------------------------------------------------------------------------------
//  Verify if squares C8,D8,E8 are attacked by white player
//-------------------------------------------------------------------------------------------------
int	is_black_queenside_attacked(void)
{
    int     square;
    U64     pc_qr;
    U64     pc_qb;

    // Attacked by knight
	if (board.state[WHITE].piece[KNIGHT].u64 & BB_BQSN)
		return 1;
    // Attacked by rook, bishop or queen
    pc_qr = (board.state[WHITE].piece[QUEEN].u64 | board.state[WHITE].piece[ROOK].u64);
    pc_qb = (board.state[WHITE].piece[QUEEN].u64 | board.state[WHITE].piece[BISHOP].u64);
	for (square = C8; square <= E8; square++)
    {
        if (bb_rook_attacks(square, board.occupied_squares) & pc_qr)
            return 1;
        if (bb_bishop_attacks(square, board.occupied_squares) & pc_qb)
            return 1;
	}
    // Attacked by pawn
	if (board.state[WHITE].piece[PAWN].u64 & BB_BQSP)
		return 1;
    // Attacked by king
	if (board.state[WHITE].piece[KING].u64 & BB_BQSK)
		return 1;
    // No attacks
	return 0;
}

//END
