/*-------------------------------------------------------------------------------
tucano is a XBoard chess playing engine developed by Alcides Schulz.
Copyright (C) 2011-present - Alcides Schulz

tucano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

tucano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Bitboard functions.
//  Note: For portability reasons, it uses the BBIX structure for bitboard functions when int type 
//        is 4 bytes and short type is 2 bytes. Otherwise we use a different approach.
//        UINT_MAX and USHRT_MAX are defined in limits.h.
//-------------------------------------------------------------------------------------------------

#if (UINT_MAX == 0xffffffff && USHRT_MAX == 0xffff)
#define USE_BBIX_STRUCT
#else
#undef USE_BBIX_STRUCT
#endif

S8      first_index_table[4][65536];
S8      last_index_table[4][65536];
U8      bit_count_table[65536];

U8      util_bb_count(U64 bb);
S8      ushort_first(U16 value);
S8      ushort_last(U16 value);

//-------------------------------------------------------------------------------------------------
//  Initializations
//-------------------------------------------------------------------------------------------------
void bb_init(void)
{
    int	    r, f;
    U64     j;
    UINT    k;

    // set/clear bits tables.
    memset(&bb_table_set, 0, sizeof(bb_table_set));
    memset(&bb_table_clr, 0, sizeof(bb_table_clr));

    for (r = 0; r < 8; r++)
    {
        for (f = 0; f < 8; f++)
        {
            bb_table_set[r * 8 + f] = (U64)1 << (63 - (r * 8 + f));
            bb_table_clr[r * 8 + f]	= ~bb_table_set[r * 8 + f];
        }
    }

    // tables for faster first/last/count
    first_index_table[0][0] = -1;
    first_index_table[1][0] = -1;
    first_index_table[2][0] = -1;
    first_index_table[3][0] = -1;
    last_index_table[0][0]  = -1;
    last_index_table[1][0]  = -1;
    last_index_table[2][0]  = -1;
    last_index_table[3][0]  = -1;

    for (k = 1; k < 65536; k++)
    {
        first_index_table[0][k] = ushort_first((U16)k);
        first_index_table[1][k] = ushort_first((U16)k) + 16;
        first_index_table[2][k] = ushort_first((U16)k) + 32;
        first_index_table[3][k] = ushort_first((U16)k) + 48;

        last_index_table[0][k]  = ushort_last((U16)k);
        last_index_table[1][k]  = ushort_last((U16)k) + 16;
        last_index_table[2][k]  = ushort_last((U16)k) + 32;
        last_index_table[3][k]  = ushort_last((U16)k) + 48;
    }

    for (j = 0; j < 65536; j++)
    {
        bit_count_table[j] = util_bb_count(j);
    }

}

//-------------------------------------------------------------------------------------------------
//  Return index of the first "1" bit of BBIX: 0 to 63, or -1 when bitboard is 0
//-------------------------------------------------------------------------------------------------
SINT bb_first(BBIX bbix)
{
#if defined(USE_BBIX_STRUCT)
    if (bbix.u32[1])
    {
        if (bbix.u16[3])
            return (SINT)first_index_table[0][bbix.u16[3]];
        else
            return (SINT)first_index_table[1][bbix.u16[2]];
    }
    else
    {
        if (bbix.u16[1])
            return (SINT)first_index_table[2][bbix.u16[1]];
        else
            return (SINT)first_index_table[3][bbix.u16[0]];
    }
#else
    if (bbix.u64 & (U64)0xFFFF000000000000)
        return first_index_table[0][(bbix.u64 & (U64)0xFFFF000000000000) >> 48];
    if (bbix.u64 & (U64)0x0000FFFF00000000)
        return first_index_table[1][(bbix.u64 & (U64)0x0000FFFF00000000) >> 32];
    if (bbix.u64 & (U64)0x00000000FFFF0000)
        return first_index_table[2][(bbix.u64 & (U64)0x00000000FFFF0000) >> 16];
    return first_index_table[3][(bbix.u64 & (U64)0x000000000000FFFF)];
#endif
}

//-------------------------------------------------------------------------------------------------
//  Return index of the last "1" bit of BBIX: 0 to 63, or -1 when bitboard is 0
//-------------------------------------------------------------------------------------------------
SINT bb_last(BBIX bbix)
{
#ifdef USE_BBIX_STRUCT
    if (bbix.u32[0])
    {
        if (bbix.u16[0])
            return (SINT)last_index_table[3][bbix.u16[0]];
        else
            return (SINT)last_index_table[2][bbix.u16[1]];
    }
    else
    {
        if (bbix.u16[2])
            return (SINT)last_index_table[1][bbix.u16[2]];
        else
            return (SINT)last_index_table[0][bbix.u16[3]];
    }
#else
    if (bbix.u64 & (U64)0x000000000000FFFF)
        return last_index_table[3][bbix.u64 & (U64)0x000000000000FFFF];
    if (bbix.u64 & (U64)0x00000000FFFF0000)
        return last_index_table[2][(bbix.u64 & (U64)0x00000000FFFF0000) >> 16];
    if (bbix.u64 & (U64)0x0000FFFF00000000)
        return last_index_table[1][(bbix.u64 & (U64)0x0000FFFF00000000) >> 32];
    return last_index_table[0][(bbix.u64 & (U64)0xFFFF000000000000) >> 48];
#endif
}

//-------------------------------------------------------------------------------------------------
//  Number of "1" bits in the BBIX
//-------------------------------------------------------------------------------------------------
UINT bb_count(BBIX bbix)
{
#ifdef USE_BBIX_STRUCT
    return (UINT)(bit_count_table[bbix.u16[0]] + 
                  bit_count_table[bbix.u16[1]] +
                  bit_count_table[bbix.u16[2]] + 
                  bit_count_table[bbix.u16[3]]);
#else
    return (UINT)(bit_count_table[(bbix.u64 & (U64)0xFFFF000000000000) >> 48] + 
                  bit_count_table[(bbix.u64 & (U64)0x0000FFFF00000000) >> 32] +
                  bit_count_table[(bbix.u64 & (U64)0x00000000FFFF0000) >> 16] + 
                  bit_count_table[(bbix.u64 & (U64)0x000000000000FFFF)]);
#endif
}

//-------------------------------------------------------------------------------------------------
//  Number of "1" bits in the bitboard
//-------------------------------------------------------------------------------------------------
U8 util_bb_count(U64 bb)
{
    U8   count = 0;

    while (bb)
    {
        count++;
        bb &= bb - 1;
    }
    return count;
}

//-------------------------------------------------------------------------------------------------
//  Return index of the first "1" bit for a short integer, or -1 when 0.
//-------------------------------------------------------------------------------------------------
S8 ushort_first(U16 value)
{
    S8  i;

    for (i = 0; i < 16; i++)
        if (value & (1 << (15 - i)))
            return i;

    return -1;
}

//-------------------------------------------------------------------------------------------------
//  Return index of the last "1" bit for a short integer, or -1 when 0.
//-------------------------------------------------------------------------------------------------
S8 ushort_last(U16 value)
{
    S8  i;

    for (i = 0; i < 16; i++)
        if (value & (1 << i))
            return (S8)(15 - i);

    return -1;
}

//End