/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Board functionality, make/undo move, state information.
//-------------------------------------------------------------------------------------------------

void    move_piece(int color, int type, int frsq, int tosq);
void    set_piece(int color, int type, int tosq);
void    remove_piece(int color, int type, int frsq);
void    move_piece_undo(int color, int type, int frsq, int tosq);
void    set_piece_undo(int color, int type, int index);
void    remove_piece_undo(int color, int type, int index);

//-------------------------------------------------------------------------------------------------
//  Setup a new game based on a FEN position.
//-------------------------------------------------------------------------------------------------
void new_game(char *fen) 
{
	int i, r, f, inc;

	memset(&board, 0, sizeof(board));
	memset(history, 0, sizeof(history));
    move_hist_init();
    tt_clear();
	clear_eval_table();

	// Pieces
	i = r = f = inc = 0;
	while(fen[i] && fen[i] != ' ') 
    {
		if (fen[i] == '/') 
        {
			if (f > 0) 
            {
				f = 0;
				r++;
			}
			i++;
			continue;
		}
		inc = 1;
		switch(fen[i]) 
        {
			case 'r': set_piece(BLACK, ROOK, r * 8 + f);   break;
			case 'n': set_piece(BLACK, KNIGHT, r * 8 + f); break;
			case 'b': set_piece(BLACK, BISHOP, r * 8 + f); break;
			case 'q': set_piece(BLACK, QUEEN, r * 8 + f);  break;
			case 'k': set_piece(BLACK, KING, r * 8 + f);   break;
			case 'p': set_piece(BLACK, PAWN, r * 8 + f);   break;
			case 'R': set_piece(WHITE, ROOK, r * 8 + f);   break;
			case 'N': set_piece(WHITE, KNIGHT, r * 8 + f); break;
			case 'B': set_piece(WHITE, BISHOP, r * 8 + f); break;
			case 'Q': set_piece(WHITE, QUEEN, r * 8 + f);  break;
			case 'K': set_piece(WHITE, KING, r * 8 + f);   break;
			case 'P': set_piece(WHITE, PAWN, r * 8 + f);   break;
			default: 
				if (fen[i] >= '1' && fen[i] <= '8') inc = fen[i] - '0'; 
				break;
		}
		f += inc;
		if (f > 7) 
        {
			r++;
			f = f - 8;
		}
		i++;
	}

	// White or black to play
	while (fen[i] && fen[++i] == ' ');
	if (fen[i] == 'w') 
		board.side_on_move = WHITE;
	if (fen[i] == 'b') 
    {
		board.side_on_move = BLACK;
		board.key ^= zkColor;
        board.pawn_key ^= zkColor;
	}

	// Castle status
	while (fen[i] && fen[++i] == ' ');
	while (fen[i] && fen[i] != ' ') 
    {
		switch(fen[i]) 
        {
			case 'K': board.state[WHITE].can_castle_ks = 1; break;
			case 'Q': board.state[WHITE].can_castle_qs = 1; break;
			case 'k': board.state[BLACK].can_castle_ks = 1; break;
			case 'q': board.state[BLACK].can_castle_qs = 1; break;
		}
		i++;
	}
	board.key ^= zkCastleKS[board.side_on_move][STM_STATE.can_castle_ks];
	board.key ^= zkCastleQS[board.side_on_move][STM_STATE.can_castle_qs];

    // En-passant square
	board.ep_square.u64 = 0;
	if (fen[i] && fen[++i] == '-')
		i++;
	else 
    {
		if (fen[i] >= 'a' && fen[i] <= 'h' && fen[i+1] >= '1' && fen[i+1] <= '8')
			bb_set_rf(board.ep_square.u64, '8' - fen[i+1], fen[i] - 'a');
		i += 2;
	}
	if (board.ep_square.u64)
		board.key ^= zkEnPassant[bb_first(board.ep_square)];

	// 50 move count
	while(fen[i] && fen[i] == ' ')
		i++;

	board.fifty_move_rule = 0;
	while(fen[i] && fen[i] != ' ') 
    {
		if (fen[i] >= '0' && fen[i] <= '9')
			board.fifty_move_rule = board.fifty_move_rule * 10 + (fen[i] - '0');
		i++;
	}

    if (board.fifty_move_rule < 0 || board.fifty_move_rule > 100)
        board.fifty_move_rule = 0;

	// Setup additional state data.
	board.ply = 0;
	histply = 0;
	for (i = 0; i < 7; i++) 
    {
		board.state[WHITE].all_pieces |= board.state[WHITE].piece[i].u64;
		board.state[BLACK].all_pieces |= board.state[BLACK].piece[i].u64;
	}
	board.occupied_squares = board.state[WHITE].all_pieces | board.state[BLACK].all_pieces;
    board.empty_squares = ~board.occupied_squares;
    board.state[WHITE].king_square = bb_first(board.state[WHITE].piece[KING]);
    board.state[BLACK].king_square = bb_first(board.state[BLACK].piece[KING]);
}

//-------------------------------------------------------------------------------------------------
//  Make a move, update state and save history data.
//-------------------------------------------------------------------------------------------------
void make_move(MOVE move)
{
    int type = TYPE(move);
    int frsq = FRSQ(move);
    int tosq = TOSQ(move);
    int mvpc = MVPC(move);

	// Save history data to be used during move undo.
	history[histply].move             = move;
	history[histply].can_castle_ks    = STM_STATE.can_castle_ks;
	history[histply].can_castle_qs    = STM_STATE.can_castle_qs;
    history[histply].king_square      = STM_STATE.king_square;
	history[histply].ep_square        = board.ep_square.u64;
	history[histply].occupied_squares = board.occupied_squares;
	history[histply].board_key        = board.key;
    history[histply].pawn_key         = board.pawn_key;
	history[histply].fifty_move_rule  = board.fifty_move_rule;

    // Key
	board.key ^= zkColor;
    board.pawn_key ^= zkColor;
	board.key ^= zkCastleKS[board.side_on_move][STM_STATE.can_castle_ks];
	board.key ^= zkCastleQS[board.side_on_move][STM_STATE.can_castle_qs];

	// Clear ep square. The 2sq pawn move sets it.
	if (board.ep_square.u64) 
    {
		board.key ^= zkEnPassant[bb_first(board.ep_square)];
		board.ep_square.u64 = 0;
	}

    // Make move according its type
	switch(type) 
    {
		case MT_QUIET:
            move_piece(board.side_on_move, mvpc, frsq, tosq);
			break;
		case MT_CAPPC:
			remove_piece(flip_color(board.side_on_move), CPPC(move), tosq);
			move_piece(board.side_on_move, mvpc, frsq, tosq);
			break;
		case MT_EPCAP:
			remove_piece(flip_color(board.side_on_move), PAWN, PWSQ(move));
			move_piece(board.side_on_move, PAWN, frsq, tosq);
			break;
		case MT_PAWN2:
			move_piece(board.side_on_move, PAWN, frsq, tosq);
			bb_set(board.ep_square.u64, EPSQ(move));
			board.key ^= zkEnPassant[bb_first(board.ep_square)];
			break;
		case MT_PROMO:
            remove_piece(board.side_on_move, PAWN, frsq);
			set_piece(board.side_on_move, PROM(move), tosq);
			break;
		case MT_CPPRM:
            remove_piece(board.side_on_move, PAWN, frsq);
			remove_piece(flip_color(board.side_on_move), CPPC(move), tosq);
			set_piece(board.side_on_move, PROM(move), tosq);
			break;
		case MT_CSWKS:
            move_piece(WHITE, KING, E1, G1);
            move_piece(WHITE, ROOK, H1, F1);
            break;
		case MT_CSWQS:
            move_piece(WHITE, KING, E1, C1);
            move_piece(WHITE, ROOK, A1, D1);
            break;
		case MT_CSBKS:
            move_piece(BLACK, KING, E8, G8);
            move_piece(BLACK, ROOK, H8, F8);
            break;
		case MT_CSBQS:
            move_piece(BLACK, KING, E8, C8);
            move_piece(BLACK, ROOK, A8, D8);
            break;
        case MT_NULL:
            break;
	}

    // Update state data.
	board.occupied_squares = board.state[WHITE].all_pieces | board.state[BLACK].all_pieces;
    board.empty_squares = ~board.occupied_squares;

    // Castle status and king square update.
	if (mvpc == KING) 
    {
		STM_STATE.can_castle_ks = 0;
		STM_STATE.can_castle_qs = 0;
        STM_STATE.king_square = tosq;
	}
	else
    {
		if (mvpc == ROOK) 
        {
			if ((board.side_on_move == WHITE && frsq == A1) || (board.side_on_move == BLACK && frsq == A8))
				STM_STATE.can_castle_qs = 0;
			if ((board.side_on_move == WHITE && frsq == H1) || (board.side_on_move == BLACK && frsq == H8))
				STM_STATE.can_castle_ks = 0;
		}
    }

	// fifty move rule
	if (type == MT_CAPPC || mvpc == PAWN)
		board.fifty_move_rule = 0;
	else
		board.fifty_move_rule++;

    position[board.ply].curr_move = move;

	board.ply++;
	histply++;

	board.side_on_move = flip_color(board.side_on_move);

	board.key ^= zkCastleKS[board.side_on_move][STM_STATE.can_castle_ks];
	board.key ^= zkCastleQS[board.side_on_move][STM_STATE.can_castle_qs];

    assert(zk_board_key() == board.key);
    assert(board_state_is_ok());
}

//-------------------------------------------------------------------------------------------------
//  Undo last move based and restore state based on history.
//-------------------------------------------------------------------------------------------------
void undo_move(void)
{
    int     frsq;
    int     tosq;
	
	board.ply--;
	histply--;

	// Restore board state.
	board.side_on_move          = flip_color(board.side_on_move);
	STM_STATE.can_castle_ks     = history[histply].can_castle_ks;
	STM_STATE.can_castle_qs     = history[histply].can_castle_qs;
    STM_STATE.king_square       = history[histply].king_square;
	board.ep_square.u64         = history[histply].ep_square;
	board.occupied_squares      = history[histply].occupied_squares;
	board.empty_squares         = ~board.occupied_squares;
	board.key                   = history[histply].board_key;
    board.pawn_key              = history[histply].pawn_key;
	board.fifty_move_rule       = history[histply].fifty_move_rule;

    // Undo move
    frsq = FRSQ(history[histply].move);
    tosq = TOSQ(history[histply].move);

	switch(TYPE(history[histply].move))
    {
		case MT_QUIET:
            move_piece_undo(board.side_on_move, STM_STATE.square[tosq], tosq, frsq);
			break;
		case MT_CAPPC:
            move_piece_undo(board.side_on_move, STM_STATE.square[tosq], tosq, frsq);
			set_piece_undo(flip_color(board.side_on_move), CPPC(history[histply].move), tosq);
			break;
		case MT_EPCAP:
            move_piece_undo(board.side_on_move, PAWN, tosq, frsq);
			set_piece_undo(flip_color(board.side_on_move), PAWN, PWSQ(history[histply].move));
			break;
		case MT_PAWN2:
            move_piece_undo(board.side_on_move, PAWN, tosq, frsq);
			break;
		case MT_PROMO:
			remove_piece_undo(board.side_on_move, PROM(history[histply].move), tosq);
			set_piece_undo(board.side_on_move, PAWN, frsq);
			break;
		case MT_CPPRM:
			remove_piece_undo(board.side_on_move, PROM(history[histply].move), tosq);
			set_piece_undo(board.side_on_move, PAWN, frsq);
			set_piece_undo(flip_color(board.side_on_move), CPPC(history[histply].move), tosq);
			break;
		case MT_CSWKS:
            move_piece_undo(WHITE, KING, G1, E1);
            move_piece_undo(WHITE, ROOK, F1, H1);
			break;
		case MT_CSWQS:
            move_piece_undo(WHITE, KING, C1, E1);
            move_piece_undo(WHITE, ROOK, D1, A1);
			break;
		case MT_CSBKS:
            move_piece_undo(BLACK, KING, G8, E8);
            move_piece_undo(BLACK, ROOK, F8, H8);
			break;
		case MT_CSBQS:
            move_piece_undo(BLACK, KING, C8, E8);
            move_piece_undo(BLACK, ROOK, D8, A8);
			break;
		case MT_NULL:
			break;
	}
    assert(board_state_is_ok());
}

//-------------------------------------------------------------------------------------------------
//  Return the full material value for side of color.
//-------------------------------------------------------------------------------------------------
int material_value(int color)
{
    return (bb_count(board.state[color].piece[QUEEN])  * VALUE_QUEEN) +
           (bb_count(board.state[color].piece[ROOK])   * VALUE_ROOK) +
           (bb_count(board.state[color].piece[KNIGHT]) * VALUE_KNIGHT) +
           (bb_count(board.state[color].piece[BISHOP]) * VALUE_BISHOP) +
           (bb_count(board.state[color].piece[PAWN])   * VALUE_PAWN);
}

//-------------------------------------------------------------------------------------------------
//  Return the number of pieces for side of color.
//-------------------------------------------------------------------------------------------------
int pieces_count(int color)
{
    return bb_count(board.state[color].piece[QUEEN])  +
           bb_count(board.state[color].piece[ROOK])   +
           bb_count(board.state[color].piece[KNIGHT]) +
           bb_count(board.state[color].piece[BISHOP]);
}

//-------------------------------------------------------------------------------------------------
//  Check if current board is draw.
//-------------------------------------------------------------------------------------------------
int is_draw(void)
{
	int i, reps = 0;

    if (board.fifty_move_rule < 3)
		return 0;

	for (i = histply -  board.fifty_move_rule; i < histply; i++)
		if (history[i].board_key == board.key)
			reps++;

	if (reps >= 2)
		return 1;

    if (board.fifty_move_rule >= 100)
        return 1;

   	if (insufficient_material())
		return 1;

	return 0;
}

//-------------------------------------------------------------------------------------------------
//  Check if there's enough pieces for mate
//-------------------------------------------------------------------------------------------------
int insufficient_material(void) 
{
    if (material_value(WHITE) != 0 || material_value(BLACK) != 0)
		return 0;
    else
	    return 1;
}

//-------------------------------------------------------------------------------------------------
//  Move a piece and update color state.
//-------------------------------------------------------------------------------------------------
void move_piece(int color, int type, int frsq, int tosq)
{
    board.state[color].square[frsq] = 0;
    bb_clear(board.state[color].piece[type].u64, frsq);
    bb_clear(board.state[color].all_pieces, frsq);
    board.key ^= zkBase[color][type][frsq];
    board.state[color].square[tosq] = type; 
    bb_set(board.state[color].piece[type].u64, tosq); 
    bb_set(board.state[color].all_pieces, tosq);
    board.key ^= zkBase[color][type][tosq];
    if (type == PAWN)
    {
        board.pawn_key ^= zkBase[color][PAWN][frsq];
        board.pawn_key ^= zkBase[color][PAWN][tosq];
    }
}

//-------------------------------------------------------------------------------------------------
//  Place a piece on square and update color state.
//-------------------------------------------------------------------------------------------------
void set_piece(int color, int type, int tosq)
{
    board.state[color].square[tosq] = type; 
    bb_set(board.state[color].piece[type].u64, tosq); 
    bb_set(board.state[color].all_pieces, tosq);
    board.key ^= zkBase[color][type][tosq];
    if (type == PAWN)
    {
        board.pawn_key ^= zkBase[color][PAWN][tosq];
    }
}

//-------------------------------------------------------------------------------------------------
//  Remove a piece from square and update color state.
//-------------------------------------------------------------------------------------------------
void remove_piece(int color, int type, int frsq)
{
    board.state[color].square[frsq] = 0;
    bb_clear(board.state[color].piece[type].u64, frsq);
    bb_clear(board.state[color].all_pieces, frsq);
    board.key ^= zkBase[color][type][frsq];
    if (type == PAWN)
    {
        board.pawn_key ^= zkBase[color][PAWN][frsq];
    }
}

//-------------------------------------------------------------------------------------------------
//  Move a piece during undo method. Does not need to update some data.
//-------------------------------------------------------------------------------------------------
void move_piece_undo(int color, int type, int frsq, int tosq)
{
    board.state[color].square[frsq] = 0;
    bb_clear(board.state[color].piece[type].u64, frsq);
    bb_clear(board.state[color].all_pieces, frsq);
    board.state[color].square[tosq] = type; 
    bb_set(board.state[color].piece[type].u64, tosq); 
    bb_set(board.state[color].all_pieces, tosq);
}

//-------------------------------------------------------------------------------------------------
//  Place a piece on square during undo method. Does not need to update some data.
//-------------------------------------------------------------------------------------------------
void set_piece_undo(int color, int type, int index)
{
	board.state[color].square[index] = type;
	bb_set(board.state[color].piece[type].u64, index);
	bb_set(board.state[color].all_pieces, index);
}

//-------------------------------------------------------------------------------------------------
//  Remove a piece from square during undo method. Does not need to update some data.
//-------------------------------------------------------------------------------------------------
void remove_piece_undo(int color, int type, int index)
{
	board.state[color].square[index] = 0;
	bb_clear(board.state[color].piece[type].u64, index);
	bb_clear(board.state[color].all_pieces, index);
}

//-------------------------------------------------------------------------------------------------
//  Calculate the distance between 2 squares
//-------------------------------------------------------------------------------------------------
int square_distance(int square1, int square2)
{
	int		file_distance = square_file(square1) - square_file(square2);
	int		rank_distance = square_rank(square1) - square_rank(square2);

	file_distance = ABS(file_distance);
	rank_distance = ABS(rank_distance);

	return MAX(file_distance, rank_distance);
}

//end
