/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Main evaluation function
//-------------------------------------------------------------------------------------------------

void eval_material(void);
void eval_pawns(void);
void eval_passed(void);
void eval_kings(void);
void eval_pieces(void);

//-------------------------------------------------------------------------------------------------
//  Calculate the score for current position.
//-------------------------------------------------------------------------------------------------
int evaluate(void)
{
	int		opening = 0;
	int		endgame = 0;
	int		score;
    EVTB    *pet;

    //  Return score from eval table if available.
    pet = eval_table + (board.key % ET_SIZE);
    if (pet->key == board.key && !ep_on)
    {
        return pet->score;
    }

    //  Prepare eval_values.
    memset(&eval_values, 0, sizeof(eval_values));

    //  Material.
	eval_material();

    //  Position is draw.
	if (eval_values.draw_adjust[WHITE] == 0 && eval_values.draw_adjust[BLACK] == 0)
    {
        pet->key = board.key;
        pet->score = 0;
		return 0;
    }

    //  Evaluation.
	eval_pawns();
    eval_passed();
	eval_kings();
	eval_pieces();

    //  Score calculation
	opening += OPENING(eval_values.material[WHITE]) - OPENING(eval_values.material[BLACK]);
	endgame += ENDGAME(eval_values.material[WHITE]) - ENDGAME(eval_values.material[BLACK]);

	opening += OPENING(eval_values.king[WHITE]) - OPENING(eval_values.king[BLACK]);
	endgame += ENDGAME(eval_values.king[WHITE]) - ENDGAME(eval_values.king[BLACK]);

	opening += OPENING(eval_values.pawn[WHITE]) - OPENING(eval_values.pawn[BLACK]);
	endgame += ENDGAME(eval_values.pawn[WHITE]) - ENDGAME(eval_values.pawn[BLACK]);

	opening += OPENING(eval_values.passed[WHITE]) - OPENING(eval_values.passed[BLACK]);
	endgame += ENDGAME(eval_values.passed[WHITE]) - ENDGAME(eval_values.passed[BLACK]);

	opening += OPENING(eval_values.pieces[WHITE]) - OPENING(eval_values.pieces[BLACK]);
	endgame += ENDGAME(eval_values.pieces[WHITE]) - ENDGAME(eval_values.pieces[BLACK]);

	opening += OPENING(eval_values.mobility[WHITE]) - OPENING(eval_values.mobility[BLACK]);
	endgame += ENDGAME(eval_values.mobility[WHITE]) - ENDGAME(eval_values.mobility[BLACK]);

    //  Adjust score according material and draw adjustment.
    score = ((opening * (24 - eval_values.phase)) + (endgame * eval_values.phase)) / 24;
	if (score > 0)
		score = score * eval_values.draw_adjust[WHITE] / 64;
	else
		score = score * eval_values.draw_adjust[BLACK] / 64;

    //  Adjustment for side on move.
  	if (board.side_on_move == BLACK)
		score = -score;

    assert(score >= -MAX_SCORE && score <= MAX_SCORE);

    //  Save to eval table and return.
    pet->key = board.key;
    pet->score = score;

	return score;
}

//------------------------------------------------------------------------------------
//  Format score value.
//------------------------------------------------------------------------------------
void eval_print_value(double value)
{
    char    v[100];

    sprintf(v, "%3.2f", value);
    printf("%6s ", v);
}

//------------------------------------------------------------------------------------
//  Print evalulation component.
//------------------------------------------------------------------------------------
void eval_print_item(char *desc, int item[2])
{
    double  value[2][2];
    double  total[2];
    int     op[2];
    int     eg[2];

    op[WHITE] = OPENING(item[WHITE]);
    eg[WHITE] = ENDGAME(item[WHITE]);
    op[BLACK] = OPENING(item[BLACK]);
    eg[BLACK] = ENDGAME(item[BLACK]);

    value[WHITE][OP] = op[WHITE] / 100.0;
    value[WHITE][EG] = eg[BLACK] / 100.0;
    value[BLACK][OP] = op[BLACK] / 100.0;
    value[BLACK][EG] = eg[BLACK] / 100.0;

    total[OP] = value[WHITE][OP] - value[BLACK][OP];
    total[EG] = value[WHITE][EG] - value[BLACK][EG];

    printf("%-16s: ", desc);
    eval_print_value(value[WHITE][OP]);
    eval_print_value(value[WHITE][EG]);
    printf(" | ");
    eval_print_value(value[BLACK][OP]);
    eval_print_value(value[BLACK][EG]);
    printf(" | ");
    eval_print_value(total[OP]);
    eval_print_value(total[EG]);
    printf("|\n");
}

//------------------------------------------------------------------------------------
//  Calculate and print evaluation of current board position. "eval" command.
//------------------------------------------------------------------------------------
void eval_print(void) 
{
    int     score;

    ep_on = 1;
    score = evaluate();

    eval_print_item("Material", eval_values.material);
    eval_print_item("Pawn", eval_values.pawn);
    eval_print_item("Pieces", eval_values.pieces);
    eval_print_item("Mobility", eval_values.mobility);
    eval_print_item("Passed pawns", eval_values.passed);
    eval_print_item("King safety", eval_values.king);

    printf("\n%-16s: %d\n", "Final score", score);
    ep_on = 0;
}

// END