/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//------------------------------------------------------------------------------------
//  King evaluation.
//  Reference: fruit engine.
//------------------------------------------------------------------------------------

#define P_PAWN_SHIELD	MAKE_SCORE(10, 0)
#define P_PAWN_STORM1	MAKE_SCORE(10, 0)
#define P_PAWN_STORM2	MAKE_SCORE(20, 0)
#define P_PAWN_STORM3	MAKE_SCORE(30, 0)

static const int FIRST_RANK[2] = {RANK1, RANK8};
static const int LAST_RANK[2] = {RANK8, RANK1};
static const int FRONT_SQUARE[2] = {-8, +8};
static const U64 BB_FILES_QS = ((U64)0xF0F0F0F0F0F0F0F0);
static const U64 BB_FILES_KS = ((U64)0x0F0F0F0F0F0F0F0F);

int pawn_shelter_penalty(int square, int color);

//------------------------------------------------------------------------------------
//  Eval king pieces.
//------------------------------------------------------------------------------------
void eval_kings(void)
{
	int		myc;
	int		opp;
	int		pcsq;
    int     rank;
    int     file;
    U64     pawns;

    assert(eval_values.flag_king_safety[WHITE] == 0 || eval_values.flag_king_safety[WHITE] == 1);
    assert(eval_values.flag_king_safety[BLACK] == 0 || eval_values.flag_king_safety[BLACK] == 1);
	
    pawns = board.state[WHITE].piece[PAWN].u64 | board.state[BLACK].piece[PAWN].u64;

	for (myc = WHITE; myc <= BLACK; myc++)
	{
		opp = flip_color(myc);

		assert(bb_count(board.state[myc].piece[KING]) == 1);

		//pcsq = bb_first(board.state[myc].piece[KING]);
		pcsq = board.state[myc].king_square;
        rank = square_rank(pcsq);
        file = square_file(pcsq);

        assert(pcsq >= 0 && pcsq < 64);
        assert(board.state[myc].square[pcsq] == KING);
        assert(file >= 0 && file <= 7);
        assert(rank >= 0 && rank <= 7);

        // king pst (endgame pst based on pawns position)
        if (!pawns || ((pawns & BB_FILES_QS) && (pawns & BB_FILES_KS)))
        {
			eval_values.king[myc] += pst_king[myc][pcsq];
        }
        else
        {
            if (pawns & BB_FILES_QS)
            {
                eval_values.king[myc] += pst_king_qs[myc][pcsq];
            }
            else
            {
                eval_values.king[myc] += pst_king_ks[myc][pcsq];
            }
        }

        if (eval_values.flag_king_safety[myc] && rank != LAST_RANK[myc])
        {
            eval_values.king[myc] -= pawn_shelter_penalty(pcsq + FRONT_SQUARE[myc], myc);
            if (file != FILEA)
                eval_values.king[myc] -= pawn_shelter_penalty(pcsq + FRONT_SQUARE[myc] - 1, myc);
            if (file != FILEH)
                eval_values.king[myc] -= pawn_shelter_penalty(pcsq + FRONT_SQUARE[myc] + 1, myc);
        }

	}
}

//------------------------------------------------------------------------------------
//  Eval pawn protection and opponent pawn storm.
//------------------------------------------------------------------------------------
int pawn_shelter_penalty(int pcsq, int color)
{
    int     penalty = 0;
    BBIX    opp_pawn;

    assert(valid_square(pcsq));
    assert(valid_color(color));

    // pawn shield
    if (board.state[color].square[pcsq] != PAWN)
    {
        penalty += P_PAWN_SHIELD * 2;
        if (square_rank(pcsq) != LAST_RANK[color])
        {
            if (board.state[color].square[pcsq + FRONT_SQUARE[color]] != PAWN)
                penalty += P_PAWN_SHIELD;
        }
    }

    // opp pawn storm
    if (color == WHITE)
    {
        opp_pawn.u64 = rankfile_north[pcsq] & board.state[flip_color(color)].piece[PAWN].u64;
        if (opp_pawn.u64)
        {
            switch(square_rank(bb_last(opp_pawn)))
            {
            case RANK5: penalty += P_PAWN_STORM1; break;
            case RANK4: penalty += P_PAWN_STORM2; break;
            case RANK3: penalty += P_PAWN_STORM3; break;
            }
        }
    }
    else
    {
        opp_pawn.u64 = rankfile_south[pcsq] & board.state[flip_color(color)].piece[PAWN].u64;
        if (opp_pawn.u64)
        {
            switch(square_rank(bb_first(opp_pawn)))
            {
            case RANK4: penalty += P_PAWN_STORM1; break;
            case RANK5: penalty += P_PAWN_STORM2; break;
            case RANK6: penalty += P_PAWN_STORM3; break;
            }
        }
    }

    return penalty;
}

// END
