/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Material evaluation
//-------------------------------------------------------------------------------------------------

#define B_BISHOP_PAIR   MAKE_SCORE(50, 50)

//-------------------------------------------------------------------------------------------------
//  Calculate material score and draw adjustment.
//-------------------------------------------------------------------------------------------------
void eval_material(void)
{
	int		wq = bb_count(board.state[WHITE].piece[QUEEN]);
	int		wr = bb_count(board.state[WHITE].piece[ROOK]);
	int		wb = bb_count(board.state[WHITE].piece[BISHOP]);
	int		wn = bb_count(board.state[WHITE].piece[KNIGHT]);
	int		wp = bb_count(board.state[WHITE].piece[PAWN]);

	int		bq = bb_count(board.state[BLACK].piece[QUEEN]);
	int		br = bb_count(board.state[BLACK].piece[ROOK]);
	int		bb = bb_count(board.state[BLACK].piece[BISHOP]);
	int		bn = bb_count(board.state[BLACK].piece[KNIGHT]);
	int		bp = bb_count(board.state[BLACK].piece[PAWN]);

	int		wmaj;
	int		bmaj;
	int		wmin;
	int		bmin;

    //  Indicates that king safety should be evaluated.
    //  Side has queen and at least one piece.
    if (wq >= 1 && (wr + wb + wn) >= 1)
        eval_values.flag_king_safety[BLACK] = 1;
    if (bq >= 1 && (br + bb + bn) >= 1)
        eval_values.flag_king_safety[WHITE] = 1;

	eval_values.material[WHITE] = wq * SCORE_QUEEN
		                        + wr * SCORE_ROOK
								+ wb * SCORE_BISHOP
								+ wn * SCORE_KNIGHT
								+ wp * SCORE_PAWN
								+ (wb >= 2 ? B_BISHOP_PAIR : 0);
	eval_values.material[BLACK] = bq * SCORE_QUEEN
		                        + br * SCORE_ROOK
								+ bb * SCORE_BISHOP
								+ bn * SCORE_KNIGHT
								+ bp * SCORE_PAWN
								+ (bb >= 2 ? B_BISHOP_PAIR : 0);

    //  Save values.

    //  Calculate the phase of the game based on material.
    //  Opening = 0, Endgame = 24
	//  Base value is 24: 2Q * 4 + 4R * 2 + 4B * 1 + 4N * 1
	eval_values.phase = 24 - ((wq + bq) * 4 + (wr + br) * 2 + (wb + bb + wn + bn) * 1);

    //  Draw adjustment factor to reduce the material advantage in some cases.
    eval_values.draw_adjust[WHITE] = 64;
    eval_values.draw_adjust[BLACK] = 64;

	//  Don't calculate draw adjustment when in middle game or enough pawns
	if (wp > 2 || bp > 2 || eval_values.material[WHITE] > 1400 || eval_values.material[BLACK] > 1400)
	{
		return;
	}

	//  Bare kings draw.
	if (eval_values.material[WHITE] == 0 && eval_values.material[BLACK] == 0)
	{
		eval_values.draw_adjust[WHITE] = 0;
		eval_values.draw_adjust[BLACK] = 0;
		return;
	}

	//  Major and minor pieces count.
	wmin = wb + wn;
	wmaj = wq + wr;
	bmin = bb + bn;
	bmaj = bq + br;

	//  Draw ajustment for no pawns.
	if (wp == 0 && bp == 0)
	{
		//  Bishop/knight and no queen/rook.
		if (wmaj == 0 && bmaj == 0)
		{
			// Minor or 2 knights
			if (wmin == 1 || (wn == 2 && wb == 0))
			{
				eval_values.draw_adjust[WHITE] = 0;
			}
			if (bmin == 1 || (bn == 2 && bb == 0))
			{
				eval_values.draw_adjust[BLACK] = 0;
			}
			//  2 minors vs minor
            if (wn == 1 && wb == 1 && bmin == 1)
			{
				eval_values.draw_adjust[WHITE] = 8;
			}
			if (bn == 1 && bb == 1 && wmin == 1)
			{
				eval_values.draw_adjust[BLACK] = 8;
			}
			//  2 bishops vs bishop
			if (wb == 2 && wn == 0 && bb == 1 && bn == 0)
			{
				eval_values.draw_adjust[WHITE] = 8;
			}
			if (bb == 2 && bn == 0 && wb == 1 && wn == 0)
			{
				eval_values.draw_adjust[BLACK] = 8;
			}
			return;
		}

		//  Major pieces and no pawns.
		//  2 minors vs rook.
		if (wn == 1 && wb == 1 && br == 1 && bq == 0)
		{
			eval_values.draw_adjust[WHITE] = 8;
		}
		if (bn == 1 && bb == 1 && wr == 1 && wq == 0)
        {
			eval_values.draw_adjust[BLACK] = 8;
		}
		//  2 bishops vs rook.
		if (wb == 2 && wn == 0 && br == 1 && bq == 0)
		{
			eval_values.draw_adjust[WHITE] = 8;
		}
		if (bb == 2 && bn == 0 && wr == 1 && wq == 0)
		{
			eval_values.draw_adjust[BLACK] = 8;
		}
		//  Rook+minor vs rook.
		if (wr == 1 && wq == 0 && wmin == 1 && br == 1 && bq == 0 && bmin == 0)
		{
			eval_values.draw_adjust[WHITE] = 0;
		}
		if (br == 1 && bq == 0 && bmin == 1 && wr == 1 && wq == 0 && wmin == 0)
		{
			eval_values.draw_adjust[BLACK] = 0;
		}

		return;
	}

	//  End games with maximum of 2 pawns 

    //  Rook vs minor/minor+pawn
    if (wp == 0 && wmin == 0 && wr == 1 && wq == 0 && bmaj == 0 && (bmin > 0 || (bmin == 1 && bp > 0)))
    {
        eval_values.draw_adjust[WHITE] = 32;
    }
    if (bp == 0 && bmin == 0 && br == 1 && bq == 0 && wmaj == 0 && (wmin > 0 || (wmin == 1 && wp > 0)))
    {
        eval_values.draw_adjust[BLACK] = 32;
    }

    //  Rook+minor vs rook/rook+pawn(s)
    if (wp == 0 && wr == 1 && wq == 0 && wmin == 1 && br == 1 && bq == 0 && bmin == 0)
    {
        eval_values.draw_adjust[WHITE] = 32;
    }
    if (bp == 0 && br == 1 && bq == 0 && bmin == 1 && wr == 1 && wq == 0 && wmin == 0)
    {
        eval_values.draw_adjust[BLACK] = 32;
    }

    //  Queen+minor vs queen/queen+pawn(s)
    if (wp == 0 && wr == 0 && wq == 1 && wmin == 1 && br == 0 && bq == 1 && bmin == 0)
    {
        eval_values.draw_adjust[WHITE] = 32;
    }
    if (bp == 0 && br == 0 && bq == 1 && bmin == 1 && wr == 0 && wq == 1 && wmin == 0)
    {
        eval_values.draw_adjust[BLACK] = 32;
    }
}

// END
