/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Passed pawns evaluation
//-------------------------------------------------------------------------------------------------

#define MS(op, eg)	MAKE_SCORE(op, eg)

static const int B_PASSED[RANKS]    = {MS(0,0), MS(0,0), MS(4,8), MS(8,16), MS(16,32), MS(32,64), MS(64,128), MS(0,0)};
static const int B_UNBLOCKED[RANKS] = {MS(0,0), MS(0,0), MS(4,8), MS(8,20), MS(20,40), MS(40,80), MS(80,120), MS(0,0)};

#define P_KING_CLOSE_MYC	MAKE_SCORE(0,3)
#define B_KING_CLOSE_OPP	MAKE_SCORE(0,5)
#define B_ROOK_PAWN	    	MAKE_SCORE(8,8)
#define P_ROOK_PAWN 		MAKE_SCORE(12,12)
#define B_UNSTOPPABLE_PAWN	MAKE_SCORE((VALUE_QUEEN - VALUE_PAWN), (VALUE_QUEEN - VALUE_PAWN))

static const int rank_index[COLORS][RANKS] = {{7, 6, 5, 4, 3, 2, 1, 0}, {0, 1, 2, 3, 4, 5, 6, 7}};
static const int prom_base[COLORS] = {A8, A1};
static const int square_forward[COLORS] = {-8, +8};

#define EVAL_TRACE(d, s)    {printf("eval_trace: square=%d\n", s); board_print(d);}

//-------------------------------------------------------------------------------------------------
//  Evaluate passed pawns.
//-------------------------------------------------------------------------------------------------
void eval_passed(void)
{
    BBIX    pawns;
    int     pcsq;
    int     rank;
    int     file;
    int     myc;
    int     opp;
    int     prom_square;
    int     prom_distance;
    int     king_distance;
	U64 	*forward_path;
    BBIX    blockers;
    int     fwsq;
    int     king_sq[2];
    int     king_dist[2];

    for (myc = WHITE; myc <= BLACK; myc++)
    {
        opp = flip_color(myc);
        forward_path = (myc == WHITE) ? rankfile_north : rankfile_south;
        king_sq[myc] = board.state[myc].king_square;
        king_sq[opp] = board.state[opp].king_square;

        // additional bonus/penalties for passed pawns considering king/pieces.
        pawns.u64 = eval_values.bb_passers[myc];
        while(pawns.u64)
        {
            pcsq = bb_first(pawns);
            rank = square_rank(pcsq);
            file = square_file(pcsq);
            prom_square = prom_base[myc] + file;

            assert(board.state[myc].square[pcsq] == PAWN);
            assert(pawn_is_passed(pcsq, myc));
            
			eval_values.passed[myc] += B_PASSED[rank_index[myc][rank]];

            // bonus for unblocked passed pawns
            fwsq = pcsq + square_forward[myc];
            if (!(bb_square[fwsq] & board.occupied_squares))
            {
                if (see_sign(myc, pcsq, fwsq) >= 0)
                {
                    eval_values.passed[myc] += B_UNBLOCKED[rank_index[myc][square_rank(pcsq)]];
                }
            }

            // bonus/penalty according king distance to square in front
            fwsq = pcsq + square_forward[myc];
            king_dist[myc] = square_distance(king_sq[myc], fwsq);
            king_dist[opp] = square_distance(king_sq[opp], fwsq);
            if (board.side_on_move == opp)
                king_dist[myc] += 1;

			eval_values.passed[myc] -= (king_dist[myc] * P_KING_CLOSE_MYC);
			eval_values.passed[myc] += (king_dist[opp] * B_KING_CLOSE_OPP);

            // bonus/penalty for rook pawns
            if (file == FILEA || file == FILEH)
            {
                if (pieces_count(opp) == 1 && board.state[opp].piece[KNIGHT].u64)
                {
                    // opponent has only knight bonus
					eval_values.passed[myc] += B_ROOK_PAWN;
                }
                else
                {
                    if (board.state[opp].piece[QUEEN].u64 || board.state[opp].piece[ROOK].u64)
                    {
                        // opponent has queen and/or rook penalty
						eval_values.passed[myc] -= P_ROOK_PAWN;
                    }
                }
            }

            // unstoppable passers
            if (pieces_count(opp) == 0)
            {
                prom_square = prom_base[myc] + file;
                prom_distance = square_distance(pcsq, prom_square);
                if (board.side_on_move == opp)
                    prom_distance += 1;
                blockers.u64 = forward_path[pcsq] & board.state[myc].all_pieces;
                prom_distance += bb_count(blockers);
                king_distance = square_distance(bb_first(board.state[opp].piece[KING]), prom_square);
                if (king_distance > prom_distance)
                {
					eval_values.passed[myc] += B_UNSTOPPABLE_PAWN;
                }
            }

            bb_clear(pawns.u64, pcsq);
        }
    }

}

// END