/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//------------------------------------------------------------------------------------
//	Pawn structure evaluation.
//------------------------------------------------------------------------------------

#define MS(op, eg)	MAKE_SCORE(op, eg)
static const int B_CANDIDATE[RANKS] = {MS(0,0), MS(0,0), MS(1,2), MS(2,4), MS(4,8), MS(8,16), MS(16,32), MS(0,0)};
#define B_CONNECTED MAKE_SCORE(8, 10)
#define P_DOUBLED   MAKE_SCORE(8, 14)
#define P_ISOLATED  MAKE_SCORE(8, 16)
#define P_BACKWARD  MAKE_SCORE(8, 12)

static const int rank_index[COLORS][RANKS] = {{7, 6, 5, 4, 3, 2, 1, 0}, {0, 1, 2, 3, 4, 5, 6, 7}};

#define MY_PAWNS	board.state[myc].piece[PAWN].u64
#define OP_PAWNS	board.state[opp].piece[PAWN].u64

#define EVAL_TRACE(d, s)    {printf("eval_trace: square=%d\n", s); board_print(d);}

//------------------------------------------------------------------------------------
//	Evaluate pawns and indicate what pawns are passed to be evaluted later.
//------------------------------------------------------------------------------------
void eval_pawns(void)
{
	int		myc;
	int		opp;
	BBIX 	pawns;
	int		pcsq;
    int     file;
    int     rank;
	U64 	*forward_path;
	U64 	*backward_path;
    BBIX    bb_my_pawn;
    BBIX    bb_op_pawn;
	int		is_doub;
	int		is_conn;
	int		is_pass;
	int		is_weak;
	int		is_isol;
	int		is_back;
    int     is_oppo;
    int     is_cand;
    PWTB    *ppt;
	
    // Probe pawn evaluation table.
    ppt = pawn_table + (board.pawn_key % PT_SIZE);
    if (board.pawn_key != 0 && ppt->key == board.pawn_key && !ep_on)
    {
        eval_values.pawn[WHITE] = ppt->pawn_eval[WHITE];
        eval_values.pawn[BLACK] = ppt->pawn_eval[BLACK];
        eval_values.bb_passers[WHITE] = ppt->bb_passers[WHITE];
        eval_values.bb_passers[BLACK] = ppt->bb_passers[BLACK];
        return;
    }

    // Evaluate each pawn.
	for (myc = WHITE; myc <= BLACK; myc++)
	{
		opp = flip_color(myc);
		forward_path  = (myc == WHITE) ? rankfile_north : rankfile_south;
		backward_path = (myc == WHITE) ? rankfile_south : rankfile_north;

		pawns.u64 = MY_PAWNS;
		while(pawns.u64)
		{
			pcsq = bb_first(pawns);
            file = square_file(pcsq);
            rank = square_rank(pcsq);

	        assert(board.state[myc].square[pcsq] == PAWN);

			eval_values.pawn[myc] += pst_pawn[myc][pcsq];

			is_doub = (forward_path[pcsq] & MY_PAWNS) ? 1 : 0;
			is_conn = (bb_pawn_conn[pcsq] & MY_PAWNS) ? 1 : 0;
			is_pass = (!is_doub && !(bb_pawn_pass[myc][pcsq] & OP_PAWNS)) ? 1 : 0;
			is_weak = (!(bb_pawn_weak[myc][pcsq] & MY_PAWNS)) ? 1 : 0;
			is_isol = (is_weak && !(bb_pawn_isol[pcsq] & MY_PAWNS)) ? 1 : 0;
            is_oppo = (forward_path[pcsq] & OP_PAWNS) ? 1 : 0;

            // backward pawn
            is_back = 0;
            if (!is_isol && is_weak && !is_pass && !is_conn)
            {
                if (!((bb_pawn_capt[myc][pcsq] & OP_PAWNS)))
                {
                    bb_my_pawn.u64 = bb_op_pawn.u64 = 0;
                    if (file > 0)
                    {
                        bb_my_pawn.u64 |= (forward_path[pcsq - 1] & MY_PAWNS);
                        bb_op_pawn.u64 |= (forward_path[pcsq - 1] & OP_PAWNS);
                    }
                    if (file < 7)
                    {
                        bb_my_pawn.u64 |= (forward_path[pcsq + 1] & MY_PAWNS);
                        bb_op_pawn.u64 |= (forward_path[pcsq + 1] & OP_PAWNS);
                    }
                    if (bb_op_pawn.u64)
                    {
                        if (!bb_my_pawn.u64)
                        {
                            is_back = 1;
                        }
                        else
                        {
                            if (myc == WHITE && square_rank(bb_last(bb_op_pawn)) > square_rank(bb_last(bb_my_pawn)))
                            {
                                is_back = 1;
                            }
                            if (myc == BLACK && square_rank(bb_first(bb_op_pawn)) < square_rank(bb_first(bb_my_pawn)))
                            {
                                is_back = 1;
                            }
                        }
                    }
                }
            }

            // Candidate passer
            // - Not passed, not isolated, not backward, no opp pawns in front
            // - Has more own pawns behind than opp pawns in files forward
            is_cand = 0;
            if (!is_pass && !is_isol && !is_back && !is_oppo)
            {
                bb_my_pawn.u64 = bb_op_pawn.u64 = 0;
                if (file > 0)
                {
                    bb_my_pawn.u64 |= (bb_square[pcsq - 1] & MY_PAWNS);
                    bb_my_pawn.u64 |= (backward_path[pcsq - 1] & MY_PAWNS);
                    bb_op_pawn.u64 |= (forward_path[pcsq - 1] & OP_PAWNS);
                }
                if (file < 7)
                {
                    bb_my_pawn.u64 |= (bb_square[pcsq + 1] & MY_PAWNS);
                    bb_my_pawn.u64 |= (backward_path[pcsq + 1] & MY_PAWNS);
                    bb_op_pawn.u64 |= (forward_path[pcsq + 1] & OP_PAWNS);
                }
                if (bb_count(bb_op_pawn) <= bb_count(bb_my_pawn))
                {
                    is_cand = 1;
                }
            }

            // Assign bonuses and penalties.
			if (is_doub)
			{
				eval_values.pawn[myc] -= P_DOUBLED;
                assert(pawn_is_doubled(pcsq, myc));
			}
            else
                assert(!pawn_is_doubled(pcsq, myc));

			if (is_conn)
			{
				eval_values.pawn[myc] += B_CONNECTED;
                assert(pawn_is_connected(pcsq, myc));
			}
            else
                assert(!pawn_is_connected(pcsq, myc));

            if (is_pass)
            {
                bb_set(eval_values.bb_passers[myc], pcsq);
                if (is_conn)
                {
					eval_values.pawn[myc] += B_CONNECTED;
                }
                assert(pawn_is_passed(pcsq, myc));
            }
            else
                assert(!pawn_is_passed(pcsq, myc));

			if (is_isol)
			{
				eval_values.pawn[myc] -= P_ISOLATED;
                assert(pawn_is_isolated(pcsq, myc));
			}
            else
                assert(!pawn_is_isolated(pcsq, myc));

            if (is_back)
            {
				eval_values.pawn[myc] -= P_BACKWARD;
                assert(pawn_is_backward(pcsq, myc));
            }
            else
                assert(!pawn_is_backward(pcsq, myc));

            if (is_cand)
            {
				eval_values.pawn[myc] += B_CANDIDATE[rank_index[myc][rank]];
                assert(pawn_is_candidate(pcsq, myc));
            }
            else
                assert(!pawn_is_candidate(pcsq, myc));

			bb_clear(pawns.u64, pcsq);
		}
	}

    // Save information to pawn table.
    ppt->key = board.pawn_key;
    ppt->pawn_eval[WHITE] = eval_values.pawn[WHITE];
    ppt->pawn_eval[BLACK] = eval_values.pawn[BLACK];
    ppt->bb_passers[WHITE] = eval_values.bb_passers[WHITE];
    ppt->bb_passers[BLACK] = eval_values.bb_passers[BLACK];
}

// END