/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Queen, rook, bishop and knight evaluation
//-------------------------------------------------------------------------------------------------

#define B_KNIGHT_OUTPOST	MAKE_SCORE(8,8)
#define B_KNIGHT_FORK		MAKE_SCORE(10,10)

#define B_ROOK_SEMI_OPEN	MAKE_SCORE(5,12)
#define B_ROOK_FULL_OPEN	MAKE_SCORE(10,20)
#define B_ROOK_RANK_8		MAKE_SCORE(5,8)
#define B_ROOK_RANK_7		MAKE_SCORE(20,40)
#define B_ROOK_RANK_6		MAKE_SCORE(4,7)
#define P_ROOK_TRAP			MAKE_SCORE(50,0)

#define B_QUEEN_RANK_7		MAKE_SCORE(20,30)

#define P_BISHOP_TRAP		MAKE_SCORE(150,150)

#define B_QUEEN_MOBILITY	MAKE_SCORE(2,3)
#define B_ROOK_MOBILITY		MAKE_SCORE(3,4)
#define B_BISHOP_MOBILITY	MAKE_SCORE(4,5)
#define B_KNIGHT_MOBILITY	MAKE_SCORE(5,6)

#define KING_ATTACK_KNIGHT  1
#define KING_ATTACK_BISHOP  1
#define KING_ATTACK_ROOK    2
#define KING_ATTACK_QUEEN   4
static const int KING_ATTACK_MULTI[16] = {0, 0, 50, 80, 110, 140, 140, 140, 140, 140, 140, 140, 140, 140, 140, 140};

#define B_KING_ATTACK		MAKE_SCORE(25,0)

#define B_DOUBLE_RANK_7		MAKE_SCORE(5,10)
#define P_MINOR_BLOCK_PAWN	MAKE_SCORE(20,0)
#define P_BLOCKED_PAWNS		MAKE_SCORE(3,6)

#define B_KNIGHT_GUARD_KING	MAKE_SCORE(4,2)
#define B_BISHOP_GUARD_KING	MAKE_SCORE(3,1)
#define B_ROOK_GUARD_KING	MAKE_SCORE(4,2)
#define B_QUEEN_GUARD_KING	MAKE_SCORE(5,2)

#define P_PAWN_ATK_KING		MAKE_SCORE(1,1)
#define P_PAWN_ATK_KNIGHT	MAKE_SCORE(2,4)
#define P_PAWN_ATK_BISHOP	MAKE_SCORE(2,4)
#define P_PAWN_ATK_ROOK		MAKE_SCORE(3,5)
#define P_PAWN_ATK_QUEEN	MAKE_SCORE(4,6)

#define B_KNIGHT_ATK_PAWN	MAKE_SCORE(2,2)
#define B_KNIGHT_ATK_KNIGHT	MAKE_SCORE(2,2)
#define B_KNIGHT_ATK_BISHOP	MAKE_SCORE(3,3)
#define B_KNIGHT_ATK_ROOK	MAKE_SCORE(5,5)
#define B_KNIGHT_ATK_QUEEN	MAKE_SCORE(8,8)

#define B_BISHOP_ATK_PAWN	MAKE_SCORE(2,2)
#define B_BISHOP_ATK_BISHOP	MAKE_SCORE(2,2)
#define B_BISHOP_ATK_KNIGHT	MAKE_SCORE(3,3)
#define B_BISHOP_ATK_ROOK	MAKE_SCORE(6,6)
#define B_BISHOP_ATK_QUEEN	MAKE_SCORE(8,8)

#define B_ROOK_ATK_PAWN		MAKE_SCORE(2,2)
#define B_ROOK_ATK_ROOK 	MAKE_SCORE(2,2)
#define B_ROOK_ATK_KNIGHT	MAKE_SCORE(3,3)
#define B_ROOK_ATK_BISHOP	MAKE_SCORE(3,3)
#define B_ROOK_ATK_QUEEN	MAKE_SCORE(6,6)

#define B_QUEEN_ATK_PAWN	MAKE_SCORE(2,2)
#define B_QUEEN_ATK_ROOK 	MAKE_SCORE(1,1)
#define B_QUEEN_ATK_KNIGHT	MAKE_SCORE(2,2)
#define B_QUEEN_ATK_BISHOP	MAKE_SCORE(1,1)
#define B_QUEEN_ATK_QUEEN	MAKE_SCORE(2,2)

#define B_MULTIPLE_ATTACK	MAKE_SCORE(2,2)

static const int MY_RANK_6[COLORS] = {RANK6, RANK3};
static const int MY_RANK_7[COLORS] = {RANK7, RANK2};
static const int MY_RANK_8[COLORS] = {RANK8, RANK1};
static const U64 MY_BB_RANK_7[COLORS] = {BB_RANK_7, BB_RANK_2};
static const U64 MY_BB_RANK_8[COLORS] = {BB_RANK_8, BB_RANK_1};
static const U64 MY_BB_RANK_678[COLORS] = {BB_RANK_6 | BB_RANK_7 | BB_RANK_8, BB_RANK_3 | BB_RANK_2 | BB_RANK_1};
static const U64 *PATH_FORWARD[COLORS] = {rankfile_north, rankfile_south};
static const U64 BB_OUTPOST[COLORS] = {((U64)0x007E7E7E00000000), ((U64)0x000000007E7E7E00)};
static const U64 BB_BLOCK_PAWN[COLORS] = {((U64)0x0000000000180000), ((U64)0x0000180000000000)};
static const int SQUARE_BEHIND[COLORS] = {+8, -8};
static const int BISHOP_TRAP_A7[COLORS] = {A7, A2};
static const int BISHOP_TRAP_A7_PAWN[COLORS][2] = {{B6, C7}, {B3, C2}};
static const int BISHOP_TRAP_H7[COLORS] = {H7, H2};
static const int BISHOP_TRAP_H7_PAWN[COLORS][2] = {{G6, F7}, {G3, F2}};

static const int ROOK_TRAP_A1_RSQ[COLORS] = {A1, A8};
static const int ROOK_TRAP_B1_RSQ[COLORS] = {B1, B8};
static const U64 ROOK_TRAP_A1_KSQ[COLORS] = {((U64)0x0000000000000060), ((U64)0x6000000000000000)}; // B1-C1, B8-C8
static const int ROOK_TRAP_H1_RSQ[COLORS] = {H1, H8};
static const int ROOK_TRAP_G1_RSQ[COLORS] = {G1, G8};
static const U64 ROOK_TRAP_H1_KSQ[COLORS] = {((U64)0x0000000000000006), ((U64)0x0600000000000000)}; // G1-F1, G8-F8

#define EVAL_TRACE(d, s)    {printf("eval_trace: square=%d\n", s); board_print(d);}

U64     get_rankfile_mobility(int pcsq, int op_color, U64 target);
U64     get_diagonal_mobility(int pcsq, int op_color, U64 target);
void    eval_pieces_prepare(void);
void    eval_knights(int myc, int opp);
void    eval_bishops(int myc, int opp);
void    eval_rooks(int myc, int opp);
void    eval_queens(int myc, int opp);
void    eval_pieces_finalize(int myc, int opp);
void    eval_pieces_threat(U64 attacks, int piece, int myc, int opp, int bonus);
void    eval_pieces_king_attack(U64 attacks, int from, int myc, int opp, int attack_value);
void    eval_pieces_guards_king(U64 attacks, int myc, int bonus);
void    eval_pieces_attacked_by_pawn(int myc, int opp, int pcsq, int penalty);

//-------------------------------------------------------------------------------------------------
//  Evaluate pieces.
//-------------------------------------------------------------------------------------------------
void eval_pieces(void)
{
    eval_pieces_prepare();

    eval_knights(WHITE, BLACK);
    eval_knights(BLACK, WHITE);

    eval_bishops(WHITE, BLACK);
    eval_bishops(BLACK, WHITE);

    eval_rooks(WHITE, BLACK);
    eval_rooks(BLACK, WHITE);

    eval_queens(WHITE, BLACK);
    eval_queens(BLACK, WHITE);

    eval_pieces_finalize(WHITE, BLACK);
    eval_pieces_finalize(BLACK, WHITE);
}

//-------------------------------------------------------------------------------------------------
//  Initialize data used by pieces evaluation
//-------------------------------------------------------------------------------------------------
void eval_pieces_prepare(void)
{
    eval_values.pawn_attacks[WHITE] = (board.state[WHITE].piece[PAWN].u64 & BB_NO_AFILE) << 9 |
	                                  (board.state[WHITE].piece[PAWN].u64 & BB_NO_HFILE) << 7;
    eval_values.pawn_attacks[BLACK] = (board.state[BLACK].piece[PAWN].u64 & BB_NO_HFILE) >> 9 |
	                                  (board.state[BLACK].piece[PAWN].u64 & BB_NO_AFILE) >> 7;
    eval_values.strong_minors[WHITE] = (board.state[WHITE].piece[KNIGHT].u64 | board.state[WHITE].piece[BISHOP].u64)
                                     & BB_OUTPOST[WHITE] 
                                     & eval_values.pawn_attacks[WHITE]
                                     & ~eval_values.pawn_attacks[BLACK];
    eval_values.strong_minors[BLACK] = (board.state[BLACK].piece[KNIGHT].u64 | board.state[BLACK].piece[BISHOP].u64)
                                     & BB_OUTPOST[BLACK] 
                                     & eval_values.pawn_attacks[BLACK]
                                     & ~eval_values.pawn_attacks[WHITE];
    eval_values.undefended[WHITE] = board.state[WHITE].all_pieces & ~eval_values.pawn_attacks[WHITE];
    eval_values.undefended[BLACK] = board.state[BLACK].all_pieces & ~eval_values.pawn_attacks[BLACK];

    position[board.ply].threats[WHITE] = 0;
    position[board.ply].threats[BLACK] = 0;
}

//-------------------------------------------------------------------------------------------------
//  Calculate final eval terms for pieces
//-------------------------------------------------------------------------------------------------
void eval_pieces_finalize(int myc, int opp)
{
    BBIX    blocked_pawns;
    int     a;

    // pawns attacking king zone
    if (eval_values.pawn_attacks[myc] & king_moves[board.state[opp].king_square] & ~board.state[opp].piece[PAWN].u64)
    {
        eval_values.king[opp] -= P_PAWN_ATK_KING;
    }

    //  penalty for blocked pawns
    if (myc == WHITE)
        blocked_pawns.u64 = board.state[myc].piece[PAWN].u64 << 8 & board.occupied_squares;
    else
        blocked_pawns.u64 = board.state[myc].piece[PAWN].u64 >> 8 & board.occupied_squares;
    eval_values.pieces[myc] -= bb_count(blocked_pawns) * P_BLOCKED_PAWNS;

    // calculate king attack
    if (eval_values.flag_king_safety[opp])
    {
        a = (int)(eval_values.king_attack_value[myc] * 30 * KING_ATTACK_MULTI[eval_values.king_attack_count[myc]] / 100);
        eval_values.king[opp] -= MAKE_SCORE(a, 0);
        //if (eval_values.king_attack_count[myc] > 1)
        //{
        //    printf("%d %d\n", eval_values.king_attack_value[myc], eval_values.king_attack_count[myc]);
        //    printf("1. op=%d eg=%d\n", OPENING(eval_values.king[opp]), ENDGAME(eval_values.king[opp]));
        //    printf("%d %d\n", opp, a);
        //    printf("2. op=%d eg=%d\n", OPENING(eval_values.king[opp]), ENDGAME(eval_values.king[opp]));
        //    board_print("aa");
        //}
    }

    // multiple attack bonus
    position[board.ply].threats[myc] = bb_count(eval_values.threats[myc]);
    eval_values.pieces[myc] += position[board.ply].threats[myc] * B_MULTIPLE_ATTACK;
}

//-------------------------------------------------------------------------------------------------
//  Knights
//-------------------------------------------------------------------------------------------------
void eval_knights(int myc, int opp)
{
    U64     path_forward;
    BBIX    piece;
    int     pcsq;
    U64     mob_target;
    BBIX    mobility;
    U64     fork_pieces = 0;
    BBIX    forks;

    fork_pieces |= board.state[opp].piece[QUEEN].u64;
    fork_pieces |= board.state[opp].piece[ROOK].u64;
    fork_pieces |= board.state[opp].piece[BISHOP].u64 & eval_values.undefended[opp];

    piece.u64 = board.state[myc].piece[KNIGHT].u64;
    while(piece.u64)
    {
        pcsq = bb_first(piece);
        mobility.u64 = 0;

        assert(board.state[myc].square[pcsq] == KNIGHT);

        // pst
		eval_values.pieces[myc] += pst_knight[myc][pcsq];

        // mobility
        mob_target = (board.empty_squares | board.state[opp].all_pieces) & ~eval_values.pawn_attacks[opp];
        mobility.u64 = knight_moves[pcsq] & mob_target;
        eval_values.mobility[myc] += bb_count(mobility) * B_KNIGHT_MOBILITY;

        //  threats
        eval_pieces_threat(mobility.u64, PAWN,   myc, opp, B_KNIGHT_ATK_PAWN);
        eval_pieces_threat(mobility.u64, KNIGHT, myc, opp, B_KNIGHT_ATK_KNIGHT);
        eval_pieces_threat(mobility.u64, BISHOP, myc, opp, B_KNIGHT_ATK_BISHOP);
        eval_pieces_threat(mobility.u64, ROOK,   myc, opp, B_KNIGHT_ATK_ROOK);
        eval_pieces_threat(mobility.u64, QUEEN,  myc, opp, B_KNIGHT_ATK_QUEEN);

        // king attack
        eval_pieces_king_attack(mobility.u64, pcsq, myc, opp, KING_ATTACK_KNIGHT);

        //  guards own king
        eval_pieces_guards_king(mobility.u64, myc, B_KNIGHT_GUARD_KING);

        // penalty when attacked by pawn
        eval_pieces_attacked_by_pawn(myc, opp, pcsq, P_PAWN_ATK_KNIGHT);

        //  forks
        forks.u64 = mobility.u64 & fork_pieces;
        if (!(bb_square[pcsq] & eval_values.pawn_attacks[opp]) && bb_count(forks) > 1)
        {
			eval_values.pieces[myc] += B_KNIGHT_FORK;
        }

        // knight outpost
        if (bb_square[pcsq] & BB_OUTPOST[myc] & eval_values.pawn_attacks[myc])
        {
            path_forward = 0;
            if (square_file(pcsq) > 0)
                path_forward |= PATH_FORWARD[myc][pcsq - 1];
            if (square_file(pcsq) < 7)
                path_forward |= PATH_FORWARD[myc][pcsq + 1];
            if (!(path_forward & board.state[opp].piece[PAWN].u64) && !(knight_moves[pcsq] & board.state[opp].piece[KNIGHT].u64))
            {
                eval_values.pieces[myc] += B_KNIGHT_OUTPOST;

                if (!board.state[opp].piece[KNIGHT].u64 && !(square_color(pcsq) & board.state[opp].piece[BISHOP].u64))
                {
                    eval_values.pieces[myc] += B_KNIGHT_OUTPOST;
                }
            }
        }
        //  Blocking central pawns
        if (bb_square[pcsq] & BB_BLOCK_PAWN[myc] && board.state[myc].square[pcsq + SQUARE_BEHIND[myc]] == PAWN)
        {
            eval_values.pieces[myc] -= P_MINOR_BLOCK_PAWN;
        }

        bb_clear(piece.u64, pcsq);
    }
}

//-------------------------------------------------------------------------------------------------
//  Bishops
//-------------------------------------------------------------------------------------------------
void eval_bishops(int myc, int opp)
{
    BBIX    piece;
    int     pcsq;
    BBIX    mobility;
    U64     mob_target;

    piece.u64 = board.state[myc].piece[BISHOP].u64;
    while(piece.u64)
    {
        pcsq = bb_first(piece);

        assert(board.state[myc].square[pcsq] == BISHOP);

        // pst
        eval_values.pieces[myc] += pst_bishop[myc][pcsq];

        // mobility
        mob_target = board.occupied_squares & ~board.state[myc].piece[QUEEN].u64;
        mobility.u64 = bb_bishop_attacks(pcsq, mob_target) & ~board.state[myc].all_pieces;
        mobility.u64 &= ~eval_values.pawn_attacks[opp];
        eval_values.mobility[myc] += bb_count(mobility) * B_BISHOP_MOBILITY;

        //  threats
        eval_pieces_threat(mobility.u64, PAWN,   myc, opp, B_BISHOP_ATK_PAWN);
        eval_pieces_threat(mobility.u64, KNIGHT, myc, opp, B_BISHOP_ATK_KNIGHT);
        eval_pieces_threat(mobility.u64, BISHOP, myc, opp, B_BISHOP_ATK_BISHOP);
        eval_pieces_threat(mobility.u64, ROOK,   myc, opp, B_BISHOP_ATK_ROOK);
        eval_pieces_threat(mobility.u64, QUEEN,  myc, opp, B_BISHOP_ATK_QUEEN);

        // king attack
        eval_pieces_king_attack(mobility.u64, pcsq, myc, opp, KING_ATTACK_BISHOP);

        //  guards own king
        eval_pieces_guards_king(mobility.u64, myc, B_BISHOP_GUARD_KING);

        // penalty when attacked by pawn
        eval_pieces_attacked_by_pawn(myc, opp, pcsq, P_PAWN_ATK_BISHOP);

        // Blocking central pawns
        if (bb_square[pcsq] & BB_BLOCK_PAWN[myc] && board.state[myc].square[pcsq + SQUARE_BEHIND[myc]] == PAWN)
        {
            eval_values.pieces[myc] -= P_MINOR_BLOCK_PAWN;
        }

        // Bishop trapped at A2, A7
        if (pcsq == BISHOP_TRAP_A7[myc])
        {
            if (board.state[opp].square[BISHOP_TRAP_A7_PAWN[myc][0]] == PAWN 
                && board.state[opp].square[BISHOP_TRAP_A7_PAWN[myc][1]] == PAWN)
            {
                eval_values.pieces[myc] -= P_BISHOP_TRAP;
            }
        }

        // Bishop trapped at H2, H7
        if (pcsq == BISHOP_TRAP_H7[myc])
        {
            if (board.state[opp].square[BISHOP_TRAP_H7_PAWN[myc][0]] == PAWN 
                && board.state[opp].square[BISHOP_TRAP_H7_PAWN[myc][1]] == PAWN)
            {
                eval_values.pieces[myc] -= P_BISHOP_TRAP;
            }
        }

        bb_clear(piece.u64, pcsq);
    }
}

//-------------------------------------------------------------------------------------------------
//  Rooks
//-------------------------------------------------------------------------------------------------
void eval_rooks(int myc, int opp)
{
    BBIX    piece;
    int     pcsq;
    BBIX    mobility;
    U64     mob_target;

    piece.u64 = board.state[myc].piece[ROOK].u64;
    while(piece.u64)
    {
        pcsq = bb_first(piece);

        assert(board.state[myc].square[pcsq] == ROOK);

        // pst
		eval_values.pieces[myc] += pst_rook[myc][pcsq];

        // mobility
        mob_target = board.occupied_squares & ~(board.state[myc].piece[QUEEN].u64 | board.state[myc].piece[ROOK].u64);
        mobility.u64 = bb_rook_attacks(pcsq, mob_target) & ~board.state[myc].all_pieces;
        mobility.u64 &= ~eval_values.pawn_attacks[opp];
		eval_values.mobility[myc] += bb_count(mobility) * B_ROOK_MOBILITY;

        //  threats
        eval_pieces_threat(mobility.u64, PAWN,   myc, opp, B_ROOK_ATK_PAWN);
        eval_pieces_threat(mobility.u64, KNIGHT, myc, opp, B_ROOK_ATK_KNIGHT);
        eval_pieces_threat(mobility.u64, BISHOP, myc, opp, B_ROOK_ATK_BISHOP);
        eval_pieces_threat(mobility.u64, ROOK,   myc, opp, B_ROOK_ATK_ROOK);
        eval_pieces_threat(mobility.u64, QUEEN,  myc, opp, B_ROOK_ATK_QUEEN);

        // king attack
        eval_pieces_king_attack(mobility.u64, pcsq, myc, opp, KING_ATTACK_ROOK);

        //  guards own king
        eval_pieces_guards_king(mobility.u64, myc, B_ROOK_GUARD_KING);

        // penalty when attacked by pawn
        eval_pieces_attacked_by_pawn(myc, opp, pcsq, P_PAWN_ATK_ROOK);

        // rook attacking king on last rank
        if (square_rank(pcsq) == MY_RANK_8[myc])
        {
            if (MY_BB_RANK_8[myc] & board.state[opp].piece[KING].u64)
            {
                eval_values.pieces[myc] += B_ROOK_RANK_8;
            }
        }

        // rook on 7th rank
        if (square_rank(pcsq) == MY_RANK_7[myc])
        {
            if ((MY_BB_RANK_7[myc] & board.state[opp].piece[PAWN].u64) || 
                (MY_BB_RANK_8[myc] & board.state[opp].piece[KING].u64))
            {
                eval_values.pieces[myc] += B_ROOK_RANK_7;
            }
        }

        // rook on 6th rank
        if (square_rank(pcsq) == MY_RANK_6[myc])
        {
            if ((MY_BB_RANK_678[myc] & board.state[opp].piece[PAWN].u64) || 
                (MY_BB_RANK_678[myc] & board.state[opp].piece[KING].u64))
            {
                eval_values.pieces[myc] += B_ROOK_RANK_6;
            }
        }

        // rook in open file
        if (!(bb_files[square_file(pcsq)] & board.state[myc].piece[PAWN].u64))
        {
            if (!(bb_files[square_file(pcsq)] & board.state[opp].piece[PAWN].u64))
            {
                eval_values.pieces[myc] += B_ROOK_FULL_OPEN;
            }
            else
            {
                eval_values.pieces[myc] += B_ROOK_SEMI_OPEN;
            }
        }

        // trapped rook
        if (pcsq == ROOK_TRAP_A1_RSQ[myc] && (board.state[myc].piece[KING].u64 & ROOK_TRAP_A1_KSQ[myc]))
        {
            eval_values.pieces[myc] -= P_ROOK_TRAP;
        }
        if (pcsq == ROOK_TRAP_B1_RSQ[myc] && (board.state[myc].piece[KING].u64 & ROOK_TRAP_A1_KSQ[myc]))
        {
            eval_values.pieces[myc] -= P_ROOK_TRAP;
        }
        if (pcsq == ROOK_TRAP_H1_RSQ[myc] && (board.state[myc].piece[KING].u64 & ROOK_TRAP_H1_KSQ[myc]))
        {
            eval_values.pieces[myc] -= P_ROOK_TRAP;
        }
        if (pcsq == ROOK_TRAP_G1_RSQ[myc] && (board.state[myc].piece[KING].u64 & ROOK_TRAP_H1_KSQ[myc]))
        {
            eval_values.pieces[myc] -= P_ROOK_TRAP;
        }

        bb_clear(piece.u64, pcsq);
    }
}

void eval_queens(int myc, int opp)
{
    BBIX    piece;
    int     pcsq;
    BBIX    mobility;
    U64     mob_target;

    piece.u64 = board.state[myc].piece[QUEEN].u64;
    while(piece.u64)
    {
        pcsq = bb_first(piece);

        assert(board.state[myc].square[pcsq] == QUEEN);

        // pst
		eval_values.pieces[myc] += pst_queen[myc][pcsq];

        // mobility
        mob_target = board.occupied_squares & ~board.state[myc].piece[ROOK].u64;
        mobility.u64 = (bb_rook_attacks(pcsq, mob_target) & ~board.state[myc].all_pieces);
        mob_target = board.occupied_squares & ~board.state[myc].piece[BISHOP].u64;
        mobility.u64 |= (bb_bishop_attacks(pcsq, mob_target) & ~board.state[myc].all_pieces);

        mobility.u64 &= ~eval_values.pawn_attacks[opp];
        eval_values.mobility[myc] += bb_count(mobility) * B_QUEEN_MOBILITY;

        //  threats
        eval_pieces_threat(mobility.u64, PAWN,   myc, opp, B_QUEEN_ATK_PAWN);
        eval_pieces_threat(mobility.u64, KNIGHT, myc, opp, B_QUEEN_ATK_KNIGHT);
        eval_pieces_threat(mobility.u64, BISHOP, myc, opp, B_QUEEN_ATK_BISHOP);
        eval_pieces_threat(mobility.u64, ROOK,   myc, opp, B_QUEEN_ATK_ROOK);
        eval_pieces_threat(mobility.u64, QUEEN,  myc, opp, B_QUEEN_ATK_QUEEN);

        // king attack
        eval_pieces_king_attack(mobility.u64, pcsq, myc, opp, KING_ATTACK_QUEEN);

        //  guards own king
        eval_pieces_guards_king(mobility.u64, myc, B_QUEEN_GUARD_KING);

        // penalty when attacked by pawn
        eval_pieces_attacked_by_pawn(myc, opp, pcsq, P_PAWN_ATK_QUEEN);

        // queen on 7th rank
        if (square_rank(pcsq) == MY_RANK_7[myc])
        {
            if ((MY_BB_RANK_7[myc] & board.state[opp].piece[PAWN].u64) || 
                (MY_BB_RANK_8[myc] & board.state[opp].piece[KING].u64))
            {
                eval_values.pieces[myc] += B_QUEEN_RANK_7;

                if (MY_BB_RANK_7[myc] & board.state[myc].piece[ROOK].u64)
                {
                    eval_values.pieces[myc] += B_DOUBLE_RANK_7;
                }
            }
        }

        bb_clear(piece.u64, pcsq);
    }
}

//-------------------------------------------------------------------------------------------------
//  Give bonus of piece is attacking opponents undefended piece.
//-------------------------------------------------------------------------------------------------
void eval_pieces_threat(U64 attacks, int piece, int myc, int opp, int bonus)
{
    U64     attacked = attacks & board.state[opp].piece[piece].u64 & eval_values.undefended[opp];

    assert(piece >= PAWN && piece <= KING);
    assert(valid_color(myc));
    assert(valid_color(opp));

    if (attacked)
    {
        eval_values.pieces[myc] += bonus;
        eval_values.threats[myc].u64 |= attacked;
    }
}


//1r4k1/2RbQppp/p7/3Pp3/4P3/6P1/4NPBK/8 b - -
//-------------------------------------------------------------------------------------------------
//  Give bonus of piece is attacking opponents king area (king moves)
//-------------------------------------------------------------------------------------------------
void eval_pieces_king_attack(U64 attacks, int from, int myc, int opp, int attack_value)
{
    U64     king_zone_attacked = attacks & king_moves[board.state[opp].king_square];

    assert(valid_color(myc));
    assert(valid_color(opp));

    if (king_zone_attacked)
    {
        if (bb_square[from] & eval_values.pawn_attacks[opp])
        {
            //EVAL_TRACE("atk by pawn", from);
            return;
        }
        if (board.state[myc].square[from] == QUEEN)
        {
            if (king_zone_attacked & board.state[opp].piece[BISHOP].u64 & diagonal_rays[from])
            {
                //EVAL_TRACE("qxb", board.state[opp].king_square);
                return;
            }
            if (king_zone_attacked & board.state[opp].piece[ROOK].u64 & rankfile_rays[from])
            {
                //EVAL_TRACE("qxr", board.state[opp].king_square);
                return;
            }
            //EVAL_TRACE("q-ok", board.state[opp].king_square);
        }
        eval_values.king_attack_value[myc] += attack_value;
        eval_values.king_attack_count[myc]++;
    }
}

//-------------------------------------------------------------------------------------------------
//  Give bonus of piece is guarding own king area
//-------------------------------------------------------------------------------------------------
void eval_pieces_guards_king(U64 attacks, int myc, int bonus)
{
    assert(valid_color(myc));

    if (attacks & king_moves[board.state[myc].king_square])
    {
        eval_values.pieces[myc] += bonus;
    }
}

//-------------------------------------------------------------------------------------------------
//  Attacked by pawn penalty
//-------------------------------------------------------------------------------------------------
void eval_pieces_attacked_by_pawn(int myc, int opp, int pcsq, int penalty)
{
    U64     threat = eval_values.pawn_attacks[opp] & bb_square[pcsq];

    assert(valid_color(myc));
    assert(valid_color(opp));
    assert(valid_square(pcsq));

    if (threat)
    {
        eval_values.pieces[myc] -= penalty;
        eval_values.threats[opp].u64 |= threat;
    }
}

// END
