/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//	Generate capture and promotion moves.
//-------------------------------------------------------------------------------------------------

void	gen_white_pawn_captures(MOVES *ml);
void	gen_black_pawn_captures(MOVES *ml);

//-------------------------------------------------------------------------------------------------
//  Generate capture moves.
//-------------------------------------------------------------------------------------------------
void gen_caps(MOVES *ml)
{
	BBIX    piece, attacks;
	int     from, to;

    //  Pawns
    if (board.side_on_move == WHITE)
        gen_white_pawn_captures(ml);
    else
        gen_black_pawn_captures(ml);

    //  Knight
	piece.u64 = STM_STATE.piece[KNIGHT].u64;
	while(piece.u64)
    {
		from = bb_first(piece);
		attacks.u64 = knight_moves[from] & (OPP_STATE.all_pieces);
		while(attacks.u64)
        {
			to = bb_first(attacks);
			MOVES_ADD(ml, pack_cappc(from, to));
			bb_clear(attacks.u64, to);
		}
		bb_clear(piece.u64, from);
	}

    //  Rook and queen
	piece.u64 = STM_STATE.piece[ROOK].u64 | STM_STATE.piece[QUEEN].u64;
	while(piece.u64)
	{
		from = bb_first(piece);

        attacks.u64 = bb_rook_attacks(from, board.occupied_squares) & OPP_STATE.all_pieces;
		while(attacks.u64)
		{
			to = bb_first(attacks);
			MOVES_ADD(ml, pack_cappc(from, to));
			bb_clear(attacks.u64, to);
		}

		bb_clear(piece.u64, from);
	}

    //  Bishop and Queen
    piece.u64 = STM_STATE.piece[BISHOP].u64 | STM_STATE.piece[QUEEN].u64;
	while(piece.u64)
	{
		from = bb_first(piece);

        attacks.u64 = bb_bishop_attacks(from, board.occupied_squares) & OPP_STATE.all_pieces;
		while(attacks.u64)
		{
			to = bb_first(attacks);
			MOVES_ADD(ml, pack_cappc(from, to));
			bb_clear(attacks.u64, to);
		}

		bb_clear(piece.u64, from);
	}

    //  King
	from = STM_STATE.king_square;
	attacks.u64 = king_moves[from] & (OPP_STATE.all_pieces);
	while(attacks.u64)
    {
		to = bb_first(attacks);
		MOVES_ADD(ml, pack_cappc(from, to));
		bb_clear(attacks.u64, to);
	}
}

//-------------------------------------------------------------------------------------------------
//  White pawn captures and promotions
//-------------------------------------------------------------------------------------------------
void gen_white_pawn_captures(MOVES *ml)
{
    BBIX    moves, attacks, ep_capture;
	int     from, to;

    // promotions
    moves.u64 = ((board.state[WHITE].piece[PAWN].u64 & BB_RANK_7) << 8) & board.empty_squares;
	while(moves.u64)
	{
		to = bb_first(moves);
        MOVES_ADD(ml, pack_promo(to + 8, to, QUEEN));
        MOVES_ADD(ml, pack_promo(to + 8, to, ROOK));
        MOVES_ADD(ml, pack_promo(to + 8, to, BISHOP));
        MOVES_ADD(ml, pack_promo(to + 8, to, KNIGHT));
		bb_clear(moves.u64, to);
	}

    //  attacks to northwest
	attacks.u64 = (board.state[WHITE].piece[PAWN].u64 & BB_NO_AFILE) << 9;
	ep_capture.u64 = attacks.u64 & board.ep_square.u64;
	attacks.u64 &= board.state[BLACK].all_pieces;
	while(attacks.u64)
	{
		to = bb_first(attacks);
		from = to + 9;
		if (to < 8)
        {
			MOVES_ADD(ml, pack_cpprm(from, to, QUEEN));
			MOVES_ADD(ml, pack_cpprm(from, to, ROOK));
			MOVES_ADD(ml, pack_cpprm(from, to, BISHOP));
			MOVES_ADD(ml, pack_cpprm(from, to, KNIGHT));
        }
		else
			MOVES_ADD(ml, pack_cappc(from, to));
		bb_clear(attacks.u64, to);
	}
	if (ep_capture.u64)
	{
		to = bb_last(ep_capture);
		MOVES_ADD(ml, pack_epcap(to + 9, to, to + 8));
	}

	// attacks northeast
	attacks.u64 = (board.state[WHITE].piece[PAWN].u64 & BB_NO_HFILE) << 7;
	ep_capture.u64 = attacks.u64 & board.ep_square.u64;
	attacks.u64 &= board.state[BLACK].all_pieces;
	while(attacks.u64)
	{
		to = bb_first(attacks);
		from = to + 7;
		if (to < 8)
        {
			MOVES_ADD(ml, pack_cpprm(from, to, QUEEN));
			MOVES_ADD(ml, pack_cpprm(from, to, ROOK));
			MOVES_ADD(ml, pack_cpprm(from, to, BISHOP));
			MOVES_ADD(ml, pack_cpprm(from, to, KNIGHT));
        }
		else
			MOVES_ADD(ml, pack_cappc(from, to));
		bb_clear(attacks.u64, to);
	}
	if (ep_capture.u64)
	{
		to = bb_last(ep_capture);
		MOVES_ADD(ml, pack_epcap(to + 7, to, to + 8));
	}
}

//-------------------------------------------------------------------------------------------------
//  Black pawn captures and promotions
//-------------------------------------------------------------------------------------------------
void gen_black_pawn_captures(MOVES *ml) 
{
    BBIX    moves, attacks, ep_capture;
	int     from, to;

	//  promotions
   	moves.u64 = ((board.state[BLACK].piece[PAWN].u64 & BB_RANK_2) >> 8) & board.empty_squares;
	while(moves.u64)
	{
		to = bb_first(moves);
        MOVES_ADD(ml, pack_promo(to - 8, to, QUEEN));
        MOVES_ADD(ml, pack_promo(to - 8, to, ROOK));
        MOVES_ADD(ml, pack_promo(to - 8, to, BISHOP));
        MOVES_ADD(ml, pack_promo(to - 8, to, KNIGHT));
		bb_clear(moves.u64, to);
	}
	// attacks southeast
	attacks.u64 = (board.state[BLACK].piece[PAWN].u64 & BB_NO_HFILE) >> 9;
	ep_capture.u64 = attacks.u64 & board.ep_square.u64;
	attacks.u64 &= board.state[WHITE].all_pieces;
	while(attacks.u64)
	{
		to = bb_first(attacks);
		from = to - 9;
		if (to > 55)
        {
			MOVES_ADD(ml, pack_cpprm(from, to, QUEEN));
			MOVES_ADD(ml, pack_cpprm(from, to, ROOK));
			MOVES_ADD(ml, pack_cpprm(from, to, BISHOP));
			MOVES_ADD(ml, pack_cpprm(from, to, KNIGHT));
        }
		else
			MOVES_ADD(ml, pack_cappc(from, to));
		bb_clear(attacks.u64, to);
	}
	if (ep_capture.u64)
	{
		to = bb_last(ep_capture);
		MOVES_ADD(ml, pack_epcap(to - 9, to, to - 8));
	}
	// attacks southwest
	attacks.u64 = (board.state[BLACK].piece[PAWN].u64 & BB_NO_AFILE) >> 7;
	ep_capture.u64 = attacks.u64 & board.ep_square.u64;
	attacks.u64 &= board.state[WHITE].all_pieces;
	while(attacks.u64)
	{
		to = bb_first(attacks);
		from = to - 7;
		if (to > 55)
        {
			MOVES_ADD(ml, pack_cpprm(from, to, QUEEN));
			MOVES_ADD(ml, pack_cpprm(from, to, ROOK));
			MOVES_ADD(ml, pack_cpprm(from, to, BISHOP));
			MOVES_ADD(ml, pack_cpprm(from, to, KNIGHT));
        }
		else
			MOVES_ADD(ml, pack_cappc(from, to));
		bb_clear(attacks.u64, to);
	}
	if (ep_capture.u64)
	{
		to = bb_last(ep_capture);
		MOVES_ADD(ml, pack_epcap(to - 7, to, to - 8));
	}
}

//end
