/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Generate check evasion moves
//-------------------------------------------------------------------------------------------------
void	gm_evasion_king_moves(MOVES *ml, int king_square);

//-------------------------------------------------------------------------------------------------
//  When in check generate evasion moves.
//-------------------------------------------------------------------------------------------------
void gen_check_evasions(MOVES *ml)
{
	BBIX    temp;
    int     idx;
	int     king_square = bb_first(STM_STATE.piece[KING]);
	int     attack_count = 0;
	int     attack_square = 0;
	BBIX    attack_path;
	int     block_square;
	U64     pawns, moves, moves2sq;
    int     ep_pos[COLORS] = {-8, +8};

    assert(is_incheck(board.side_on_move));

	// step 1: count how many pieces are attacking the king
    // knight attacks (we can't have 2 knights attacking the king)
	temp.u64 = knight_moves[king_square] & OPP_STATE.piece[KNIGHT].u64;
	if (temp.u64)
    {
		idx = bb_last(temp);
		attack_square = idx;
		attack_count++;
	}
    // queen or rook attack.
    temp.u64 = bb_rook_attacks(king_square, board.occupied_squares) & (OPP_STATE.piece[QUEEN].u64 | OPP_STATE.piece[ROOK].u64);
	while(temp.u64) 
    {
		idx = bb_last(temp);
		attack_square = idx;
		attack_count++;
		bb_clear(temp.u64, idx);
	}
    //  queen or bishop attacks
    temp.u64 = bb_bishop_attacks(king_square, board.occupied_squares) & (OPP_STATE.piece[QUEEN].u64 | OPP_STATE.piece[BISHOP].u64);
	while(temp.u64)
    {
		idx = bb_last(temp);
		attack_square = idx;
		attack_count++;
		bb_clear(temp.u64, idx);
	}
    //  pawn attacks
	temp.u64 = pawn_attack[flip_color(board.side_on_move)][king_square] & OPP_STATE.piece[PAWN].u64;
	while(temp.u64)
    {
		idx = bb_last(temp);
		attack_square = idx;
		attack_count++;
		bb_clear(temp.u64, idx);
	}

    // Step 2: generate king escape moves
    gm_evasion_king_moves(ml, king_square);

	// When there is more than one attacker, only king moves are possible.
	if (attack_count > 1)
		return;

    // Step 3: generate moves that capture the attacker
    //  capture with knight
	temp.u64 = knight_moves[attack_square] & STM_STATE.piece[KNIGHT].u64;
	while (temp.u64)
    {
		idx = bb_last(temp);
		MOVES_ADD(ml, pack_cappc(idx, attack_square));
		bb_clear(temp.u64, idx);
	}
    //  capture with queen or rook
    temp.u64 = bb_rook_attacks(attack_square, board.occupied_squares) & (STM_STATE.piece[QUEEN].u64 | STM_STATE.piece[ROOK].u64);
	while(temp.u64)
    {
		idx = bb_last(temp);
        MOVES_ADD(ml, pack_cappc(idx, attack_square));
		bb_clear(temp.u64, idx);
	}
    //  capture with queen or bishop
    temp.u64 = bb_bishop_attacks(attack_square, board.occupied_squares) & (STM_STATE.piece[QUEEN].u64 | STM_STATE.piece[BISHOP].u64);
	while(temp.u64)
    {
		idx = bb_last(temp);
        MOVES_ADD(ml, pack_cappc(idx, attack_square));
		bb_clear(temp.u64, idx);
	}
    //  capture with pawn including capture/promotion
	temp.u64 = pawn_attack[board.side_on_move][attack_square] & STM_STATE.piece[PAWN].u64;
	while(temp.u64) 
    {
		idx = bb_last(temp);
        if (attack_square < 8 || attack_square > 55)
        {
		    MOVES_ADD(ml, pack_cpprm(idx, attack_square, QUEEN));
		    MOVES_ADD(ml, pack_cpprm(idx, attack_square, ROOK));
		    MOVES_ADD(ml, pack_cpprm(idx, attack_square, BISHOP));
		    MOVES_ADD(ml, pack_cpprm(idx, attack_square, KNIGHT));
        }
        else
		    MOVES_ADD(ml, pack_cappc(idx, attack_square));
        bb_clear(temp.u64, idx);
	}
    //  if attacker is pawn verify if can be ep-captured.
	if (OPP_STATE.square[attack_square] == PAWN && board.ep_square.u64)
    {
		idx = bb_first(board.ep_square);
        if (attack_square + ep_pos[board.side_on_move] == idx)
        {
            if (square_file(attack_square) > FILEA && STM_STATE.square[attack_square - 1] == PAWN)
                MOVES_ADD(ml, pack_epcap(attack_square - 1, idx, attack_square));
            if (square_file(attack_square) < FILEH && STM_STATE.square[attack_square + 1] == PAWN)
                MOVES_ADD(ml, pack_epcap(attack_square + 1, idx, attack_square));
        }
	}

	//  Step 4: generate moves that place a piece between king and attacker.
	attack_path.u64 = from_to_path[king_square][attack_square];
	while(attack_path.u64)
    {
		block_square = bb_first(attack_path);

        //  block with knight
		temp.u64 = knight_moves[block_square] & STM_STATE.piece[KNIGHT].u64;
		while (temp.u64)
        {
			idx = bb_last(temp);
            MOVES_ADD(ml, pack_quiet(idx, block_square));
			bb_clear(temp.u64, idx);
		}
        //  block with queen or rook
        temp.u64 = bb_rook_attacks(block_square, board.occupied_squares) & (STM_STATE.piece[QUEEN].u64 | STM_STATE.piece[ROOK].u64);
		while(temp.u64)
        {
			idx = bb_last(temp);
            MOVES_ADD(ml, pack_quiet(idx, block_square));
			bb_clear(temp.u64, idx);
		}
        //  block with queen or bishop
        temp.u64 = bb_bishop_attacks(block_square, board.occupied_squares) & (STM_STATE.piece[QUEEN].u64 | STM_STATE.piece[BISHOP].u64);
		while(temp.u64) 
        {
			idx = bb_last(temp);
            MOVES_ADD(ml, pack_quiet(idx, block_square));
			bb_clear(temp.u64, idx);
		}
		//  block with pawn moves, including promotions
		if (board.side_on_move == WHITE)
        {
			pawns = board.state[WHITE].piece[PAWN].u64;
			moves = (pawns << 8) & board.empty_squares;
			moves2sq = (((moves & BB_RANK_3) << 8) & board.empty_squares);
			if (moves & bb_square[block_square])
            {
                if (block_square < 8)
                {
                    MOVES_ADD(ml, pack_promo(block_square + 8, block_square, QUEEN));
                    MOVES_ADD(ml, pack_promo(block_square + 8, block_square, ROOK));
                    MOVES_ADD(ml, pack_promo(block_square + 8, block_square, BISHOP));
                    MOVES_ADD(ml, pack_promo(block_square + 8, block_square, KNIGHT));
                }
                else
                    MOVES_ADD(ml, pack_quiet(block_square + 8, block_square));
            }
			if (moves2sq & bb_square[block_square])
				MOVES_ADD(ml, pack_pawn2(block_square + 16, block_square, block_square + 8));
		}
		else
        {
			pawns = board.state[BLACK].piece[PAWN].u64;
			moves = (pawns >> 8) & board.empty_squares;
			moves2sq = (((moves & BB_RANK_6) >> 8) & board.empty_squares);
			if (moves & bb_square[block_square])
            {
                if (block_square > 55)
                {
                    MOVES_ADD(ml, pack_promo(block_square - 8, block_square, QUEEN));
                    MOVES_ADD(ml, pack_promo(block_square - 8, block_square, ROOK));
                    MOVES_ADD(ml, pack_promo(block_square - 8, block_square, BISHOP));
                    MOVES_ADD(ml, pack_promo(block_square - 8, block_square, KNIGHT));
                }
                else
                    MOVES_ADD(ml, pack_quiet(block_square - 8, block_square));
            }
			if (moves2sq & bb_square[block_square])
				MOVES_ADD(ml, pack_pawn2(block_square - 16, block_square, block_square - 8));
		}

		bb_clear(attack_path.u64, block_square);
	}
}

//-------------------------------------------------------------------------------------------------
//  Generate king evasion moves.
//-------------------------------------------------------------------------------------------------
void gm_evasion_king_moves(MOVES *ml, int king_square)
{
	BBIX    temp;
	int     to;

	temp.u64 = king_moves[king_square] & OPP_STATE.all_pieces;
	while(temp.u64)
    {
		to = bb_first(temp);
		MOVES_ADD(ml, pack_cappc(king_square, to));
		bb_clear(temp.u64, to);
	}
	temp.u64 = king_moves[king_square] & board.empty_squares;
	while(temp.u64)
    {
		to = bb_first(temp);
        MOVES_ADD(ml, pack_quiet(king_square, to));
		bb_clear(temp.u64, to);
	}
}

//end
