/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Generate quiet moves.
//-------------------------------------------------------------------------------------------------

void	gen_white_pawn_moves(MOVES *ml);
void	gen_black_pawn_moves(MOVES *ml);

//-------------------------------------------------------------------------------------------------
//  Generate moves.
//-------------------------------------------------------------------------------------------------
void gen_moves(MOVES *ml)
{
    int     from, to;
    BBIX    piece;
    BBIX    moves;

    //  Pawn
    if (board.side_on_move == WHITE)
        gen_white_pawn_moves(ml);
    else
        gen_black_pawn_moves(ml);

    //  Knight
    piece.u64 = STM_STATE.piece[KNIGHT].u64;
    while(piece.u64)
    {
        from = bb_last(piece);
        moves.u64 = knight_moves[from] & board.empty_squares;
        while(moves.u64)
        {
            to = bb_first(moves);
            MOVES_ADD(ml, pack_quiet(from, to));
            bb_clear(moves.u64, to);
        }
        bb_clear(piece.u64, from);
    }

    //  Rook and Queen
    piece.u64 = STM_STATE.piece[ROOK].u64 | STM_STATE.piece[QUEEN].u64;
    while(piece.u64)
    {
        from = bb_first(piece);

		for (to = from + 1; to < 64 && bb_square[to] & rankfile_east[from] & board.empty_squares; to += 1)
			MOVES_ADD(ml, pack_quiet(from, to));
		for (to = from + 8; to < 64 && bb_square[to] & rankfile_south[from] & board.empty_squares; to += 8)
			MOVES_ADD(ml, pack_quiet(from, to));
		for (to = from - 1; to >= 0 && bb_square[to] & rankfile_west[from] & board.empty_squares; to -= 1)
			MOVES_ADD(ml, pack_quiet(from, to));
		for (to = from - 8; to >= 0 && bb_square[to] & rankfile_north[from] & board.empty_squares; to -= 8)
			MOVES_ADD(ml, pack_quiet(from, to));

        bb_clear(piece.u64, from);
    }

    //  Bishop and Queen
    piece.u64 = STM_STATE.piece[BISHOP].u64 | STM_STATE.piece[QUEEN].u64;
    while(piece.u64)
    {
        from = bb_first(piece);

		for (to = from + 9; to < 64 && bb_square[to] & diagonal_se[from] & board.empty_squares; to += 9)
			MOVES_ADD(ml, pack_quiet(from, to));
		for (to = from + 7; to < 64 && bb_square[to] & diagonal_sw[from] & board.empty_squares; to += 7)
			MOVES_ADD(ml, pack_quiet(from, to));
		for (to = from - 9; to >= 0 && bb_square[to] & diagonal_nw[from] & board.empty_squares; to -= 9)
			MOVES_ADD(ml, pack_quiet(from, to));
		for (to = from - 7; to >= 0 && bb_square[to] & diagonal_ne[from] & board.empty_squares; to -= 7)
			MOVES_ADD(ml, pack_quiet(from, to));

        bb_clear(piece.u64, from);
    }

    //  King
    from = STM_STATE.king_square;
    moves.u64 = king_moves[from] & board.empty_squares;
    while(moves.u64)
    {
        to = bb_first(moves);
        MOVES_ADD(ml, pack_quiet(from, to));
        bb_clear(moves.u64, to);
    }

    //  Castle
    if (board.side_on_move == WHITE)
    {
        if (board.state[WHITE].can_castle_ks && board.state[WHITE].square[H1] == ROOK)
            if (!((U64)0x0000000000000006 & board.occupied_squares))
                MOVES_ADD(ml, pack_castl(from, G1, MT_CSWKS));
        if (board.state[WHITE].can_castle_qs && board.state[WHITE].square[A1] == ROOK)
            if (!((U64)0x0000000000000070 & board.occupied_squares))
                MOVES_ADD(ml, pack_castl(from, C1, MT_CSWQS));
    }
    else
    {
        if (board.state[BLACK].can_castle_ks && board.state[BLACK].square[H8] == ROOK)
            if (!((U64)0x0600000000000000 & board.occupied_squares))
                MOVES_ADD(ml, pack_castl(from, G8, MT_CSBKS));
        if (board.state[BLACK].can_castle_qs && board.state[BLACK].square[A8] == ROOK)
            if (!((U64)0x7000000000000000 & board.occupied_squares))
                MOVES_ADD(ml, pack_castl(from, C8, MT_CSBQS));
    }
}

//-------------------------------------------------------------------------------------------------
//  White pawns
//-------------------------------------------------------------------------------------------------
void gen_white_pawn_moves(MOVES *ml)
{
    BBIX    moves1sq;
    BBIX    moves2sq;
    int     to;

    moves1sq.u64 = ((board.state[WHITE].piece[PAWN].u64 & ~BB_RANK_7) << 8) & board.empty_squares;
    moves2sq.u64 = (((moves1sq.u64 & BB_RANK_3) << 8) & board.empty_squares);
    //  1 square move
    while(moves1sq.u64)
    {
        to = bb_first(moves1sq);
        MOVES_ADD(ml, pack_quiet(to + 8, to));
        bb_clear(moves1sq.u64, to);
    }
    // 2 square move
    while(moves2sq.u64)
    {
        to = bb_first(moves2sq);
        MOVES_ADD(ml, pack_pawn2(to + 16, to, to + 8));
        bb_clear(moves2sq.u64, to);
    }
}

//-------------------------------------------------------------------------------------------------
//  Black pawn
//-------------------------------------------------------------------------------------------------
void gen_black_pawn_moves(MOVES *ml) 
{
    BBIX    moves1sq;
    BBIX    moves2sq;
    int     to;

    moves1sq.u64 = ((board.state[BLACK].piece[PAWN].u64 & ~BB_RANK_2) >> 8) & board.empty_squares;
    moves2sq.u64 = (((moves1sq.u64 & BB_RANK_6) >> 8) & board.empty_squares);

    // 1 square move
    while(moves1sq.u64)
    {
        to = bb_first(moves1sq);
        MOVES_ADD(ml, pack_quiet(to - 8, to));
        bb_clear(moves1sq.u64, to);
    }
    // 2 square move
    while(moves2sq.u64)
    {
        to = bb_first(moves2sq);
        MOVES_ADD(ml, pack_pawn2(to - 16, to, to - 8));
        bb_clear(moves2sq.u64, to);
    }
}

//end