/*-------------------------------------------------------------------------------
  Tobler is XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  Tobler is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Tobler is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

//-------------------------------------------------------------------------------------------------
// All definitions and declarations.
//-------------------------------------------------------------------------------------------------

// Avoid warnings for microsoft compiler (visual studio)
#define _CRT_SECURE_NO_WARNINGS

// Comment next line to use "assert" functions. "assert" functions are helpfull during debug.
#define NDEBUG

#pragma warning (disable : 4706)
#pragma warning (disable : 4711)
#pragma warning (disable : 4255)

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <memory.h>
#include <ctype.h>
#include <limits.h>
#include <stdlib.h>
#include <assert.h>
#include <setjmp.h>
//#include <direct.h>

// Hack to define variables only once in this file.
#ifndef EXTERN
#define EXTERN extern
#endif

// Type definitions
typedef unsigned long long  U64; // bitboard
typedef signed int          SINT;
typedef unsigned int        UINT;
typedef signed char         S8;
typedef unsigned char       U8;
typedef signed short        S16;
typedef unsigned short      U16;

// Startup Fen
#define FEN_NEW_GAME		"rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"

// Colors
#define WHITE		0
#define BLACK		1
#define COLORS      2

#define flip_color(color)   ((color) ^ 1)

// Search values
#define MAX_PLY		64
#define	MAX_DEPTH	32
#define	MAX_HIST	1000

#define MAX_SCORE	9000
#define MATE_VALUE	9999
#define INFINITY	32767

#define POST_NONE		0
#define POST_DEFAULT	1
#define	POST_XBOARD		2

// Piece index
#define EMPTY		0
#define PAWN		1
#define	KNIGHT		2
#define BISHOP		3
#define	ROOK		4
#define	QUEEN		5
#define	KING		6
#define NUM_PIECES	7

// Ranks/Files as represented in this program.
enum	{FILEA, FILEB, FILEC, FILED, FILEE, FILEF, FILEG, FILEH};
enum	{RANK8, RANK7, RANK6, RANK5, RANK4, RANK3, RANK2, RANK1};

#define RANKS       8
#define FILES       8

// Bitboard
EXTERN  U64 bb_table_set[64];
EXTERN  U64 bb_table_clr[64];

// This union is used to get first and last index, and bit counts.
typedef union u_bitboard_index
{
    U64     u64;
    UINT    u32[2];
    U16     u16[4];
}   BBIX;

void	bb_init(void);
SINT	bb_first(BBIX bbix);
SINT	bb_last(BBIX bbix);
UINT	bb_count(BBIX bbix);
void	bb_print(char *msg, U64 bb);

#define bb_set_rf(bitboard, rank, file)     ((bitboard) |= bb_table_set[(rank) * 8 + (file)])
#define bb_clear_rf(bitboard, rank, file)   ((bitboard) &= bb_table_clr[(rank) * 8 + (file)])
#define bb_set(bitboard, index)		        ((bitboard) |= bb_table_set[(index)])
#define bb_clear(bitboard, index)		    ((bitboard) &= bb_table_clr[(index)])
#define bb_is_one(bitboard, index)		    ((bitboard) & bb_table_set[(index)] ? 1 : 0)

U64 bb_rook_attacks(int sq, U64 occup);
U64 bb_bishop_attacks(int sq, U64 occup);

// move types
#define MT_QUIET    0
#define MT_CAPPC    1
#define MT_PROMO    2
#define MT_CPPRM    3
#define MT_PAWN2    4
#define MT_EPCAP    5
#define MT_CSWKS    6
#define MT_CSWQS    7
#define MT_CSBKS    8
#define MT_CSBQS    9
#define MT_NULL     10
#define MT_SIZE     11

typedef unsigned int MOVE;

// Move Structure (from right to left)
// tosq: 6 bits - to square
// frsq: 6 bits - from square
// type: 4 bits - move type (quiet, capture, etc)
// mvpc: 3 bits - moving piece
// cppc: 3 bits - captured piece (Capture. Ep Capture, Capture/Promotion)
// prom: 3 bits - promotion piece (Promotion, Capture/Promotion)
// epsq or pwsq: 6 bits - ep square or pawn square (ep square when is pawn 2 square move, pawn square when is EP Capture)
// quiet: 1 bit - flag to indicate quiet moves

#define pack_quiet(frsq, tosq)          (MOVE)(0x80000000 | STM_STATE.square[frsq] << 16 | MT_QUIET << 12 | (frsq) << 6 | (tosq))
#define pack_pawn2(frsq, tosq, epsq)    (MOVE)(0x80000000 | (epsq) << 25 | PAWN << 16 | MT_PAWN2 << 12 | (frsq) << 6 | (tosq))
#define pack_castl(frsq, tosq, cast)    (MOVE)(0x80000000 | KING << 16 | cast << 12 | (frsq) << 6 | (tosq))
#define pack_cappc(frsq, tosq)          (MOVE)(OPP_STATE.square[tosq] << 19 | STM_STATE.square[frsq] << 16 | MT_CAPPC << 12 | (frsq) << 6 | (tosq))
#define pack_epcap(frsq, tosq, pwsq)    (MOVE)((pwsq) << 25 | PAWN << 19 | PAWN << 16 | MT_EPCAP << 12 | (frsq) << 6 | (tosq))
#define pack_promo(frsq, tosq, prom)    (MOVE)((prom) << 22 | PAWN << 16 | MT_PROMO << 12 | (frsq) << 6 | (tosq))
#define pack_cpprm(frsq, tosq, prom)    (MOVE)((prom) << 22 | OPP_STATE.square[tosq] << 19| PAWN << 16 | MT_CPPRM << 12 | (frsq) << 6 | (tosq))

#define MOVE_NULL   (MT_NULL << 12)
#define MOVE_NONE   0

#define TOSQ(m)         ((m)       & 0x3F)
#define FRSQ(m)         ((m >>  6) & 0x3F)
#define TYPE(m)         ((m >> 12) & 0x0F)
#define MVPC(m)         ((m >> 16) & 0x07)
#define CPPC(m)         ((m >> 19) & 0x07)
#define PROM(m)         ((m >> 22) & 0x07)
#define EPSQ(m)         ((m >> 25) & 0x3F)
#define PWSQ(m)         ((m >> 25) & 0x3F)
#define IS_QUIET(m)     ((m) & 0x80000000)
#define IS_CASTLE(m)    (TYPE(m) >= MT_CSWKS && TYPE(m) <= MT_CSBQS)


#define	SORT_CAPS	200000000
#define	SORT_KILL	100000000

#define ABS(a)		((a) < 0 ? -(a) : (a))
#define MAX(a, b)	((a) > (b) ? (a) : (b))
#define MIN(a, b)	((a) < (b) ? (a) : (b))

//#define NULL_DEPTH(d)   ((d) - 3 - ((d) > 8 ? 1 : 0))
#define NULL_DEPTH(d)   ((d) - 4)

void	data_init(void);

// board
typedef struct s_board
{
	struct s_state
    {
		BBIX		piece[7];			// bb for each piece type
		U64			all_pieces;			// bb for all pieces combined
		UINT 		square[64];			// piece location on the board
		UINT 		can_castle_ks;		// king side
		UINT 		can_castle_qs;		// queen side
        UINT        king_square;        // calculated after king move
		int			dummy;
	}				state[COLORS];
	U64				occupied_squares;
    U64             empty_squares;
	BBIX			ep_square;
	U64				key;
    U64             pawn_key;
	int				side_on_move;
	UINT			fifty_move_rule;
    UINT            ply;
    int             dummy;
}	BOARD_TYPE;

EXTERN	BOARD_TYPE	board;

#define square_rank(square) ((square) >> 3)
#define square_file(square) ((square) & 0x07)

#define SQ_NE(sq)   ((sq) - 7)
#define SQ_NW(sq)   ((sq) - 9)
#define SQ_SE(sq)   ((sq) + 9)
#define SQ_SW(sq)   ((sq) + 7)
#define SQ_N(sq)    ((sq) - 8)
#define SQ_S(sq)    ((sq) + 8)
#define SQ_E(sq)    ((sq) + 1)
#define SQ_W(sq)    ((sq) - 1)

void	new_game(char *fen);
void	make_move(MOVE move);
void	undo_move(void);
int		is_draw(void);
int		insufficient_material(void);
int     material_value(int color);
int     pieces_count(int color);
int     is_valid(MOVE move);
int		square_distance(int square1, int square2);
void	board_print(char *text);

// history information
typedef struct s_history
{
	U64			board_key;
    U64         pawn_key;
	U64			ep_square;
	U64			occupied_squares;
	MOVE		move;
	int 		can_castle_ks;
	int 		can_castle_qs;
	int			fifty_move_rule;
    UINT        king_square;
	int			dummy;
}	HISTORY;

EXTERN	int			histply;
EXTERN	HISTORY		history[MAX_HIST];

// Position information
typedef struct s_position
{
    MOVE        curr_move;
    MOVE        killer[COLORS][3];
    MOVE        pv_line[MAX_PLY];
    int         pv_size;
    int         threats[COLORS];
}   POSITION;

EXTERN  POSITION    position[MAX_PLY];

// State of side to move
#define STM_STATE	board.state[board.side_on_move]
// State of opponent
#define OPP_STATE	board.state[flip_color(board.side_on_move)]

// Eval score definitions.
// Instead of using two variables for opening and endgame values, 
// we use the macro MAKE_SCORE to combine into one variable. 
// Didn't measure if there's any advantage, it is just a matter
// of preference to avoid handling two variables for each eval term.
// Stockfish and Gull are two engines that uses the same mechanism.
#define MAKE_SCORE(op, eg)  (((op) << 16) + (eg))
#define OPENING(s) ((((s) + 32768) & ~0xffff) / 0x10000)
#define ENDGAME(s) (((unsigned)(s) & 0x7fffu) - (int)((unsigned)(s) & 0x8000u))

// Piece Values
#define VALUE_PAWN		 90
#define	VALUE_KNIGHT	320
#define VALUE_BISHOP	320
#define VALUE_ROOK		500
#define VALUE_QUEEN		1000
#define VALUE_KING		0

#define SCORE_PAWN		MAKE_SCORE(90, 100)
#define SCORE_KNIGHT	MAKE_SCORE(320, 320)
#define SCORE_BISHOP	MAKE_SCORE(320, 320)
#define SCORE_ROOK		MAKE_SCORE(510, 530)
#define SCORE_QUEEN		MAKE_SCORE(1000, 1000)

EXTERN int	piece_value[NUM_PIECES];
EXTERN int	see_value[NUM_PIECES];
EXTERN int  capture_value[NUM_PIECES][NUM_PIECES];
EXTERN int  promotion_value[NUM_PIECES];
EXTERN int  prom_capt_value[NUM_PIECES][NUM_PIECES];
EXTERN int  is_sliding[NUM_PIECES];
EXTERN char	rank_file_notation[64][3];
EXTERN char	piece_letter[NUM_PIECES];
EXTERN char	promo_letter[NUM_PIECES];

enum e_squares
{
	A8, B8,	C8, D8, E8, F8, G8, H8,
	A7, B7,	C7, D7, E7, F7, G7, H7,
	A6, B6,	C6, D6, E6, F6, G6, H6,
	A5, B5,	C5, D5, E5, F5, G5, H5,
	A4, B4,	C4, D4, E4, F4, G4, H4,
	A3, B3,	C3, D3, E3, F3, G3, H3,
	A2, B2,	C2, D2, E2, F2, G2, H2,
	A1, B1,	C1, D1, E1, F1, G1, H1
};

// Move generation
EXTERN	U64			king_moves[64];			// possible king moves from each square
EXTERN	U64			knight_moves[64];		// possible king moves from each square
EXTERN	U64 		rankfile_west[64];		// west
EXTERN	U64 		rankfile_east[64];		// east
EXTERN	U64 		rankfile_north[64];		// north
EXTERN	U64 		rankfile_south[64];		// south
EXTERN	U64 		diagonal_se[64];		// southeast
EXTERN	U64 		diagonal_sw[64];		// southwest
EXTERN	U64 		diagonal_ne[64];		// northeast
EXTERN	U64 		diagonal_nw[64];		// northwest
EXTERN	U64			rankfile_rays[64];		// ranks and files from square
EXTERN	U64			diagonal_rays[64];		// diagonals from square
EXTERN	U64			from_to_path[64][64];	// path from square to square: eg. from [A1][A8]=A2-A7
EXTERN	U64			bb_square[64];			// bitboard for each square
EXTERN  U64			pawn_attack[2][64];     // position for pawns to attack a square

#define BB_NO_AFILE ((U64)0x7F7F7F7F7F7F7F7F)
#define BB_NO_HFILE ((U64)0xFEFEFEFEFEFEFEFE)
#define BB_RANK_6   ((U64)0x0000FF0000000000)
#define BB_RANK_3   ((U64)0x0000000000FF0000)
#define BB_RANK_8   ((U64)0xFF00000000000000)
#define BB_RANK_7   ((U64)0x00FF000000000000)
#define BB_RANK_2   ((U64)0x000000000000FF00)
#define BB_RANK_1   ((U64)0x00000000000000FF)
#define BB_CENTER	((U64)0x0000001818000000)
#define BB_LIGHT_SQ ((U64)0xAA55AA55AA55AA55)
#define BB_DARK_SQ  ((U64)0x55AA55AA55AA55AA)

#define square_color(s) (bb_square[(s)] & BB_LIGHT_SQ ? BB_LIGHT_SQ : BB_DARK_SQ)

// Move generation and selection
#define MAX_MOVE   250

typedef struct s_move_list
{
	MOVE		moves[MAX_MOVE];
	int			score[MAX_MOVE];
	int			count;
    int         incheck;
    int         caps;
    int         gen_checks;
    MOVE        ttm;
    int         next;
    int         phase;
    MOVE        bc_moves[MAX_MOVE];
    int         bc_count;
    int         bc_next;
    int         sort;
}	MOVES;

#define MOVES_ADD(ml, move) ((ml)->moves[(ml)->count++] = (move))

void	gen_moves(MOVES *ml);
void	gen_caps(MOVES *ml);
void	gen_check_evasions(MOVES *ml);
void	print_current_moves(void);

void    moves_init(MOVES *ml, int incheck, MOVE ttm, int caps);
MOVE    moves_next(MOVES *ml);
MOVE    moves_next_perft(MOVES *ml);
void    moves_select_quiet(MOVES *ml);
int     moves_is_bad_cap(MOVES *ml);

void    pv_moves_init(MOVES *ml, int incheck, MOVE ttm, int caps);
MOVE    pv_moves_next(MOVES *ml);

void    quiesce_moves_init(MOVES *ml, int incheck, int gen_checks);
MOVE    quiesce_moves_next(MOVES *ml);
int     quiesce_moves_next_check(MOVES *ml);

int     is_badcap(MOVE move);
int     is_mate_score(int score);

// see
int     see_capture(MOVE move);
int		see_move(MOVE move);
int     see_sign(int color, int frsq, int tosq);

// Search
struct s_search_control
{
	U64     nodes;			// visited nodes
	U64		qnodes;			// quiesce visited nodes
    int     moves_level;    // moves from level command
    int     moves_to_go[COLORS];    // moves per time control
	UINT    normal_move_time;	    // max time for move
    UINT    extended_move_time;    // max time for long search
	UINT    start_time;
	UINT    normal_finish_time;
	UINT    extended_finish_time;
	int		post_flag;		// output information
	int		min_depth;		// min depth for timeout
	int		max_depth;		// max depth 
    int     cur_depth;      // current depth
    int     can_stop;       // controls when to stop a search
    int     score_drop;     // controls when score drops
	int		stop;			// time out
};

EXTERN struct s_search_control search;

enum            {TT_EXACT, TT_UPPER, TT_LOWER};

#define TIME_CHECK	2047

EXTERN  jmp_buf env;

void	update_pv(MOVE move);
int	    is_free_pawn(int color, MOVE move);
int     has_pawn_on_rank7(int color);
void	check_time(void);
void	search_init(void);
void	think(int output, MOVE *move, int usebook);
void	select_next_move(int from);
void	select_next_move_hint(int from, MOVE hint);
int		search_root(MOVES *ml, UINT incheck, int alpha, int beta, int depth);
int     search_pv(UINT incheck, int alpha, int beta, int depth);
int     search_zw(UINT incheck, int beta, int depth, UINT can_null);
int     search_evasion(int beta, int depth);
int		quiesce(UINT incheck, int alpha, int beta, int depth);
void	post_info(int score, int depth);
int     move_is_killer(int color, int ply, MOVE move);
int     is_defended_by_pawn(int square, int color);
int     is_check(MOVE move);

void    move_hist_init(void);
void    move_hist_save(int color, int ply, MOVE move, int depth);
void    move_hist_bad(int color, MOVE list[], int size);
int     move_hist_val(int color, MOVE move);
int     move_hist_check_value(int color, int king_sq, int frsq);
void    move_hist_print(void);
int     move_has_bad_hist(int color, MOVE move);

// Zobrish Keys (hash)
EXTERN U64		zkBase[2][7][64]; // color, piece, index
EXTERN U64		zkColor;
EXTERN U64		zkEnPassant[64];
EXTERN U64		zkCastleKS[2][2]; // color, no/yes
EXTERN U64		zkCastleQS[2][2];

void	zk_init(void);
U64		zk_board_key(void);
U64     zk_pawn_key(void);
U64     zk_rand(void);

// Book
void	book_init(void);
MOVE    book_next_move(void);

// Evaluation
#define OP      0
#define EG      1
#define PHASES  2

EXTERN U64      bb_files[8];
EXTERN U64      bb_pawn_weak[2][64];    // mask to test weak pawns
EXTERN U64      bb_pawn_supp[2][64];    // mask to test supportted pawns
EXTERN U64		bb_pawn_pass[2][64];	// mask to test passed pawns
EXTERN U64      bb_pawn_capt[2][64];    // mask to test pawn capture
EXTERN U64		bb_pawn_isol[64];	    // mask to test isolated pawns
EXTERN U64		bb_pawn_conn[64];		// mask to test connected pawns
//EXTERN int 		pst_n_op[COLORS][64];	// knight pst
//EXTERN int 		pst_n_eg[COLORS][64];	// knight pst
//EXTERN int		pst_b_op[COLORS][64];	// bishop pst
//EXTERN int		pst_b_eg[COLORS][64];	// bishop pst
//EXTERN int		pst_q_op[COLORS][64];	// queen pst
//EXTERN int		pst_q_eg[COLORS][64];	// queen pst
//EXTERN int		pst_r_op[COLORS][64];	// rook pst
//EXTERN int		pst_r_eg[COLORS][64];	// rook pst
////EXTERN int 		pst_p_op[COLORS][64];	// pawn pst
////EXTERN int 		pst_p_eg[COLORS][64];	// pawn pst
//EXTERN int 		pst_k_op[COLORS][64];	// king opening pst
//EXTERN int		pst_k_eg[COLORS][64];	// king endgame pst
//EXTERN int		pst_k_eg_ks[COLORS][64];	// king endgame pst king side pawns
//EXTERN int		pst_k_eg_qs[COLORS][64];	// king endgame pst queen side pawns

EXTERN int		pst_pawn[COLORS][64];
EXTERN int		pst_king[COLORS][64];
EXTERN int		pst_king_ks[COLORS][64];
EXTERN int		pst_king_qs[COLORS][64];
EXTERN int		pst_knight[COLORS][64];
EXTERN int		pst_bishop[COLORS][64];
EXTERN int		pst_rook[COLORS][64];
EXTERN int		pst_queen[COLORS][64];

typedef struct s_pt
{
    U64     key;
    int     pawn_eval[COLORS];
    U64     bb_passers[COLORS];
}   PWTB;

//#define PT_SIZE 0x100000
#define PT_SIZE 32768

EXTERN PWTB    pawn_table[PT_SIZE];

//#define ET_SIZE	0x100000
#define ET_SIZE	65536

typedef struct s_et
{
	U64		key;
	int		score;
	int		dummy;
}	EVTB;

EXTERN EVTB    eval_table[ET_SIZE];

EXTERN int ep_on;

typedef struct s_eval_values
{
    int     phase;
	int		material[COLORS];
	int		pawn[COLORS];
    int     king[COLORS];
	int		passed[COLORS];
	int		pieces[COLORS];
    int     mobility[COLORS];
    int     draw_adjust[COLORS];
    int     flag_king_safety[COLORS];
    int     king_attack_value[COLORS];
	int		king_attack_count[COLORS];
    int     king_zone_value;
    int     king_zone_count;
	int		score[COLORS];
	int		dummy;
    U64     bb_passers[COLORS];
    U64     pawn_attacks[COLORS];
    U64     strong_minors[COLORS];
    U64     undefended[COLORS];
    BBIX    threats[COLORS];
}   EVAL_VALUES;

EXTERN  EVAL_VALUES     eval_values;

int     evaluate(void);
void	eval_init(void);
void	eval_pst_init(void);
void	clear_eval_table(void);
void	eval_print(void);

// transposition table
void    tt_age(void);
void    tt_init(int size_mb);
void    tt_clear(void);
void	tt_save(int depth, int search_score, S8 flag, MOVE best);
int		tt_probe(int depth, int alpha, int beta, int *search_score, MOVE *best_move, UINT *skip_null);
MOVE    tt_move(void);

// Attacks/Checks
int		is_illegal(MOVE move);
int		is_incheck(int color);
int		is_square_attacked(int square, int color);
int		is_white_kingside_attacked(void);
int		is_white_queenside_attacked(void);
int		is_black_kingside_attacked(void);
int		is_black_queenside_attacked(void);

#define GR_NOT_FINISH	0
#define GR_WHITE_WIN	1
#define GR_BLACK_WIN	2
#define GR_DRAW			3

int get_game_result(void);

// Utils
unsigned int util_get_time(void);
void	util_get_move_string(MOVE move, char *string);
void	util_get_move_desc(MOVE move, char *string, int inc_file);
void	util_parse_move(char *move_string, MOVE *parsed);
void    util_print_move(MOVE move, int new_line);

// Perf
void	perft(int depth);
void	perftx(void);
void	perfty(void);

// EPD
double	epd(char *file_name);

// tests
void    eval_test(char *file_name);

// Log File
//#define LOG
#ifdef LOG
EXTERN FILE	*log_file;
#endif

U64     zk_rand(void);

#ifndef NDEBUG
// Assert functions.
int     valid_material(void);
int     valid_rank_file(void);
int     valid_square(int square);
int     valid_color(int color);
int     board_state_is_ok(void);
int     valid_piece(int pcsq, int color, UINT type);
int     pawn_is_doubled(int pcsq, int color);
int     pawn_is_connected(int pcsq, int color);
int     pawn_is_weak(int pcsq, int color);
int     pawn_is_passed(int pcsq, int color);
int     pawn_is_isolated(int pcsq, int color);
int     pawn_is_backward(int pcsq, int color);
int     pawn_is_candidate(int pcsq, int color);
#endif

EXTERN int test_on;
EXTERN U64 test_cnt;
EXTERN U64 test_hit;

//End
