/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#define EXTERN
#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  main loop.
//-------------------------------------------------------------------------------------------------

#define VERSION "2.00"

void	verify_game_results(void);
int		is_game_ended(void);
void	calc_move_time(UINT moves_to_go, UINT max_time, UINT move_time);
void    update_moves_to_go(int color);
void	show_dev_info(void);
void    magic_init(void);
void    bench(int depth);
int     magic_generate(void);

#ifndef _WIN32
#  define __cdecl    /* nothing */
#  define __stdcall  /* nothing */
#  define __fastcall /* nothing */
#endif /* _WIN32 */

int __cdecl main(int argc, char *argv[])
{
	int		computer = -1;
	int		max_depth, max_time;
	int		post = POST_DEFAULT;
	int		use_book = 0;
	UINT    start, finish;
	double	duration;
	char	line[1000];
    char	command[512] = {'\0'};
	int		stop = 0;
	MOVE	move;
	MOVE	parsed;
	char	move_string[20];
	int		perft_depth;
	char	epd_file[1000];
	int		auto_play = 0;
    char    xboard_reply[100];
    int     time1000 = 0;
    int     i;
    int     hash_size = 64;

#ifdef LOG
	char	filename[100];
#endif

	printf("tucano chess engine by Alcides Schulz - %s\n", VERSION);

    for (i = 0; i < argc; i++)
    {
        // specific parameter for fics.org
        if (!strcmp("-time1000", argv[i]))
            time1000 = 1;
        // hash table size
        if (!strcmp("-hash", argv[i]))
        {
            if (++i < argc)
            {
                hash_size = atoi(argv[i]);
                if (hash_size < 0)
                    hash_size = 8;
            }
        }
		if (!strcmp("-nobook", argv[i]))
		{
			use_book = 0;
		}
    }

    printf("   hash table: %d MB\n", hash_size);

    //srand((unsigned int)2593452304/*time(NULL)*/);
    srand((unsigned int)19810505/*time(NULL)*/);

	bb_init();
	zk_init();
	data_init();
	eval_init();
    eval_pst_init();
    search_init();
	book_init();
    tt_init(hash_size);

    magic_init();

	sprintf(xboard_reply, "feature setboard=1 myname=\"tucano_%s\" colors=0 analyze=0 done=1", VERSION);

#ifndef NDEBUG
    printf("\nDEBUG MODE ON (running with asserts)\n");
#endif

#ifdef LOG
	_mkdir("log");
	sprintf(filename, "log/tucano.%s.log", VERSION);
    printf("Log file: %s\n", filename);
	log_file = fopen(filename, "w");
#endif

    assert(valid_rank_file());

	signal(SIGINT, SIG_IGN);
	printf("\n");
	new_game(FEN_NEW_GAME);

    assert(valid_rank_file());

	while(!stop)
    {

		fflush(stdout);

		if (board.side_on_move == computer || auto_play)
        {

			start = util_get_time();
			think(post, &move, use_book);
			finish = util_get_time();
			duration = (double)(finish - start) / (double)1000;
			if (duration == 0)
				duration = 0.0001;

			if (post == 1)
            {
                printf("\nNodes: %lld  QNodes: %lld  Time spent: %3.2f  Nodes/Sec=%8.0f\n", search.nodes + search.qnodes, search.qnodes, duration, ((search.nodes + search.qnodes) / duration));
				printf("\n");
			}
			
			if (!move)
            {
#ifdef LOG
				fprintf(log_file, "no move\n");
#endif
				computer = -1;
				auto_play = 0;
				verify_game_results();
				continue;
			}
            update_moves_to_go(board.side_on_move);
			make_move(move);
			util_get_move_string(move, move_string);
			printf("move %s\n", move_string);
			if (auto_play)
				board_print(0);
#ifdef LOG
			fprintf(log_file, "move %s\n", move_string);
			fflush(log_file);
#endif

            if (get_game_result() != GR_NOT_FINISH)
            {
				verify_game_results();
                auto_play = 0;
            }

			continue;
		}
		if (!fgets(line, 512, stdin))
			return 0;
	
		if (line[0] == '\n')
			continue;
#ifdef LOG
		fprintf(log_file, "--> %s\n", line);
		fflush(log_file);
#endif
		sscanf(line, "%s", command);
		if (!strcmp(command, "xboard")) 
        {
            printf("%s\n", xboard_reply);
			continue;
		}
		if (!strcmp(command, "new")) 
        {
			new_game(FEN_NEW_GAME);
			computer = BLACK;
			continue;
		}
		if (!strcmp(command, "quit")) 
        {
			stop = 1;
			continue;
		}
		if (!strcmp(command, "force")) 
        {
			computer = -1;
			continue;
		}
		if (!strcmp(command, "sd")) 
        {
			sscanf(line, "sd %d", &max_depth);
			search.max_depth = max_depth;
			//search.move_time = (60 * 60 * 1000); // 1 hour
			continue;
		}
		if (!strcmp(command, "st")) 
        {
			sscanf(line, "st %d", &max_time);
			max_time *= 1000;
			calc_move_time(search.moves_to_go[board.side_on_move], 0, max_time);
			//search.move_time = max_time - 100;
			//search.max_depth = MAX_DEPTH;
			continue;
		}
		if (!strcmp(command, "level")) 
        {
			// just get the moves to go. use "time" command.
			sscanf(line, "level %d", &search.moves_level);
			if (search.moves_level < 0)
				search.moves_level = 0;
            search.moves_to_go[WHITE] = search.moves_level - 1;
            if (search.moves_to_go[WHITE] < 0)
                search.moves_to_go[WHITE] = 0;
            search.moves_to_go[BLACK] = search.moves_level - 1;
            if (search.moves_to_go[BLACK] < 0)
                search.moves_to_go[BLACK] = 0;
#ifdef LOG
			fprintf(log_file, "moves_to_go from level command=%d\n", search.moves_level);
			fflush(log_file);
#endif
			continue;
		}
		if (!strcmp(command, "time")) 
        {
			sscanf(line, "time %d", &max_time);
            if (time1000)
                max_time *= 1000; // when using arena to connect fics.org needs this.
            else
			    max_time *= 10;
			calc_move_time(search.moves_to_go[board.side_on_move], max_time, 0);
#ifdef LOG
			fprintf(log_file, "max_time=%d MAXDEP=%d moves_to_go=%d move_time=%d\n", 
                max_time, MAX_DEPTH, search.moves_to_go[board.side_on_move], search.move_time);
			fflush(log_file);
#endif
			continue;
		}
		if (!strcmp(command, "otim")) 
        {
			continue;
		}
		if (!strcmp(command, "go")) 
        {
			computer = board.side_on_move;
			continue;
		}
		if (!strcmp(command, "hint")) 
        {
			think(post, &move, use_book);
			if (!move)
				continue;
			util_get_move_string(move, move_string);
			printf("Hint: %s", move_string);
			continue;
		}
		if (!strcmp(command, "random"))
			continue;
		if (!strcmp(command, "hard"))
			continue;
		if (!strcmp(command, "easy"))
			continue;
		if (!strcmp(command, "computer"))
			continue;
		if (!strcmp(command, "white")) 
			continue;
		if (!strcmp(command, "black")) 
			continue;
		if (!strcmp(command, "protover"))
        {
            printf("%s\n", xboard_reply);
			continue;
		}
		if (!strcmp(command, "accepted"))
			continue;
		if (!strcmp(command, "rejected"))
        {
#ifdef LOG
			fprintf(log_file, "rejected command received: [%s]\n", line);
#endif
            continue;
		}
		if (!strcmp(command, "result"))
			continue;
		if (!strcmp(command, "undo"))
        {
			if (histply)
				undo_move();
			continue;
		}
		if (!strcmp(command, "remove"))
        {
			if (histply)
				undo_move();
			if (histply)
				undo_move();
			continue;
		}
		if (!strcmp(command, "post1"))
        {
			post = POST_DEFAULT;
			continue;
		}
		if (!strcmp(command, "post"))
        {
			post = POST_XBOARD;
			continue;
		}
		if (!strcmp(command, "nopost"))
        {
			post = POST_NONE;
			continue;
		}
		if (!strcmp(command, "d"))
        {
			board_print(0);
            print_current_moves();
			continue;
		}
		if (!strcmp(command, "eval"))
        {
			eval_print();
			continue;
		}
		if (!strcmp(command, "perft"))
        {
			perft_depth = 0;
			sscanf(line, "perft %d", &perft_depth);
			if (perft_depth == 0)
				printf("syntax: perft <depth>\n");
			else
				perft(perft_depth);
			continue;
		}
		if (!strcmp(command, "perftx"))
        {
			perftx();
			continue;
		}
		if (!strcmp(command, "perfty"))
        {
			perfty();
			continue;
		}
		if (!strcmp(command, "dev"))
        {
			show_dev_info();
			continue;
		}
		if (!strcmp(command, "auto"))
        {
			auto_play = 1;
			continue;
		}
		if (!strcmp(command, "setboard"))
        {
			if (strlen(line) > 9)
            {
				new_game(&line[9]);
				computer = -1;
				use_book = 0;
			}
			continue;
		}
		if (!strcmp(command, "fen"))
        {
			if (strlen(line) > 4)
            {
				new_game(&line[4]);
				computer = -1;
				use_book = 0;
			}
			else
				printf("no fen position entered.\n");
			continue;
		}
		if (!strcmp(command, "book"))
        {
			use_book = !use_book;
			printf("book use is %s\n", (use_book ? "ON" : "OFF"));
			continue;
		}
		if (!strcmp(command, "bench"))
        {
            bench(12);
            continue;

            // bench(12) Signature: 12861233  Time: 80.82
        }

		if (!strcmp(command, "ttt1"))
		{
			printf("transposition table test: best move is a1b1 (found quick with good score)\n");
			printf("8/k/3p4/p2P1p2/P2P1P2/8/8/K7 w - - 0 1\n");
			new_game("8/k/3p4/p2P1p2/P2P1P2/8/8/K7 w - - 0 1");
			computer = board.side_on_move;
			continue;
		}
		if (!strcmp(command, "ttt2"))
        {
			printf("transposition table test: best move is c7\n");
			printf("2k5/8/1pP1K3/1P6/8/8/8/8 w - -\n");
			new_game("2k5/8/1pP1K3/1P6/8/8/8/8 w - -");
			computer = board.side_on_move;
			continue;
		}
		if (!strcmp(command, "epd"))
        {
			if (strlen(line) < 5)
            {
				printf("syntax: epd <epd file name>\n");
				continue;
			}
			sscanf(line, "epd %s", epd_file);
			epd(epd_file);
			continue;
		}
		if (!strcmp(command, "evtest"))
        {
			if (strlen(line) < 7) {
				printf("syntax: evtest <epd file name>\n");
				continue;
			}
			sscanf(line, "evtest %s", epd_file);
			eval_test(epd_file);
			continue;
		}
		if (!strcmp(command, "help"))
        {
			printf("You can use XBOARD commands and the following:\n");
			printf("d - show board\n");
			printf("eval - print score for current position\n");
			printf("post1 - show formatted search information\n");
			printf("epd <filename> - test epd file\n");
			printf("perft <n> - show perft move count from current position\n");
			printf("\n");
			printf("Command line options:\n");
			printf("    tucano -hash <MB> -nobook\n");
			printf("\n");
			printf("   -hash indicates the size of hash table, default = 64 MB, minimum = 8 MB\n");
			printf("   -nobook disables the internal opening book\n");
			printf("\n");
			continue;
		}
		util_parse_move(command, &parsed);
		if (!parsed)
        {
			printf("Error (unknown command): %s\n", command);
#ifdef LOG
			fprintf(log_file, "Error (unknown command): [%s]\n", command);
#endif
		}
		else
        {
            update_moves_to_go(board.side_on_move);
			make_move(parsed);
			verify_game_results();
		}
	}

#ifdef LOG
	fclose(log_file);
#endif

	return 0;
}

void update_moves_to_go(int color)
{
    if (search.moves_to_go[color])
        search.moves_to_go[color]--;
    if (search.moves_to_go[color] == 0)
        search.moves_to_go[color] = search.moves_level;
}

void calc_move_time(UINT moves_to_go, UINT max_time, UINT move_time)
{
    //  Specific move time
    if (move_time > 0)
    {
        if (move_time > 20)
            move_time -= 10;
        search.normal_move_time = move_time;
        search.extended_move_time = move_time;
        return;
    }

    //  Calculate normal move time
    max_time = (int)(max_time * 0.95);

    if (moves_to_go <= 0 || moves_to_go > 30)
        moves_to_go = 30;
    
    search.normal_move_time = (int)(max_time / moves_to_go);

    if (search.normal_move_time >= max_time) 
		move_time = max_time;

    //  Calculate extended move time
    search.extended_move_time = (int)(search.normal_move_time * 4);
    if (search.extended_move_time > max_time)
        search.extended_move_time = search.normal_move_time;
}

int get_game_result(void)
{
	int     can_move = 0;
    MOVE    move;
    MOVES   ml;


	board.ply = 0;
	
    moves_init(&ml, is_incheck(board.side_on_move), 0, 0);

	while((move = moves_next(&ml)))
    {
		make_move(move);
        if (is_illegal(move))
        {
            undo_move();
            continue;
        }
		can_move = 1;
		undo_move();
	}

	if (can_move == 0) 
    {
		if (is_incheck(board.side_on_move)) 
        {
			if (board.side_on_move == WHITE)
				return GR_BLACK_WIN;
			else
				return GR_WHITE_WIN;
		}
		else
			return GR_DRAW;
	}
	else 
    {
        if (board.fifty_move_rule >= 100)
            return GR_DRAW;
        else
            if (insufficient_material())
				return GR_DRAW;

            else
                if (is_draw())
					return GR_DRAW;
	}  
	return GR_NOT_FINISH;
}

void verify_game_results(void)
{
	int     can_move = 0;
    MOVE    move;
    MOVES   ml;

	board.ply = 0;
	
    moves_init(&ml, is_incheck(board.side_on_move), 0, 0);

	while((move = moves_next(&ml)))
    {
		make_move(move);
        if (is_illegal(move))
        {
            undo_move();
            continue;
        }
		can_move = 1;
		undo_move();
	}

	if (can_move == 0) 
    {
		if (is_incheck(board.side_on_move)) 
        {
			if (board.side_on_move == WHITE)
				printf("0-1 {Black mates}\n");
			else
				printf("1-0 {White mates}\n");
		}
		else
			printf("1/2-1/2 {Stalemate}\n");
	}
	else 
    {
        if (board.fifty_move_rule >= 100)
            printf("1/2-1/2 {Draw by fifty move rule}\n");
        else
            if (insufficient_material())
                printf("1/2-1/2 {Insufficient Material}\n");
            else
                if (is_draw())
                    printf("1/2-1/2 {Draw by repetition}\n");
	}   
}

// Search a couple of positions and give the count of nodes searched (signature).
// It is very useful to test non-functional changes.
// Reference: discocheck/stockisfh engines)
void bench(int depth)
{
	char *test[] =
	{
        "3r2k1/pp2b1pp/5p2/4p3/3r4/P2P2BP/1P3PP1/1R1R1K2 b -- - 0 24",
        "rn1qk1nr/pp3pp1/2pbp1p1/8/2pP1BP1/2N5/PP1QPPBP/R3K2R b KQkq - 0 10",
        "r1bq1rk1/ppppbppp/2n5/1B6/4P3/3Q4/PPP2PPP/RNB2RK1 w -- - 0 9",
        "8/2B4k/R6p/1p3pp1/1P6/2P1r1K1/2n5/8 w -- - 0 46",
        "r1b1kbnr/6pp/p1pq1p2/3p4/4P3/5N2/PPP2PPP/RNBQR1K1 w -kq - 0 10",
        "N1bk4/bp1p4/p1nN1p1p/4p3/1PB1P1pq/2PP4/P4PP1/R2Q1RK1 w -- - 0 19",
        "1r2r1k1/3bnppp/p2q4/2RPp3/4P3/6P1/2Q1NPBP/2R3K1 w - -",
        "1qr3k1/p2nbppp/bp2p3/3p4/3P4/1P2PNP1/P2Q1PBP/1N2R1K1 b - -",
        "2b1k2r/2p2ppp/1qp4n/7B/1p2P3/5Q2/PPPr2PP/R2N1R1K b k -",
        "2b5/1p4k1/p2R2P1/4Np2/1P3Pp1/1r6/5K2/8 w - -",
        "2brr1k1/ppq2ppp/2pb1n2/8/3NP3/2P2P2/P1Q2BPP/1R1R1BK1 w - -",
        "2kr2nr/1pp3pp/p1pb4/4p2b/4P1P1/5N1P/PPPN1P2/R1B1R1K1 b - -",
        "2r1k2r/1p1qbppp/p3pn2/3pBb2/3P4/1QN1P3/PP2BPPP/2R2RK1 b k -",
        "2r1r1k1/pbpp1npp/1p1b3q/3P4/4RN1P/1P4P1/PB1Q1PB1/2R3K1 w - -",
        "2r2k2/r4p2/2b1p1p1/1p1p2Pp/3R1P1P/P1P5/1PB5/2K1R3 w - -",
        "2r3k1/5pp1/1p2p1np/p1q5/P1P4P/1P1Q1NP1/5PK1/R7 w - -",
        "r3kb1r/pp3ppp/4bn2/3p4/P7/4N1P1/1P2PPBP/R1B1K2R w KQkq -",
        "r3kbnr/1pp3pp/p1p2p2/8/3qP3/5Q1P/PP3PP1/RNB2RK1 w kq -",
        "r3kr2/pppb1p2/2n3p1/3Bp2p/4P2N/2P5/PP3PPP/2KR3R b q -",
        "r3nrk1/pp2qpb1/3p1npp/2pPp3/2P1P2N/2N3Pb/PP1BBP1P/R2Q1RK1 w - -",
        NULL
    };
	U64     signature = 0;
    UINT    start;
    UINT    end;
    double  elapsed;
    MOVE    move;
    int     i;
    int     t;

	search.max_depth = depth;
    search.normal_move_time = 300000;
    search.extended_move_time = 300000;

    printf("Benchmark (depth=%d)\n", depth);

    start = util_get_time();

    for (t = 0; test[t]; t++);

	for (i = 0; test[i]; i++)
    {
        printf("%d/%d) %s\n", i + 1, t, test[i]);

        new_game(test[i]);

        think(POST_NONE, &move, 0);

        signature += search.nodes;
	}

    end = util_get_time();
    elapsed = (double)(end - start) / (double)1000;

    printf("\nSignature: %llu  Time: %3.2f\n", signature, elapsed);
}

// for quick tests.
void show_dev_info(void) 
{

}

//END
