/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Generate, order and select the next move when in quiescence search.
//-------------------------------------------------------------------------------------------------

enum    {msMoveHint, msGenCaps, msNextCap, msGenChecks, msNextCheck};

void    quiesce_moves_assign_caps_score(MOVES *ml);
void    quiesce_moves_assign_check_score(MOVES *ml);
void    quiesce_moves_select_next(MOVES *ml);

//-------------------------------------------------------------------------------------------------
//  Preparation for move generation and selection.
//-------------------------------------------------------------------------------------------------
void quiesce_moves_init(MOVES *ml, int incheck, int gen_checks)
{
	ml->incheck = incheck;
	ml->caps = 1;
    ml->gen_checks = gen_checks;
	ml->next = 0;
    ml->bc_count = 0;
    ml->bc_next = 0;
    ml->ttm = MOVE_NONE;
    ml->phase = msGenCaps;
}

//-------------------------------------------------------------------------------------------------
//  Return the next move to be made.
//-------------------------------------------------------------------------------------------------
MOVE quiesce_moves_next(MOVES *ml)
{
	switch(ml->phase)
	{
	case msGenCaps:
        assert(ml->incheck == 0 || ml->incheck == 1);
        ml->count = 0;
        if (ml->incheck)
        {
            gen_check_evasions(ml);
            quiesce_moves_assign_check_score(ml);
        }
        else
        {
            gen_caps(ml);
            quiesce_moves_assign_caps_score(ml);
        }
        assert(ml->count < MAX_MOVE);
        ml->phase = msNextCap;
	case msNextCap:
        assert(ml->next < MAX_MOVE);
		if (ml->next < ml->count)
		{
			quiesce_moves_select_next(ml);
            return ml->moves[ml->next++];
		}
        ml->phase = msGenChecks;
	case msGenChecks:
        if (ml->incheck == 1 || ml->gen_checks == 0)
            return MOVE_NONE;
        ml->count = 0;
        gen_moves(ml);
        assert(ml->count < MAX_MOVE);
		ml->next = 0;
		ml->phase = msNextCheck;
	case msNextCheck:
        assert(ml->next < MAX_MOVE);
		if (ml->next < ml->count)
		{
			return ml->moves[ml->next++];
		}
        return MOVE_NONE;
    default:
        assert(0);
	}
    return MOVE_NONE;
}

//-------------------------------------------------------------------------------------------------
//  Indicate if phase is next check move.
//-------------------------------------------------------------------------------------------------
int quiesce_moves_next_check(MOVES *ml)
{
    return ml->phase == msNextCheck ? 1 : 0;
}

//-------------------------------------------------------------------------------------------------
//  Put the move with highest score in front of the list to be picked next.
//-------------------------------------------------------------------------------------------------
void quiesce_moves_select_next(MOVES *ml) 
{
	int     i;
    int     best_index = ml->next;
	MOVE    temp_move;
	int	    temp_score;

	for (i = ml->next + 1; i < ml->count; i++) 
    {
		if (ml->score[i] > ml->score[best_index]) 
			best_index = i;
	}
	if (best_index != ml->next)
    {
		temp_move = ml->moves[ml->next];
		ml->moves[ml->next] = ml->moves[best_index];
		ml->moves[best_index] = temp_move;

		temp_score = ml->score[ml->next];
		ml->score[ml->next] = ml->score[best_index];
		ml->score[best_index] = temp_score;
	}
}

//-------------------------------------------------------------------------------------------------
//  Assign the score.
//-------------------------------------------------------------------------------------------------
void quiesce_moves_assign_check_score(MOVES *ml)
{
    MOVE    move;
    int     score;
    int     i;

    for (i = 0; i < ml->count; i++)
    {
        move = ml->moves[i];
        if (move == ml->ttm) // exclude tt move from list.
        {
            ml->count--;
            if (i == ml->count) // last move
                break;
            move = ml->moves[i] = ml->moves[ml->count]; // replace with last move
        }
        if (IS_QUIET(move))
        {
            ml->score[i] = move_hist_val(board.side_on_move, move);
            if (move_is_killer(board.side_on_move, board.ply, move))
                ml->score[i] += SORT_KILL;
            assert(ml->score[i] < SORT_CAPS);
            continue;
        }
        switch(TYPE(move))
        {
        case MT_CAPPC:
            score = capture_value[MVPC(move)][CPPC(move)];
            break;
        case MT_EPCAP:
            score = capture_value[PAWN][PAWN];
            break;
        case MT_PROMO:
            score = promotion_value[PROM(move)];
            break;
        case MT_CPPRM:
            score = prom_capt_value[PROM(move)][CPPC(move)];
            break;
        default:
            score = 0;
        }
        ml->score[i] = SORT_CAPS + score;
    }
}

//-------------------------------------------------------------------------------------------------
//  Assign the score.
//-------------------------------------------------------------------------------------------------
void quiesce_moves_assign_caps_score(MOVES *ml)
{
	int     i;
    int     score;
    MOVE    move;

	for (i = 0; i < ml->count; i++)
    {
        move = ml->moves[i];
        if (move == ml->ttm) // exclude tt move from list.
        {
            ml->count--;
            if (i == ml->count) // last move
                break;
            move = ml->moves[i] = ml->moves[ml->count]; // replace with last move
        }

        switch(TYPE(move))
        {
        case MT_CAPPC:
            score = capture_value[MVPC(move)][CPPC(move)];
            break;
        case MT_EPCAP:
            score = capture_value[PAWN][PAWN];
            break;
        case MT_PROMO:
            score = promotion_value[PROM(move)];
            break;
        case MT_CPPRM:
            score = prom_capt_value[PROM(move)][CPPC(move)];
            break;
        default:
            score = 0;
        }

        ml->score[i] = score;
	}
}

// end
