/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Move generation, sort and selection for zero window search.
//-------------------------------------------------------------------------------------------------

enum    {msMoveHint, msGenCaps, msNextCap, msGenQuiet, msNextQuiet, msNextBadCap};

void    moves_assign_quiet_score(MOVES *ml);
void    moves_assign_caps_score(MOVES *ml);
void    moves_assign_check_score(MOVES *ml);
void    moves_select_next(MOVES *ml);

//-------------------------------------------------------------------------------------------------
//  Preparation for move generation and selection.
//-------------------------------------------------------------------------------------------------
void moves_init(MOVES *ml, int incheck, MOVE ttm, int caps)
{
	ml->incheck = incheck;
	ml->caps = caps;
	ml->next = 0;
    ml->bc_count = 0;
    ml->bc_next = 0;
    ml->sort = 1;
    if (ttm != MOVE_NONE && is_valid(ttm))
    {
        ml->ttm = ttm;
    	ml->phase = msMoveHint;
    }
    else
    {
        ml->ttm = MOVE_NONE;
        ml->phase = msGenCaps;
    }
}

//-------------------------------------------------------------------------------------------------
//  Goto gen quiet phase
//-------------------------------------------------------------------------------------------------
void moves_select_quiet(MOVES *ml)
{
    ml->phase = msGenQuiet;
}

int moves_is_bad_cap(MOVES *ml)
{
    return (ml->phase == msNextBadCap ? 1 : 0);
}

//-------------------------------------------------------------------------------------------------
//  Return the next move to be made based in the following sequence:
//  1 - return the hint move, usually a move from transposition table.
//  2 - capture moves (or check escapes)
//  3 - all other moves.
//  The idea is to generate moves as necessary and hopefully skip the work if 
//  search function gets an early cut-off.
//-------------------------------------------------------------------------------------------------
MOVE moves_next(MOVES *ml)
{
	switch(ml->phase)
	{
	case msMoveHint:
        ml->phase = msGenCaps;
        assert(is_valid(ml->ttm));
        return ml->ttm;
	case msGenCaps:
        assert(ml->incheck == 0 || ml->incheck == 1);
        ml->count = 0;
        if (ml->incheck)
        {
            gen_check_evasions(ml);
            moves_assign_check_score(ml);
        }
        else
        {
            gen_caps(ml);
            moves_assign_caps_score(ml);
        }
        assert(ml->count < MAX_MOVE);
        ml->phase = msNextCap;
        ml->next = 0;
	case msNextCap:
        assert(ml->next < MAX_MOVE);
		while(ml->next < ml->count)
		{
            if (ml->incheck == 0 && ml->caps == 0 && is_badcap(ml->moves[ml->next]))
            {
                ml->bc_moves[ml->bc_count++] = ml->moves[ml->next++];
                continue;
            }
            moves_select_next(ml);
            return ml->moves[ml->next++];
		}
        ml->phase = msGenQuiet;
	case msGenQuiet:
        if (ml->caps || ml->incheck)
            return MOVE_NONE;
        ml->count = 0;
        gen_moves(ml);
        assert(ml->count < MAX_MOVE);
        moves_assign_quiet_score(ml);
		ml->next = 0;
		ml->phase = msNextQuiet;
	case msNextQuiet:
        assert(ml->next < MAX_MOVE);
		if (ml->next < ml->count)
		{
            if (ml->sort && ml->next < 10)
            {
			    moves_select_next(ml);
                ml->sort = ml->score[ml->next];
            }
			return ml->moves[ml->next++];
		}
        ml->phase = msNextBadCap;
        ml->bc_next = 0;
	case msNextBadCap:
        assert(ml->bc_next < MAX_MOVE);
		if (ml->bc_next < ml->bc_count)
			return ml->bc_moves[ml->bc_next++];
        return MOVE_NONE;
    default:
        assert(0);
	}
    return MOVE_NONE;
}

//-------------------------------------------------------------------------------------------------
//  Return the next move to the perf function. It is a simplified faster version.
//-------------------------------------------------------------------------------------------------
MOVE moves_next_perft(MOVES *ml)
{
	switch(ml->phase)
	{
	case msMoveHint:
		ml->phase = msGenCaps;
	case msGenCaps:
        ml->count = 0;
        if (ml->incheck)
            gen_check_evasions(ml);
        else
    		gen_caps(ml);
		ml->phase = msNextCap;
        ml->next = 0;
	case msNextCap:
		if (ml->next < ml->count)
			return ml->moves[ml->next++];
		ml->phase = msGenQuiet;
        ml->next = 0;
	case msGenQuiet:
        if (ml->incheck)
            return MOVE_NONE;
        ml->count = 0;
		gen_moves(ml);
		ml->next = 0;
		ml->phase = msNextQuiet;
	case msNextQuiet:
		if (ml->next < ml->count)
			return ml->moves[ml->next++];
		return MOVE_NONE;
    default:
        assert(0);
	}
    return MOVE_NONE;
}

//-------------------------------------------------------------------------------------------------
//  Put the move with highest score in front of the list to be picked next.
//-------------------------------------------------------------------------------------------------
void moves_select_next(MOVES *ml) 
{
	int     i;
    int     best_index = ml->next;
	MOVE    temp_move;
	int	    temp_score;

	for (i = ml->next + 1; i < ml->count; i++) 
    {
		if (ml->score[i] > ml->score[best_index]) 
			best_index = i;
	}
	if (best_index != ml->next)
    {
		temp_move = ml->moves[ml->next];
		ml->moves[ml->next] = ml->moves[best_index];
		ml->moves[best_index] = temp_move;

		temp_score = ml->score[ml->next];
		ml->score[ml->next] = ml->score[best_index];
		ml->score[best_index] = temp_score;
	}
}

//------------------------------------------------------------------------------------
//  Assign the score for quiet moves.
//------------------------------------------------------------------------------------
void moves_assign_quiet_score(MOVES *ml)
{
	int     i;
    MOVE    move;

	for (i = 0; i < ml->count; i++)
    {
        move = ml->moves[i];
        if (move == ml->ttm) // exclude tt move from list.
        {
            ml->count--;
            if (i == ml->count) // last move
                break;
            move = ml->moves[i] = ml->moves[ml->count]; // replace with last move
        }

		ml->score[i] = move_hist_val(board.side_on_move, move);
        
        assert(ml->score[i] < SORT_KILL);

        if (move_is_killer(board.side_on_move, board.ply, move))
            ml->score[i] += SORT_KILL;
	}
}

//------------------------------------------------------------------------------------
//  Assign the score when in check.
//------------------------------------------------------------------------------------
void moves_assign_check_score(MOVES *ml)
{
    MOVE    move;
    int     score;
    int     i;

    for (i = 0; i < ml->count; i++)
    {
        move = ml->moves[i];
        if (move == ml->ttm) // exclude tt move from list.
        {
            ml->count--;
            if (i == ml->count) // last move
                break;
            move = ml->moves[i] = ml->moves[ml->count]; // replace with last move
        }
        if (IS_QUIET(move))
        {
            ml->score[i] = move_hist_val(board.side_on_move, move);
            if (move_is_killer(board.side_on_move, board.ply, move))
                ml->score[i] += SORT_KILL;
            assert(ml->score[i] < SORT_CAPS);
            continue;
        }
        switch(TYPE(move))
        {
        case MT_CAPPC:
            score = capture_value[MVPC(move)][CPPC(move)];
            break;
        case MT_EPCAP:
            score = capture_value[PAWN][PAWN];
            break;
        case MT_PROMO:
            score = promotion_value[PROM(move)];
            break;
        case MT_CPPRM:
            score = prom_capt_value[PROM(move)][CPPC(move)];
            break;
        default:
            score = 0;
        }
        ml->score[i] = SORT_CAPS + score;
    }
}

//------------------------------------------------------------------------------------
//  Assign the score for captures.
//------------------------------------------------------------------------------------
void moves_assign_caps_score(MOVES *ml)
{
	int     i;
    int     score;
    MOVE    move;

	for (i = 0; i < ml->count; i++)
    {
        move = ml->moves[i];
        if (move == ml->ttm) // exclude tt move from list.
        {
            ml->count--;
            if (i == ml->count) // last move
                break;
            move = ml->moves[i] = ml->moves[ml->count]; // replace with last move
        }

        switch(TYPE(move))
        {
        case MT_CAPPC:
            score = capture_value[MVPC(move)][CPPC(move)];
            break;
        case MT_EPCAP:
            score = capture_value[PAWN][PAWN];
            break;
        case MT_PROMO:
            score = promotion_value[PROM(move)];
            break;
        case MT_CPPRM:
            score = prom_capt_value[PROM(move)][CPPC(move)];
            break;
        default:
            score = 0;
        }

        ml->score[i] = score;
	}
}

// end
