/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Main search loop.
//-------------------------------------------------------------------------------------------------

U64     cnt = 0, hit = 0;

//-------------------------------------------------------------------------------------------------
//  Search start.
//-------------------------------------------------------------------------------------------------
void think(int post_flag, MOVE *move, int usebook) 
{
	int     depth;
    int     score;
    int     incheck;
    int     prev_score;
	MOVE    bookmove;
	int     min_score, max_score;
	UINT    used_time;

	assert(post_flag == POST_NONE || post_flag == POST_XBOARD || post_flag == POST_DEFAULT);
	assert(move != 0);
	assert(usebook == 0 || usebook == 1);

    test_on = 0;
    test_cnt = test_hit = 0;

	//  Try to find a move from book.
	if (usebook) 
    {
		bookmove = book_next_move();
		if (bookmove != MOVE_NONE)
        {
            *move = bookmove;
            return;
        }
	}

	//  Prepare the search
	*move = MOVE_NONE;
	board.ply = 0;
	search.post_flag = post_flag;
	search.nodes = 0;
	search.qnodes = 0;
    memset(position, 0, sizeof(position));
    move_hist_init();

    tt_age();

   	incheck = is_incheck(board.side_on_move);

	//  Start time control
	search.start_time = util_get_time();
    search.can_stop = 0;
    search.normal_finish_time = search.start_time + search.normal_move_time;
    search.extended_finish_time = search.start_time + search.extended_move_time;
	search.stop = 0;
	setjmp(env);
	if (search.stop)
    {
		while (board.ply)
			undo_move();
        goto search_end;
	}

#ifdef LOG
	fprintf(log_file, "search.start1: start_time=%u  finish_time=%u   move_time: %u current: %u\n", search.start_time, search.finish_time, search.move_time, util_get_time());
	fflush(log_file);
#endif

	if (search.post_flag == POST_DEFAULT)
		printf("Ply      Nodes  Score Time Principal Variation\n");

	min_score = -INFINITY;
	max_score = INFINITY;

#ifdef LOG
	fprintf(log_file, "search.start2: start_time=%u  finish_time=%u   move_time: %u  current: %u\n", search.start_time, search.finish_time, search.move_time, util_get_time());
	fflush(log_file);
#endif

    //  Start the iterative deepening
	for (depth = 1; depth <= search.max_depth; depth++)
    {

        search.cur_depth = depth;

#ifdef LOG
        fprintf(log_file, "depth: %d current: %u search.min_depth: %d\n", depth, util_get_time(), search.min_depth);
#endif

        //  Search root moves
        score = search_pv(incheck, min_score, max_score, depth);
#ifdef LOG
        fprintf(log_file, "search1: depth: %d score %d\n", depth, score);
#endif

        //  Check if score dropped from last iteration
        search.score_drop = 0;
        if (depth > 1 && score + 50 < prev_score)
        {
            search.score_drop = 1;
        }

        //  Get the best move from principal variation.
		post_info(score, depth);

		//  Leave if it's mate score.
		if (score > MAX_SCORE || score < -MAX_SCORE)
			break;

        //  Don't start another iteration if most of time was used.
        if (search.score_drop == 0 && depth >= search.min_depth)
        {
			used_time = util_get_time() - search.start_time;
			if ((double)used_time >= ((double)search.normal_move_time * 0.6))
				break;
		}

        prev_score = score;
	}

search_end:
    *move = position[0].pv_line[0];

#ifdef LOG
	fprintf(log_file, " search.stop: start_time=%u  finish_time=%u   move_time: %u   current: %u\n", 
		              search.start_time, search.finish_time, search.move_time, util_get_time());
#endif

    // print counters, used for testing.
    if (test_on)
    {
        cnt += test_cnt;
        hit += test_hit;
        printf("-------->  TEST  CNT=%llu HIT=%llu PCT=%3.4f\n", test_cnt, test_hit, (float)(test_cnt == 0 ? 0 : test_hit * 100.0 / test_cnt));
        printf("-------->  TOTAL CNT=%llu HIT=%llu PCT=%3.4f\n", cnt, hit, (float)(cnt == 0 ? 0 : hit * 100.0 / cnt));
    }
}

//END
