/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Principal variation search.
//-------------------------------------------------------------------------------------------------
int search_pv(UINT incheck, int alpha, int beta, int depth)
{
    MOVES   ml;
	MOVE    best_move  = MOVE_NONE;
    MOVE    hash_move  = MOVE_NONE;
    int     best_score = -INFINITY;
	int     score = 0;
	UINT    new_incheck;
    int	    total_move_count = 0;
	int		quiet_move_count = 0;
	int     extensions;
    int     ply = board.ply;
    int     som = board.side_on_move;
    MOVE    move;
    MOVE    searched[MAX_MOVE];

	assert(incheck == 0 || incheck == 1);
	assert(alpha >= -INFINITY && alpha <= INFINITY);
	assert(beta >= -INFINITY && beta <= INFINITY);
    assert(beta > alpha);
	assert(depth <= MAX_DEPTH);

    if (depth <= 0)
		return quiesce(incheck, alpha, beta, 0);

	search.nodes++;

    if (is_draw())
		return 0;

    check_time();

    assert(board.ply >= 0 && board.ply < MAX_PLY);

	position[ply].pv_size = ply;

	//  Get move hint from transposition table
    hash_move = tt_move();

	// Internal Iterative Deepening.
    if (depth > 3 && hash_move == MOVE_NONE)
    {
        score = search_pv(incheck, alpha, beta, depth - 3);
        if (score <= alpha)
        {
            score = search_pv(incheck, -INFINITY, beta, depth - 3);
        }
        hash_move = tt_move();
    }

    //  Loop through move list
    pv_moves_init(&ml, incheck, hash_move, 0);
    while((move = pv_moves_next(&ml)) != MOVE_NONE)
    {
        assert(is_valid(move));

        //  Extension if move puts opponent in check.
        if (is_check(move))
        {
            new_incheck = 1;
            extensions = 1;
        }
        else
        {
            new_incheck = 0;
            extensions = 0;

            // Pawn on 7th extension
            if (MVPC(move) == PAWN && ((som == WHITE && TOSQ(move) <= H7) || (som == BLACK && TOSQ(move) >= A2)))
			{
				if (see_move(move) >= 0)
				{
					extensions = 1;
				}
			}
        }

        // Make move.
		make_move(move);

        // Make sure king is not in check or castle is legal.
		if (is_illegal(move))
        {
			undo_move(); 
			continue;
		}

        // Search new position.
		if (best_score == -INFINITY)
        {
			score = -search_pv(new_incheck, -beta, -alpha, depth - 1 + extensions);
        }
		else 
        {
            score = -search_zw(new_incheck, -alpha, depth - 1 + extensions, 1);
            if (score > alpha)
			{
                score = -search_pv(new_incheck, -beta, -alpha, depth - 1 + extensions);
			}
		}

        undo_move();
        total_move_count++;
        if (IS_QUIET(move))
		{
            searched[quiet_move_count++] = move;
		}

        assert(board.key == zk_board_key());
        assert(board.pawn_key == zk_pawn_key());

        if (score > best_score)
        {
            if (score > alpha)
            {
                if (score >= beta)
                {
                    if (IS_QUIET(move))
                    {
                        move_hist_save(som, ply, move, depth);
                        move_hist_bad(som, searched, quiet_move_count - 1);
                    }
                    tt_save(depth, score, TT_LOWER, move);
                    return score;
                }
                update_pv(move);
                alpha = score;
                best_move = move;
            }
            best_score = score;
        }
	}

	if (total_move_count == 0)
		return (incheck ? -MATE_VALUE + ply : 0);

    if (best_move != MOVE_NONE)
    {
        tt_save(depth, best_score, TT_EXACT, best_move);
    }
    else
    {
        tt_save(depth, best_score, TT_UPPER, MOVE_NONE);
    }

	return best_score;
}

// end
