/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Quiescence search
//-------------------------------------------------------------------------------------------------

static const int USE_DELTA = 1;
static const int USE_SEE = 1;

//-------------------------------------------------------------------------------------------------
//  Search nodes until a quiet position is found.
//-------------------------------------------------------------------------------------------------
int quiesce(UINT incheck, int alpha, int beta, int depth) 
{
    MOVES   ml;
	int     score;
    UINT    total_move_count = 0;
    MOVE    move = MOVE_NONE;
    int     best_score = -INFINITY;
    int     som = board.side_on_move;
    int     cppc;
    int     new_incheck;

	search.qnodes++;

    if (is_draw())
		return 0;

	if (board.ply >= MAX_PLY - 1)
		return evaluate();

	position[board.ply].pv_size = board.ply;

	check_time();
  
    if (!incheck)
    {
        best_score = evaluate();
        if (best_score >= beta)
            return best_score;
        if (best_score > alpha)
            alpha = best_score;
    }

    quiesce_moves_init(&ml, incheck, (depth >= 0 ? 1 : 0));

    while((move = quiesce_moves_next(&ml)) != MOVE_NONE)
    {
        assert(is_valid(move));

        new_incheck = is_check(move);

        if (quiesce_moves_next_check(&ml) && !new_incheck)
            continue;

        //  Skip moves that are not going to improve the position.
        if (!incheck && !new_incheck && TYPE(move) == MT_CAPPC)
        {
            cppc = CPPC(move);

            // Skip captures that will not improve alpha (delta pruning)
            if (USE_DELTA && best_score + 100 + piece_value[cppc] <= alpha)
                continue;

            // Skip losing captures based on Static Exchange Evaluation.
            if (USE_SEE && see_value[cppc] < see_value[MVPC(move)])
            {
                if (is_defended_by_pawn(TOSQ(move), flip_color(som)))
                    continue;
                if (see_capture(move) < 0)
                {
                    continue;
                }
            }
        }

        make_move(move);
        if (is_illegal(move))
        {
            undo_move(); 
            continue;
        }

        if (!incheck && depth < -10)
        {
            undo_move();
            score = best_score + see_move(move);
        }
        else
        {
            score = -quiesce(new_incheck, -beta, -alpha, depth - 1);
            undo_move();
        }
        total_move_count++;

        if (score > best_score)
        {
            if (score > alpha) 
            {
                update_pv(move);
                if (score >= beta)
                {
                    return score;
                }
                alpha = score;
            }
            best_score = score;
        }
    }

	if (total_move_count == 0 && incheck)
		return -MATE_VALUE + board.ply;

    return best_score;
}

// END