/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Search utilities
//-------------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------------
//  Search Initializations
//-------------------------------------------------------------------------------------------------
void search_init(void) 
{
	memset(&search, 0, sizeof(search));

    search.max_depth = MAX_DEPTH;
	search.min_depth = 1;
	search.normal_move_time = 10 * 1000; // 10 seconds
	search.extended_move_time = 15 * 1000; // 15 seconds;
	search.post_flag = POST_DEFAULT;
}

//-------------------------------------------------------------------------------------------------
//  Update principal variation
//-------------------------------------------------------------------------------------------------
void update_pv(MOVE move) 
{
	int     i;

	position[board.ply].pv_line[board.ply] = move;
	for (i = board.ply + 1; i < position[board.ply + 1].pv_size; i++)
		position[board.ply].pv_line[i] = position[board.ply + 1].pv_line[i];
	position[board.ply].pv_size = position[board.ply + 1].pv_size;
}

//-------------------------------------------------------------------------------------------------
//  Check if time for search has ended.
//-------------------------------------------------------------------------------------------------
void check_time(void) 
{
    UINT    current;

	if (search.nodes & TIME_CHECK)
		return;

    current = util_get_time();

    if (search.cur_depth > 1)
    {
        search.can_stop = 1;
    }

	if (search.can_stop && current >= search.extended_finish_time) 
    {
        search.stop = 1;
    }

	if (search.can_stop && search.score_drop == 0 && current >= search.normal_finish_time) 
    {
        search.stop = 1;
	}

    if (search.stop)
    {
		longjmp(env, 0);
	}
}

//-------------------------------------------------------------------------------------------------
//  Verify if moves gives check (before making the move).
//-------------------------------------------------------------------------------------------------
int is_check(MOVE move)
{
    int     mvpc = MVPC(move);
    int     tosq = TOSQ(move);
    int     frsq = FRSQ(move);
    BBIX    attack;
    int     my_color = board.side_on_move;
    int     op_color = flip_color(my_color);
    U64     opp_king = board.state[op_color].piece[KING].u64;
    int     oksq = bb_first(board.state[op_color].piece[KING]);
    U64     occup = board.occupied_squares;
    BBIX    piece;
    int     pcsq;
    int     epsq;

    bb_clear(occup, frsq);
    bb_set(occup, tosq);

    // direct attack by pieces to the king
    switch(mvpc)
    {
    case QUEEN:
        attack.u64 = diagonal_rays[tosq] & opp_king;
        if (attack.u64 && !(from_to_path[tosq][bb_first(attack)] & occup))
            return 1;
        attack.u64 = rankfile_rays[tosq] & opp_king;
        if (attack.u64 && !(from_to_path[tosq][bb_first(attack)] & occup))
            return 1;
        break;
    case ROOK:
        attack.u64 = rankfile_rays[tosq] & opp_king;
        if (attack.u64 && !(from_to_path[tosq][bb_first(attack)] & occup))
            return 1;
        break;
    case BISHOP:
        attack.u64 = diagonal_rays[tosq] & opp_king;
        if (attack.u64 && !(from_to_path[tosq][bb_first(attack)] & occup))
            return 1;
        break;
    case KNIGHT:
        if (knight_moves[tosq] & opp_king)
            return 1;
        break;
    case PAWN:
        if (pawn_attack[my_color][oksq] & bb_square[tosq])
            return 1;
        break;
    }
    // castle, promotions
    switch(TYPE(move))
    {
    case MT_CSBQS: 
        attack.u64 = rankfile_rays[D8] & opp_king;
        if (attack.u64 && !(from_to_path[D8][bb_first(attack)] & occup))
            return 1;
        else
            return 0;
    case MT_CSBKS: 
        attack.u64 = rankfile_rays[F8] & opp_king;
        if (attack.u64 && !(from_to_path[F8][bb_first(attack)] & occup))
            return 1;
        else
            return 0;
    case MT_CSWQS: 
        attack.u64 = rankfile_rays[D1] & opp_king;
        if (attack.u64 && !(from_to_path[D1][bb_first(attack)] & occup))
            return 1;
        else
            return 0;
    case MT_CSWKS:
        attack.u64 = rankfile_rays[F1] & opp_king;
        if (attack.u64 && !(from_to_path[F1][bb_first(attack)] & occup))
            return 1;
        else
            return 0;
    case MT_PROMO:
    case MT_CPPRM:
        switch(PROM(move))
        {
        case QUEEN:
            attack.u64 = diagonal_rays[tosq] & opp_king;
            if (attack.u64 && !(from_to_path[tosq][bb_first(attack)] & occup))
                return 1;
            attack.u64 = rankfile_rays[tosq] & opp_king;
            if (attack.u64 && !(from_to_path[tosq][bb_first(attack)] & occup))
                return 1;
            break;
        case ROOK:
            attack.u64 = rankfile_rays[tosq] & opp_king;
            if (attack.u64 && !(from_to_path[tosq][bb_first(attack)] & occup))
                return 1;
            break;
        case BISHOP:
            attack.u64 = diagonal_rays[tosq] & opp_king;
            if (attack.u64 && !(from_to_path[tosq][bb_first(attack)] & occup))
                return 1;
            break;
        case KNIGHT:
            if (knight_moves[tosq] & opp_king)
                return 1;
            break;
        }
        break;
    case MT_EPCAP:
        epsq = PWSQ(move);
        bb_clear(occup, epsq);
        piece.u64 = rankfile_rays[epsq] & (board.state[my_color].piece[QUEEN].u64 | board.state[my_color].piece[ROOK].u64);
        while(piece.u64)
        {
            pcsq = bb_first(piece);
            attack.u64 = rankfile_rays[pcsq] & opp_king;
            if (attack.u64 && !(from_to_path[pcsq][oksq] & occup))
                return 1;
            bb_clear(piece.u64, pcsq);
        }
        piece.u64 = diagonal_rays[epsq] & (board.state[my_color].piece[QUEEN].u64 | board.state[my_color].piece[BISHOP].u64);
        while(piece.u64)
        {
            pcsq = bb_first(piece);
            attack.u64 = diagonal_rays[pcsq] & opp_king;
            if (attack.u64 && !(from_to_path[pcsq][oksq] & occup))
                return 1;
            bb_clear(piece.u64, pcsq);
        }
        break;
    }

    // discovered checks after moving piece
    piece.u64 = rankfile_rays[frsq] & (board.state[my_color].piece[QUEEN].u64 | board.state[my_color].piece[ROOK].u64);
    while(piece.u64)
    {
        pcsq = bb_first(piece);
        attack.u64 = rankfile_rays[pcsq] & opp_king;
        if (attack.u64 && !(from_to_path[pcsq][oksq] & occup))
            return 1;
        bb_clear(piece.u64, pcsq);
    }
    piece.u64 = diagonal_rays[frsq] & (board.state[my_color].piece[QUEEN].u64 | board.state[my_color].piece[BISHOP].u64);
    while(piece.u64)
    {
        pcsq = bb_first(piece);
        attack.u64 = diagonal_rays[pcsq] & opp_king;
        if (attack.u64 && !(from_to_path[pcsq][oksq] & occup))
            return 1;
        bb_clear(piece.u64, pcsq);
    }

    return 0;
}

//-------------------------------------------------------------------------------------------------
//  Test if square is defended by pawn of color
//-------------------------------------------------------------------------------------------------
int is_defended_by_pawn(int square, int color)
{
    if (color == WHITE)
    {
        if (square >= A2)
            return 0;
        if (square_file(square) > FILEA && board.state[WHITE].square[square + 7] == PAWN)
            return 1;
        if (square_file(square) < FILEH && board.state[WHITE].square[square + 9] == PAWN)
            return 1;
    }
    else
    {
        if (square <= H7)
            return 0;
        if (square_file(square) > FILEA && board.state[BLACK].square[square - 9] == PAWN)
            return 1;
        if (square_file(square) < FILEH && board.state[BLACK].square[square - 7] == PAWN)
            return 1;
    }
    return 0;
}

//-------------------------------------------------------------------------------------------------
//  Test if move made is a free pawn. No enemy pawns in front of it.
//-------------------------------------------------------------------------------------------------
int is_free_pawn(int color, MOVE move)
{
	if (MVPC(move) != PAWN)
        return 0;
    if (TYPE(move) == MT_PROMO || TYPE(move) == MT_CPPRM)
        return 1;
    if (!(bb_pawn_pass[color][TOSQ(move)] & board.state[flip_color(color)].piece[PAWN].u64))
        return 1;
    return 0;
}

//-------------------------------------------------------------------------------------------------
//  Test if color has pawn on its rank 7
//-------------------------------------------------------------------------------------------------
int has_pawn_on_rank7(int color)
{
    U64     rank7[2] = {BB_RANK_7, BB_RANK_2};

    return board.state[color].piece[PAWN].u64 & rank7[color] ? 1 : 0;
}

//-------------------------------------------------------------------------------------------------
//  Indicate if is mate score (MATE - ply)
//-------------------------------------------------------------------------------------------------
int is_mate_score(int score)
{
    if (score > MAX_SCORE && score != INFINITY)
        return 1;
    if (score < -MAX_SCORE && score != -INFINITY)
        return 1;
    return 0;
}

//-------------------------------------------------------------------------------------------------
//  Display search information on screen.
//-------------------------------------------------------------------------------------------------
void post_info(int score, int depth)
{
	int     pvi;
	char    move_string[20];
	char    info_line[200];
	int     move_number;
	double  time;
	char    *space;

	if (search.post_flag == POST_NONE)
		return;
	if (histply % 2)
		move_number = (histply + 1) / 2;
	else
		move_number = (histply + 2) / 2;

	if (search.post_flag == POST_DEFAULT) 
	{
		time = ((float)(util_get_time() - search.start_time) / 1000.0);
		if (time < 10.0)
			space = " ";
		else
			space = "";
		sprintf(info_line, "%3d  %9lld %6.2f %s%2.1f %d.%s", depth, search.nodes + search.qnodes, 
            ((float)(board.side_on_move == BLACK ? -score : score) / 100.0), 
            space, time,
            move_number, 
            (board.side_on_move == WHITE ? "" : ".."));
#ifdef LOG
		fprintf(log_file, info_line);
#endif
	}

	if (search.post_flag == POST_XBOARD) 
	{
		sprintf(info_line, "%d %d %d %lld %d.%s", depth, (board.side_on_move == BLACK ? -score : score), 
            (util_get_time() - search.start_time) / 10, search.nodes + search.qnodes, 
            move_number, 
            (board.side_on_move == WHITE ? "" : ".."));
#ifdef LOG
		fprintf(log_file, info_line);
#endif
	}

	if (search.post_flag != POST_NONE)
    {
		printf("%s", info_line);

		for (pvi = 0; pvi < position[0].pv_size; pvi++)
        {
			if (position[0].pv_line[pvi])
            {
				util_get_move_string(position[0].pv_line[pvi], move_string);
				printf(" %s", move_string);
#ifdef LOG
				fprintf(log_file, " %s", move_string);
#endif
			}
		}
		printf("\n");
		fflush(stdout);

#ifdef LOG
		fprintf(log_file, "\n");
		fflush(log_file);
#endif
	}
}

//END
