/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Zero window search.
//-------------------------------------------------------------------------------------------------

#define FUTILITY_DEPTH 4
static const int FUTILITY_MARGIN[FUTILITY_DEPTH] = {0, 120, 240, 360};

#define STAT_NULL_DEPTH     4
#define STAT_NULL_VALUE     300
#define STAT_NULL_MARGIN    100

#define MOVE_DEPTH_PRUNE 10
static const int MOVE_COUNT_PRUNE[MOVE_DEPTH_PRUNE] = {0, 15, 19, 23, 31, 39, 51, 63, 79, 95};

//-------------------------------------------------------------------------------------------------
//  Search
//-------------------------------------------------------------------------------------------------
int search_zw(UINT incheck, int beta, int depth, UINT can_null)
{
    MOVES   ml;
	UINT    skip_null  = 0;
    MOVE    hash_move  = MOVE_NONE;
    int     best_score = -INFINITY;
    int     eval_score = -INFINITY;
    int     total_move_count = 0;
    int     quiet_move_count = 0;
	int     score = 0;
	UINT    new_incheck;
	int     extensions;
    int     ply = board.ply;
    int     som = board.side_on_move;
    MOVE    move;
	int		som_pieces_count = 0;
    MOVE    searched[MAX_MOVE];
    int     null_depth;

	assert(incheck == 0 || incheck == 1);
	assert(beta >= -INFINITY && beta <= INFINITY);
	assert(depth <= MAX_DEPTH);
	assert(can_null == 0 || can_null == 1);

    if (depth <= 0)
		return quiesce(incheck, beta - 1, beta, 0);

    if (incheck)
        return search_evasion(beta, depth);

	search.nodes++;

    if (is_draw())
		return 0;

    check_time();

    assert(board.ply >= 0 && board.ply < MAX_PLY);

	position[ply].pv_size = ply;

	// transposition table score or move hint
    if (tt_probe(depth, beta - 1, beta, &score, &hash_move, &skip_null))
        return score;

    assert(score >= -INFINITY && score <= INFINITY);
    assert(skip_null == 0 || skip_null == 1);
    
    eval_score = evaluate();
    som_pieces_count = pieces_count(som);

    if (can_null && !skip_null && !is_mate_score(beta) && som_pieces_count > 0)
    {
        // static null move pruning, position is good for side on move
        if (depth < STAT_NULL_DEPTH && eval_score - (STAT_NULL_VALUE + depth * STAT_NULL_MARGIN) >= beta)
        {
            if (!has_pawn_on_rank7(flip_color(som)) && position[ply].threats[som] >= position[ply].threats[flip_color(som)])
            {
                return eval_score - (STAT_NULL_VALUE + depth * STAT_NULL_MARGIN);
            }
        }

        // null move
        if (eval_score >= beta && depth > 1)
        {
            null_depth = NULL_DEPTH(depth);

            make_move(MOVE_NULL);
            score = -search_zw(incheck, 1 - beta, null_depth, 0);
            undo_move();

            if (score >= beta)
            {
                if (is_mate_score(score))
                    score = beta;
                tt_save(depth, score, TT_LOWER, MOVE_NONE);
                return score;
            }
        }
    }

    moves_init(&ml, incheck, hash_move, 0);
    while((move = moves_next(&ml)) != MOVE_NONE)
    {
        assert(is_valid(move));

		// extension if move puts opponent in check
        if (is_check(move))
        {
            new_incheck = 1;
            if (depth < 12)
                extensions = 1;
            else
                extensions = 0;
        }
        else
        {
            new_incheck = 0;
            extensions = 0;
        }

        // pruning or depth reductions
        if (!new_incheck && total_move_count > 1 && !is_mate_score(beta))
        {
            assert(move != hash_move);

			if (IS_QUIET(move) && !is_free_pawn(som, move) && MVPC(move) != KING && !move_is_killer(som, ply, move))
			{

                assert(depth > 0);

				// futility pruning
				if (depth < FUTILITY_DEPTH && eval_score + FUTILITY_MARGIN[depth] < beta)
				{
                    total_move_count++;
					continue;
				}
				//// move count pruning
				if (depth < MOVE_DEPTH_PRUNE && total_move_count > MOVE_COUNT_PRUNE[depth] && move_has_bad_hist(som, move))
				{
                    total_move_count++;
                    continue;
				}
				// late move reductions
				if (total_move_count > 3 && depth > 2)
				{
					extensions--;
					if (depth > 5 && move_has_bad_hist(som, move))
					{
						extensions -= (depth / 6);
					}
				}
			}
        }

	     // Make move.
		make_move(move);
		if (is_illegal(move))
        {
			undo_move(); 
			continue;
		}
        assert(board.key == zk_board_key());
        assert(board.pawn_key == zk_pawn_key());

		score = -search_zw(new_incheck, 1 - beta, depth - 1 + extensions, 1);
		if (score >= beta && extensions < 0)
			score = -search_zw(new_incheck, 1 - beta, depth - 1, 1);

        undo_move();

        // move count update
        total_move_count++;
		if (IS_QUIET(move))
            searched[quiet_move_count++] = move;

        // score
        if (score > best_score)
        {
            if (score >= beta)
            {
                if (IS_QUIET(move) && depth < 11)
                {
                    move_hist_save(som, ply, move, depth);
                    move_hist_bad(som, searched, quiet_move_count - 1);
                }
                tt_save(depth, score, TT_LOWER, move);
                return score;
            }
            best_score = score;
        }
	}

	if (total_move_count == 0)
		return 0;

    tt_save(depth, best_score, TT_UPPER, MOVE_NONE);

	return best_score;
}

// end
