/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  Static exchange evaluation functions.
//-------------------------------------------------------------------------------------------------

int     see_square(int color, int square, U64 occup, int cppc);
void    get_small_cap(int to, int *piece, int *frsq, U64 occup, int color);

//-------------------------------------------------------------------------------------------------
//  SEE for capture moves.
//-------------------------------------------------------------------------------------------------
int see_capture(MOVE move)
{
    int     value;
    U64     occup;
    
    assert(TYPE(move) == MT_CAPPC);

    if (MVPC(move) == KING)
        return 0;
    occup = board.occupied_squares;
    bb_clear(occup, FRSQ(move));
    value = see_value[CPPC(move)] - see_square(flip_color(board.side_on_move), TOSQ(move), occup, MVPC(move));
    return value;
}

//-------------------------------------------------------------------------------------------------
//  SEE for any moves.
//-------------------------------------------------------------------------------------------------
int see_move(MOVE move)
{
    int     value;
    U64     occup;
    
    if (MVPC(move) == KING)
        return 0;
    occup = board.occupied_squares;
    bb_clear(occup, FRSQ(move));
	if (TYPE(move) == MT_CAPPC || TYPE(move) == MT_CPPRM)
		value = see_value[CPPC(move)];
    else
        if (TYPE(move) == MT_EPCAP)
            value = see_value[PAWN];
	    else
    		value = 0;
    value -= see_square(flip_color(board.side_on_move), TOSQ(move), occup, MVPC(move));
    return value;
}

//-------------------------------------------------------------------------------------------------
//  SEE for square.
//-------------------------------------------------------------------------------------------------
int see_sign(int color, int frsq, int tosq)
{
    U64     occup;
    int     piece;
    int     value;
    
    piece = board.state[color].square[frsq];
    if (piece == KING)
        return 0;
    occup = board.occupied_squares;
    bb_clear(occup, frsq);
    value = 0;//see_value[board.state[flip_color(color)].square[tosq]];
    value -= see_square(flip_color(color), tosq, occup, piece);
    return value;
}

//-------------------------------------------------------------------------------------------------
//  SEE on specific square. Used by the functions above.
//-------------------------------------------------------------------------------------------------
int see_square(int color, int square, U64 occup, int cppc)
{
    int		value = 0;
    int     frsq;
    int     piece;
    int     frsq1;
    int     piece1;

    if (cppc == KING)
        return 0;
	get_small_cap(square, &piece, &frsq, occup, color);
	if (piece == 0)
		return 0;
    bb_clear(occup, frsq);
    if (piece == KING)
    {
        get_small_cap(square, &piece1, &frsq1, occup, flip_color(color));
        if (piece1 != 0)
            return 0;
    }
    value = see_value[cppc] - see_square(flip_color(color), square, occup, piece);
    if (value < 0)
        value = 0;
    return value;
}

//-------------------------------------------------------------------------------------------------
//  Return the smallest piece capturing on square.
//-------------------------------------------------------------------------------------------------
void get_small_cap(int to, int *piece, int *frsq, U64 occup, int color)
{
    BBIX	temp;
    int		idx;

	temp.u64 = pawn_attack[color][to] & occup & board.state[color].piece[PAWN].u64;
	if (temp.u64) 
    {
		*frsq = bb_first(temp);
        *piece = PAWN;
        return;
	}
	temp.u64 = knight_moves[to] & occup & board.state[color].piece[KNIGHT].u64;
	if (temp.u64)
    {
		*frsq = bb_first(temp);
        *piece = KNIGHT;
        return;
	}
	temp.u64 = board.state[color].piece[BISHOP].u64 & diagonal_rays[to] & occup;
	while(temp.u64)
    {
		idx = bb_first(temp);
		if (!(from_to_path[to][idx] & occup))
		{
            *frsq = idx;
            *piece = BISHOP;
            return;
		}
		bb_clear(temp.u64, idx);
	}
	temp.u64 = board.state[color].piece[ROOK].u64 & rankfile_rays[to] & occup;
	while(temp.u64)
    {
		idx = bb_first(temp);
		if (!(from_to_path[to][idx] & occup))
		{
            *frsq = idx;
            *piece = ROOK;
            return;
		}
		bb_clear(temp.u64, idx);
	}
	temp.u64 = board.state[color].piece[QUEEN].u64 & (rankfile_rays[to] | diagonal_rays[to]) & occup;
	while(temp.u64) 
    {
		idx = bb_first(temp);
		if (!(from_to_path[to][idx] & occup))
		{
            *frsq = idx;
            *piece = QUEEN;
            return;
		}
		bb_clear(temp.u64, idx);
	}
	temp.u64 = king_moves[to] & board.state[color].piece[KING].u64 & occup;
	if (temp.u64)
    {
		*frsq = bb_first(temp);
        *piece = KING;
        return;
	}
	*piece = 0;
}

//END