/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
// validations when running in ASSERT mode (see NDEBUG in globals.h)
//-------------------------------------------------------------------------------------------------

#ifndef NDEBUG

int valid_piece(int pcsq, int color, UINT type)
{
    if (board.state[color].square[pcsq] != type)
        return 0;
    if (!(board.state[color].piece[type].u64 & bb_square[pcsq]))
        return 0;
    return 1;
}

int pawn_is_candidate(int pcsq, int color)
{
    int     sq;
    int     own = 0;
    int     enemy = 0;

    if (pawn_is_passed(pcsq, color))
        return 0;
    if (pawn_is_isolated(pcsq, color))
        return 0;
    if (pawn_is_backward(pcsq, color))
        return 0;
    // count supporting pawns on the side or behind.
    sq = pcsq;
    while(square_rank(sq) != RANK1 && square_rank(sq) != RANK8)
    {
        if (square_file(sq) > 0 && board.state[color].square[sq - 1] == PAWN)
            own++;
        if (square_file(sq) < 7 && board.state[color].square[sq + 1] == PAWN)
            own++;
        sq = (color == WHITE ? SQ_S(sq) : SQ_N(sq));
    }
    // count enemy pawns in front.
    sq = (color == WHITE ? SQ_N(pcsq) : SQ_S(pcsq));
    while(square_rank(sq) != RANK1 && square_rank(sq) != RANK8)
    {
        if (board.state[flip_color(color)].square[sq] == PAWN)
            return 0;
        if (square_file(sq) > 0 && board.state[flip_color(color)].square[sq - 1] == PAWN)
            enemy++;
        if (square_file(sq) < 7 && board.state[flip_color(color)].square[sq + 1] == PAWN)
            enemy++;
        sq = (color == WHITE ? SQ_N(sq) : SQ_S(sq));
    }
    if (own >= enemy)
        return 1;
    return 0;
}

int pawn_is_backward(int pcsq, int color)
{
    int     sq = (color == WHITE ? SQ_N(pcsq) : SQ_S(pcsq));
    int     own = 0;
    int     enemy = 0;

    if (pawn_is_connected(pcsq, color))
        return 0;
    if (pawn_is_passed(pcsq, color))
        return 0;
    if (!pawn_is_weak(pcsq, color))
        return 0;
    if (pawn_is_isolated(pcsq, color))
        return 0;

    //if (color == board.side_on_move)
    {
        if (square_file(sq) > 0 && board.state[flip_color(color)].square[sq - 1] == PAWN)
            return 0;
        if (square_file(sq) < 7 && board.state[flip_color(color)].square[sq + 1] == PAWN)
            return 0;
    }

    while(square_rank(sq) != RANK1 && square_rank(sq) != RANK8)
    {
        if (square_file(sq) > 0)
        {
            if (board.state[color].square[sq - 1] == PAWN)
                own = 1;
            if (board.state[flip_color(color)].square[sq - 1] == PAWN)
                enemy = 1;
        }
        if (square_file(sq) < 7)
        {
            if (board.state[color].square[sq + 1] == PAWN)
                own = 1;
            if (board.state[flip_color(color)].square[sq + 1] == PAWN)
                enemy = 1;
        }
        if (own)
            return 0;
        if (enemy)
            return 1;
        sq = (color == WHITE ? SQ_N(sq) : SQ_S(sq));
    }
    return 0;
}

int pawn_is_isolated(int pcsq, int color)
{
    int     sq = square_file(pcsq) + 8;

    while(square_rank(sq) < 7)
    {
        if (square_file(sq) > 0 && board.state[color].square[sq - 1] == PAWN)
            return 0;
        if (square_file(sq) < 7 && board.state[color].square[sq + 1] == PAWN)
            return 0;
        sq += 8;
    }
    return 1;
}

int pawn_is_weak(int pcsq, int color)
{
    int     sq = pcsq;

    while(square_rank(sq) != RANK1 && square_rank(sq) != RANK8)
    {
        if (square_file(sq) > 0)
        {
            if (board.state[color].square[sq - 1] == PAWN)
                return 0;
        }
        if (square_file(sq) < 7)
        {
            if (board.state[color].square[sq + 1] == PAWN)
                return 0;
        }
        sq = (color == WHITE ? SQ_S(sq) : SQ_N(sq));
    }
    return 1;
}

int pawn_is_passed(int pcsq, int color)
{
    int     sq = (color == WHITE ? SQ_N(pcsq) : SQ_S(pcsq));

    while(square_rank(sq) != RANK1 && square_rank(sq) != RANK8)
    {
        if (square_file(sq) > 0)
        {
            if (board.state[flip_color(color)].square[sq - 1] == PAWN)
                return 0;
        }
        if (square_file(sq) < 7)
        {
            if (board.state[flip_color(color)].square[sq + 1] == PAWN)
                return 0;
        }
        if (board.state[flip_color(color)].square[sq] == PAWN)
            return 0;
        if (board.state[color].square[sq] == PAWN)
            return 0;
        sq = (color == WHITE ? SQ_N(sq) : SQ_S(sq));
    }
    return 1;
}

int pawn_is_doubled(int pcsq, int color)
{
    if (square_rank(pcsq) == RANK1 || square_rank(pcsq) == RANK8)
        return 0;
    if (color == WHITE)
    {
        pcsq = SQ_N(pcsq);
        while(valid_square(pcsq))
        {
            if (board.state[color].square[pcsq] == PAWN)
                return 1;
            pcsq = SQ_N(pcsq);
        }
    }
    else
    {
        pcsq = SQ_S(pcsq);
        while(valid_square(pcsq))
        {
            if (board.state[color].square[pcsq] == PAWN)
                return 1;
            pcsq = SQ_S(pcsq);
        }
    }
    return 0;
}

int pawn_is_connected(int pcsq, int color)
{
    if (square_rank(pcsq) == RANK1 || square_rank(pcsq) == RANK8)
        return 0;
    if (square_file(pcsq) > 0)
    {
        if (board.state[color].square[SQ_NW(pcsq)] == PAWN)
        {
            assert(valid_piece(SQ_NW(pcsq), color, PAWN));
            return 1;
        }
        if (board.state[color].square[SQ_W(pcsq)] == PAWN)
        {
            assert(valid_piece(SQ_W(pcsq), color, PAWN));
            return 1;
        }
        if (board.state[color].square[SQ_SW(pcsq)] == PAWN)
        {
            assert(valid_piece(SQ_SW(pcsq), color, PAWN));
            return 1;
        }
    }
    if (square_file(pcsq) < 7)
    {
        if (board.state[color].square[SQ_NE(pcsq)] == PAWN)
        {
            assert(valid_piece(SQ_NE(pcsq), color, PAWN));
            return 1;
        }
        if (board.state[color].square[SQ_E(pcsq)] == PAWN)
        {
            assert(valid_piece(SQ_E(pcsq), color, PAWN));
            return 1;
        }
        if (board.state[color].square[SQ_SE(pcsq)] == PAWN)
        {
            assert(valid_piece(SQ_SE(pcsq), color, PAWN));
            return 1;
        }
    }
    return 0;
}

int valid_rank_file(void)
{
    int r,f;

    // square<->rank/file translation
	for (r = 0; r < 8; r++)
    {
		for (f = 0; f < 8; f++)
        {
			if (square_rank(r * 8 + f) != r)
            {
                printf("assert rank(%d) %d != %d \n", r * 8 + f, r, square_rank(r * 8 + f));
                return 0;
            }
			if (square_file(r * 8 + f) != f)
            {
                printf("assert file(%d) %d != %d \n", r * 8 + f, r, square_rank(r * 8 + f));
                return 0;
            }
		}
    }

    return 1;
}

//  Make sure board material representation is correct.
int pieces_value(int color)
{
    return (bb_count(board.state[color].piece[QUEEN])  * VALUE_QUEEN) +
           (bb_count(board.state[color].piece[ROOK])   * VALUE_ROOK) +
           (bb_count(board.state[color].piece[KNIGHT]) * VALUE_KNIGHT) +
           (bb_count(board.state[color].piece[BISHOP]) * VALUE_BISHOP);
}

int pawns_value(int color)
{
    return (bb_count(board.state[color].piece[PAWN]) * VALUE_PAWN);
}

int valid_material(void)
{
	int s, wm, bm;

	for (s = wm = bm = 0; s < 64; s++)
	{
		if (board.state[WHITE].square[s])
			wm += piece_value[board.state[WHITE].square[s]];
		if (board.state[BLACK].square[s])
			bm += piece_value[board.state[BLACK].square[s]];
	}

    if (wm != (pieces_value(WHITE) + pawns_value(WHITE)))
	{
		printf("white material error\n");
		return 0;
	}
	if (bm != (pieces_value(BLACK) + pawns_value(BLACK)))
	{
		printf("black material error\n");
		return 0;
	}
    if (wm != material_value(WHITE))
    {
		printf("white material error (wm != material(WHITE))\n");
		return 0;
    }
    if (bm != material_value(BLACK))
    {
		printf("black material error (bm != material(BLACK))\n");
		return 0;
    }
	return 1;
}

//  Assert valid square
int valid_square(int square)
{
    return square >= 0 && square < 64 ? 1 : 0;
}

//  Assert valid color
int valid_color(int color)
{
    return color == WHITE || color == BLACK ? 1 : 0;
}

int board_state_is_ok()
{
	//	int 		can_castle_ks;		// king side
	//	int 		can_castle_qs;		// queen side
	//	U64			piece[7];			// bb for each piece type
	//	U64			all_pieces;			// bb for all pieces combined
	//	int 		square[64];			// piece location on the board
	//	int			material;			// piece material value 
	//	int			pawn_mat;			// pawn material value
	//}				player_state[2];	// 0-white, 1-black

    int     i;
    BBIX    temp;
    int     c;
    int     t;
    int     m, p;

    for (c = 0; c < 2; c++)
    {
        temp.u64 = m = p = 0;
        for (i = 0; i < 7; i++)
            temp.u64 |= board.state[c].piece[i].u64;
        if (temp.u64 != board.state[c].all_pieces)
        {
            board_print(NULL);
            printf("piece[7] != all_pieces for player: %d\n", c);
            return 0;
        }

        // king
        if (bb_count(board.state[c].piece[KING]) != 1)
        {
            board_print(NULL);
            printf("king(%d) count is not 1\n", c);
            return 0;
        }
        if (bb_first(board.state[c].piece[KING]) != (SINT)board.state[c].king_square)
        {
            board_print(NULL);
            printf("king(%d) square does not match\n", c);
            return 0;
        }

        // pieces
        for (t = PAWN; t < NUM_PIECES; t++)
        {
            temp.u64 = board.state[c].piece[t].u64;
            while(temp.u64)
            {
                if (board.state[c].square[bb_first(temp)] != (UINT)t)
                {
                    board_print(NULL);
                    printf("player %c piece %d not on square: %d\n", c, t, bb_first(temp));
                    return 0;
                }
                bb_clear(temp.u64, bb_first(temp));
                if (t == PAWN)
                    p += piece_value[t];
                else
                    m += piece_value[t];
            }
        }
        if (p != pawns_value(c))
        {
            board_print(NULL);
            printf("diff pawn mat: %d calc: %d  board: %d\n", c, p, pawns_value(c));
            return 0;
        }
        if (m != pieces_value(c))
        {
            board_print(NULL);
            printf("diff mat: color: %d calc: %d board: %d\n", c, m, pieces_value(c));
            return 0;
        }

        for (i = 0; i < 64; i++)
        {
            if (board.state[c].square[i])
            {
                if (board.state[c].square[i] < 1 || board.state[c].square[i] > 6)
                {
                    board_print(NULL);
                    printf("player %c invalid piece: type %d square: %d\n", c, board.state[c].square[i], i);
                    return 0;
                }
                if (!bb_square[i] & board.state[c].piece[board.state[c].square[i]].u64)
                {
                    board_print(NULL);
                    printf("player %c invalid square: type %d square %d\n", c, board.state[c].square[i], i);
                    return 0;
                }
            }
        }
    }

    return 1;

}

#endif
