/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//  test the evaluation function to make sure is symmetrical (testing).
//-------------------------------------------------------------------------------------------------

void set_piece(int color, int type, int tosq);
void remove_piece(int color, int type, int frsq);
void eval_test_print(int score, char *text);
void flip_ranks(void);
void flip_files(void);

void eval_test(char *file)
{
	FILE	*f;
	char	line[1000];
	char	fen[1000];
	int		i;
	int		count = 0;
	int		correct_ranks = 0;
	int		correct_files = 0;
	int		eval_normal;
	int		eval_ranks;
	int		eval_files;
	
	f = fopen(file, "r");
	if (f == NULL)
    {
		printf("cannot open epd file: %s\n", file);
		return;
	}

	while(fgets(line, 1000, f) != NULL)
    {
		count++;
		i = 0;
		fen[i] = 0;
		while(i < 1000 && line[i] && line[i] != '\n' && strncmp(&line[i], " bm ", 4))
        {
			fen[i] = line[i];
			i++;
			fen[i] = 0;
		}
        fen[i] = 0;


        new_game(fen);
        printf("/-------------------------------------------------------------------\\\n");
        printf("game %3d %s\n", count, fen);
        board_print(NULL);

		eval_normal = evaluate();
        eval_test_print(eval_normal, "NORMAL");
        eval_print();

        flip_ranks();
		eval_ranks = evaluate();
		if (eval_normal == eval_ranks)
			correct_ranks++;
		else
        {
            eval_test_print(eval_ranks, "RANKS");
            eval_print();
            board_print("flip_ranks");
        }

        flip_files();
		eval_files = evaluate();
		if (eval_normal == eval_files)
			correct_files++;
		else
        {
            eval_test_print(eval_files, "FILES");
            eval_print();
            board_print("flip_files");
        }
        printf("\\-------------------------------------------------------------------/\n");
	}

    printf("number of tests: %d   ranks: %d  files: %d\n", count, correct_ranks, correct_files);

	fclose(f);
}

int get_piece_color(int square)
{
    if (board.state[WHITE].square[square] != EMPTY)
        return WHITE;
    if (board.state[BLACK].square[square] != EMPTY)
        return BLACK;
    return -1;
}

int get_piece_type(int square)
{
    if (board.state[WHITE].square[square] != EMPTY)
        return board.state[WHITE].square[square];
    if (board.state[BLACK].square[square] != EMPTY)
        return board.state[BLACK].square[square];
    return EMPTY;
}

void flip_ranks(void)
{
    int r, f;
    int piece_color_from, piece_type_from;
    int piece_color_to, piece_type_to;
    int square_from, square_to;
    //int opp_color[2] = {BLACK, WHITE};
    int castle_qs, castle_ks;

    castle_ks = board.state[WHITE].can_castle_ks;
    castle_qs = board.state[WHITE].can_castle_qs;

    board.state[WHITE].can_castle_ks = board.state[BLACK].can_castle_ks;
    board.state[WHITE].can_castle_qs = board.state[BLACK].can_castle_qs;

    board.state[BLACK].can_castle_ks = castle_ks;
    board.state[BLACK].can_castle_qs = castle_qs;

	board.side_on_move = flip_color(board.side_on_move);

    for (r = 0; r < 4; r++)
    {
        for (f = 0; f < 8; f++)
        {
            square_from = r * 8 + f;
            square_to = (7 - r) * 8 + f;
            piece_color_from = get_piece_color(square_from);
            piece_type_from = get_piece_type(square_from);
            piece_color_to = get_piece_color(square_to);
            piece_type_to = get_piece_type(square_to);
            if (piece_color_from != -1)
                remove_piece(piece_color_from, piece_type_from, square_from);
            if (piece_color_to != -1)
                remove_piece(piece_color_to, piece_type_to, square_to);
            if (piece_color_from != -1)
                set_piece(flip_color(piece_color_from), piece_type_from, square_to);
            if (piece_color_to != -1)
                set_piece(flip_color(piece_color_to), piece_type_to, square_from);
        }
    }

    board.occupied_squares = board.state[WHITE].all_pieces | board.state[BLACK].all_pieces;
    board.empty_squares = ~board.occupied_squares;
    board.key = zk_board_key();
    board.pawn_key = zk_pawn_key();
	board.state[WHITE].king_square = bb_first(board.state[WHITE].piece[KING]);
	board.state[BLACK].king_square = bb_first(board.state[BLACK].piece[KING]);
}

void flip_files(void)
{
    int r, f;
    int piece_color_from, piece_type_from;
    int piece_color_to, piece_type_to;
    int square_from, square_to;

    board.state[WHITE].can_castle_ks = 0;
    board.state[WHITE].can_castle_qs = 0;

    board.state[BLACK].can_castle_ks = 0;
    board.state[BLACK].can_castle_qs = 0;

    for (r = 0; r < 8; r++)
    {
        for (f = 0; f < 4; f++)
        {
            square_from = r * 8 + f;
            square_to = r * 8 + (7 - f);
            piece_color_from = get_piece_color(square_from);
            piece_type_from = get_piece_type(square_from);
            piece_color_to = get_piece_color(square_to);
            piece_type_to = get_piece_type(square_to);
            if (piece_color_from != -1)
                remove_piece(piece_color_from, piece_type_from, square_from);
            if (piece_color_to != -1)
                remove_piece(piece_color_to, piece_type_to, square_to);
            if (piece_color_from != -1)
                set_piece(piece_color_from, piece_type_from, square_to);
            if (piece_color_to != -1)
                set_piece(piece_color_to, piece_type_to, square_from);
        }
    }

    board.occupied_squares = board.state[WHITE].all_pieces | board.state[BLACK].all_pieces;
    board.empty_squares = ~board.occupied_squares;
    board.key = zk_board_key();
    board.pawn_key = zk_pawn_key();
	board.state[WHITE].king_square = bb_first(board.state[WHITE].piece[KING]);
	board.state[BLACK].king_square = bb_first(board.state[BLACK].piece[KING]);
}

void eval_test_print(int score, char *text)
{
    printf("******* %s: %d\n", text, score);
    //printf("       eval qrbn: %d\n", eval_qrbn());
    //printf("       eval king: %d\n", eval_king());
    //printf("       eval pawn: %d\n", eval_pawn());
    //printf("   eval-mobility: white: %d  black: %d\n", get_mobility(WHITE, BLACK), get_mobility(BLACK, WHITE));
    //printf("     qr-mobility: white: %d  black: %d\n", get_qr_mobility(WHITE, BLACK), get_qr_mobility(BLACK, WHITE));
    //printf("     qb-mobility: white: %d  black: %d\n", get_qb_mobility(WHITE, BLACK), get_qb_mobility(BLACK, WHITE));
    //printf("      n-mobility: white: %d  black: %d\n", get_knight_mobility(WHITE, BLACK), get_knight_mobility(BLACK, WHITE));
}

//end