/*-------------------------------------------------------------------------------
  tucano is a XBoard chess playing engine developed by Alcides Schulz.
  Copyright (C) 2011-present - Alcides Schulz

  tucano is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  tucano is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You can find the GNU General Public License at http://www.gnu.org/licenses/
-------------------------------------------------------------------------------*/

#include "globals.h"

//-------------------------------------------------------------------------------------------------
//	Several utilities used by the engine.
//-------------------------------------------------------------------------------------------------

static const char *FILE_LETTER = "abcdefgh";
static const char *RANK_NUMBER = "87654321";

#define file_letter(sq)    FILE_LETTER[square_file((sq))]
#define rank_number(sq)    RANK_NUMBER[square_rank((sq))]

//-------------------------------------------------------------------------------------------------
//	Print move.
//-------------------------------------------------------------------------------------------------
void util_print_move(MOVE move, int new_line)
{
	char str[50];
	util_get_move_string(move, str);
	printf("%s", str);
    if (new_line)
        printf("\n");
}

//-------------------------------------------------------------------------------------------------
//	Return move notation.
//-------------------------------------------------------------------------------------------------
void util_get_move_string(MOVE move, char *string)
{
	char promo_piece;

	if (TYPE(move) == MT_PROMO || TYPE(move) == MT_CPPRM)
    {
		promo_piece = promo_letter[PROM(move)];
		sprintf(string, "%c%c%c%c%c", file_letter(FRSQ(move)), rank_number(FRSQ(move)),
		                              file_letter(TOSQ(move)), rank_number(TOSQ(move)),
									  promo_piece);
	}
	else
		sprintf(string, "%c%c%c%c", file_letter(FRSQ(move)), rank_number(FRSQ(move)),
   	                                file_letter(TOSQ(move)), rank_number(TOSQ(move)));
}

//-------------------------------------------------------------------------------------------------
//	Return move description.
//-------------------------------------------------------------------------------------------------
void util_get_move_desc(MOVE move, char *string, int inc_file)
{
	char promo_piece;
	char moving_piece[2];

	if (MVPC(move) == PAWN)
		moving_piece[0] = 0;
	else
    {
		moving_piece[0] = piece_letter[MVPC(move)];
		moving_piece[1] = 0;
	}
	if (TYPE(move) == MT_PROMO || TYPE(move) == MT_CPPRM)
    {
		promo_piece = promo_letter[PROM(move)];
		sprintf(string, "%s%c%c%c", moving_piece,
		                            file_letter(TOSQ(move)), rank_number(TOSQ(move)),
									promo_piece);
        return;
	}
    if (TYPE(move) == MT_CSWKS || TYPE(move) == MT_CSBKS)
    {
        sprintf(string, "O-O");
        return;
    }
    if (TYPE(move) == MT_CSWQS || TYPE(move) == MT_CSBQS)
    {
        sprintf(string, "O-O-O");
        return;
    }
    if (inc_file)
        sprintf(string, "%s%c%c%c", moving_piece, file_letter(FRSQ(move)), file_letter(TOSQ(move)), rank_number(TOSQ(move)));
    else
        sprintf(string, "%s%c%c", moving_piece, file_letter(TOSQ(move)), rank_number(TOSQ(move)));
}

//-------------------------------------------------------------------------------------------------
//  Parse a move from string and create a MOVE.
//-------------------------------------------------------------------------------------------------
void util_parse_move(char *move_string, MOVE *parsed)
{
	UINT	from;
	UINT	to;
	MOVE    move;
	int     promo_piece;
	int     temp_piece;
    MOVES   ml;

	*parsed = 0;

	if (strlen(move_string) < 4)
		return;

	if (move_string[0] < 'a' || move_string[0] > 'h' || move_string[1] < '0' || move_string[1] > '8' ||
		move_string[2] < 'a' || move_string[2] > 'h' || move_string[3] < '0' || move_string[3] > '8')
		return;

	from = ('8' - move_string[1]) * 8 + (move_string[0] - 'a');
	to = ('8' - move_string[3]) * 8 + (move_string[2] - 'a');

	moves_init(&ml, 0, 0, 0);
	while((move = moves_next(&ml)) != MOVE_NONE)
    {
		if (FRSQ(move) == from && TOSQ(move) == to)
        {
			if (TYPE(move) == MT_PROMO || TYPE(move) == MT_CPPRM)
            {
				promo_piece = promo_letter[PROM(move)];
				if (strlen(move_string) < 5)
					temp_piece = promo_letter[QUEEN];
				else
					temp_piece = tolower(move_string[4]);
				if (promo_piece != temp_piece)
					continue;
			}
			*parsed = move;
			break;
		}
	}

}
//-------------------------------------------------------------------------------------------------
//  Utility to print a bitboard to screen.
//-------------------------------------------------------------------------------------------------
void bb_print(char *msg, U64 b)
{
	int     i, j;
	char    s1[200];
	char    s2[200];
	int	    li;
	int     fi;
	int     bc;
    BBIX    bbix;

    bbix.u64 = b;

	li = bb_last(bbix);
	fi = bb_first(bbix);
	bc = bb_count(bbix);

	memset(s1, 0, sizeof(s1));
	memset(s2, 0, sizeof(s2));

	for (i = 0; i < 64; i++) {
		if (b & 1)
			s1[63 - i] = '1';
		else
			s1[63 - i] = '0';
		b >>= 1;
	}

	for (i = j = 0; i < 64; i++, j++) {
		s2[j] = s1[i];
		if (i > 0 && i < 56 && (((i+1) % 8) == 0))
			s2[++j] = '\n';
	}

	//printf("%s %s fi=%d li=%d\n", m, s2, fi, li);
    printf("%s firstindex=%d lastindex=%d  bitcount=%d Hex: ((U64)0x%016llX)\n", msg, fi, li, bc, bbix.u64);
	printf("%s\n", s2);
}

//-------------------------------------------------------------------------------------------------
//  Print current board.
//-------------------------------------------------------------------------------------------------
void board_print(char *text)
{
	char    b[64];
	int     r, f, i;
	BBIX    piece;
	int	    index;
	char    p[100];
    char    fen[1024];
    int     idx = 0;

	memset(b, ' ', sizeof(b));

	if (board.state[BLACK].piece[KING].u64)
		b[bb_last(board.state[BLACK].piece[KING])] = 'k';
	if (board.state[WHITE].piece[KING].u64)
		b[bb_last(board.state[WHITE].piece[KING])] = 'K';
	
	piece.u64 = board.state[BLACK].piece[KNIGHT].u64;
	while (piece.u64) 
    {
		index = bb_first(piece);
		b[index] = 'n';
		bb_clear(piece.u64, index);
	}
	piece.u64 = board.state[BLACK].piece[ROOK].u64;
	while (piece.u64)
    {
		index = bb_first(piece);
		b[index] = 'r';
		bb_clear(piece.u64, index);
	}
	piece.u64 = board.state[BLACK].piece[BISHOP].u64;
	while (piece.u64)
    {
		index = bb_first(piece);
		b[index] = 'b';
		bb_clear(piece.u64, index);
	}
	piece.u64 = board.state[BLACK].piece[QUEEN].u64;
	while (piece.u64)
    {
		index = bb_first(piece);
		b[index] = 'q';
		bb_clear(piece.u64, index);
	}
	piece.u64 = board.state[BLACK].piece[PAWN].u64;
	while (piece.u64)
    {
		index = bb_first(piece);
		b[index] = 'p';
		bb_clear(piece.u64, index);
	}
	piece.u64 = board.state[WHITE].piece[KNIGHT].u64;
	while (piece.u64)
    {
		index = bb_first(piece);
		b[index] = 'N';
		bb_clear(piece.u64, index);
	}
	piece.u64 = board.state[WHITE].piece[ROOK].u64;
	while (piece.u64)
    {
		index = bb_first(piece);
		b[index] = 'R';
		bb_clear(piece.u64, index);
	}
	piece.u64 = board.state[WHITE].piece[BISHOP].u64;
	while (piece.u64)
    {
		index = bb_first(piece);
		b[index] = 'B';
		bb_clear(piece.u64, index);
	}
	piece.u64 = board.state[WHITE].piece[QUEEN].u64;
	while (piece.u64)
    {
		index = bb_first(piece);
		b[index] = 'Q';
		bb_clear(piece.u64, index);
	}
	piece.u64 = board.state[WHITE].piece[PAWN].u64;
	while (piece.u64)
    {
		index = bb_first(piece);
		b[index] = 'P';
		bb_clear(piece.u64, index);
	}
	i = 0;
	if (text)
    {
        printf("%s\n", text);
	}
	printf("       +---+---+---+---+---+---+---+---+\n");
	for (r = 0; r < 8; r++)
    {
		printf("%d (%2d) ", 8 - r, r * 8);
		for (f = 0; f < 8; f++)
        {
			printf("| %c ", b[i++]);
		}
		printf("|\n       +---+---+---+---+---+---+---+---+\n");
	}
	printf("         a   b   c   d   e   f   g   h  \n");

    // fen
    f = 0;
    for (i = 0, r = 0; i < 64; i++, r++)
    {
        if (b[i] == ' ')
            f++;
        else
        {
            if (f)
                fen[idx++] = '0' + (char)f;
            fen[idx++] = b[i];
            f = 0;
        }
        if (r == 7)
        {
            if (f)
                fen[idx++] = '0' + (char)f;
            f = 0;
            if (i != 63)
                fen[idx++] = '/';
            r = -1;
        }
    }
    fen[idx++] = ' ';
    if (board.side_on_move == WHITE)
        fen[idx++] = 'w';
    else
        fen[idx++] = 'b';
    fen[idx++] = ' ';
    if (board.state[WHITE].can_castle_ks || board.state[WHITE].can_castle_qs || board.state[BLACK].can_castle_ks || board.state[BLACK].can_castle_qs)
    {
        if (board.state[WHITE].can_castle_ks)
            fen[idx++] = 'K';
        if (board.state[WHITE].can_castle_qs)
            fen[idx++] = 'Q';
        if (board.state[BLACK].can_castle_ks)
            fen[idx++] = 'k';
        if (board.state[BLACK].can_castle_qs)
            fen[idx++] = 'q';
    }
    else
    {
        fen[idx++] = '-';
    }
    fen[idx++] = ' ';
    if (board.ep_square.u64)
    {
        fen[idx++] = '-';
        fen[idx++] = file_letter(bb_first(board.ep_square));
        fen[idx++] = rank_number(bb_first(board.ep_square));
    }
    else
    {
        fen[idx++] = '-';
    }   

    fen[idx] = 0;
    printf("%s\n", fen);

	if (board.side_on_move == WHITE)
		printf("White to move  ");
	else
		printf("Black to move  ");

    printf("Ply: %d HistPly: %d", board.ply, histply);
	printf("     Castle ");
	printf("%c", board.state[WHITE].can_castle_ks ? 'K' : ' ');
	printf("%c", board.state[WHITE].can_castle_qs ? 'Q' : ' ');
	printf("%c", board.state[BLACK].can_castle_ks ? 'k' : ' ');
	printf("%c", board.state[BLACK].can_castle_qs ? 'q' : ' ');
	printf("  Player To Move in Check?: %d ", is_incheck(board.side_on_move));
	printf("\n");
	printf("-----------------------------------------------------------------\n");
	if (text)
		gets(p);
}

//-------------------------------------------------------------------------------------------------
//  Print current move list.
//-------------------------------------------------------------------------------------------------
void print_current_moves(void) 
{
	int     i = 1;
    int     j = 0;
    MOVE    move;
    MOVES   ml;
    
    moves_init(&ml, is_incheck(board.side_on_move), MOVE_NONE, 0);

	while((move = moves_next(&ml)) != MOVE_NONE)
    {
        if (!is_valid(move))
        {
            util_print_move(move, 1);
            board_print("invalid_move");
        }
		printf("%2d) %s", i, rank_file_notation[FRSQ(move)]);
		if (TYPE(move) == MT_CAPPC || TYPE(move) == MT_EPCAP)
			printf("x");
		else
			printf("-");
		printf("%s %9d", rank_file_notation[TOSQ(move)], 0);
		if (j == 3) {
			printf("\n");
			j = -1;
		}
		else 
			printf(" ");
        i++;
        j++;
	}
	printf("\n");
}

#if defined(_MSC_VER) || defined(_WINDOWS_)

#pragma warning (disable : 4820)
#pragma warning (disable : 4255)
#pragma warning (disable : 4668)

#include <Windows.h>

#pragma warning (default : 4820)
#pragma warning (default : 4668)
#pragma warning (default : 4255)

//-------------------------------------------------------------------------------------------------
//  Current time (Windows).
//-------------------------------------------------------------------------------------------------
UINT util_get_time(void)
{
    FILETIME ft;
    GetSystemTimeAsFileTime(&ft);
    return (unsigned int) ((((U64)ft.dwHighDateTime << 32) | ft.dwLowDateTime) / 10000);
}
#else
#include <sys/time.h>

//-------------------------------------------------------------------------------------------------
//  Current time.
//-------------------------------------------------------------------------------------------------
UINT util_get_time(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}
#endif

//end